// このソースはパブリックドメインです。
// 作成 徳丸浩 2008/01/24 
// 利用にあたり、function pl0Println(a) を表示関数として定義してください
//
var IDENT     = 1;
var NUMBER    = 2;

var R_PROGRAM   = 3;
var R_PROCEDURE = 4;
var R_BEGIN     = 5;
var R_END       = 6;
var R_VAR       = 7;
var R_CALL      = 8;
var R_IF        = 9;
var R_THEN      = 10;
var R_WHILE     = 11;
var R_DO        = 12;
var R_ODD       = 18;
var R_PRINT     = 19;

var SEMICOLON = 20;
var COMMA     = 21;
var PERIOD    = 22;
var LPAREN    = 23;
var RPAREN    = 24;

var OP_BECOMES = 30;
var OP_ADD     = 31;
var OP_MUL     = 32;
var OP_COMP    = 33;

var keywords = {
	"begin" 	: R_BEGIN,
	"end" 		: R_END,
	"program"	: R_PROGRAM,
	"procedure"	: R_PROCEDURE,
	"var"		: R_VAR,
	"if"		: R_IF,
	"then"		: R_THEN,
	"while"		: R_WHILE,
	"do"		: R_DO,
	"call"		: R_CALL,
	"odd"		: R_ODD,
	"print"		: R_PRINT,
};


var F_LIT = 1;
var F_OPR = 2;
var F_LOD = 3;
var F_STO = 4;
var F_CAL = 5;
var F_INT = 6;
var F_JMP = 7;
var F_JPC = 8;

var codeName = ["NOP", "LIT", "OPR", "LOD", "STO", "CAL", "INT", "JMP", "JPC"];

var V_LES =  1;
var V_GTR =  2;
var V_EQL =  3;
var V_NEQ =  4;
var V_LEQ =  5;
var V_GEQ =  6;
var V_ADD =  7;
var V_SUB =  8;
var V_MUL =  9;
var V_DIV = 10;
var V_RET = 11;
var V_PRT = 12;
var V_ODD = 13;
var V_NEG = 14;

var opName = ["", "<", ">", "=", "<>", "<=", ">=", "+", "-", "*", "/", "ret", "print", "odd", "-"];
var optable = {
	"<" : V_LES,
	">" : V_GTR,
	"=" : V_EQL,
	"<>" : V_NEQ,
	"<=" : V_LEQ,
	">=" : V_GEQ,
	"+" : V_ADD,
	"-" : V_SUB,
	"*" : V_MUL,
	"/" : V_DIV,
}

var K_VARIABLE  = 1;
var K_PROCEDURE = 2;
var K_CONST     = 3;

var OTHERS = 98;
var EOF    = 99;

var pTable = [
	{	pattern : /^([a-z_][a-z0-9_]*)/i,	kind : IDENT	},
	{	pattern : /^([0-9]+)/,				kind : NUMBER	},
	{	pattern : /^(:=)/,					kind : OP_BECOMES},
	{	pattern : /^(>=)/,					kind : OP_COMP	},
	{	pattern : /^(<=)/,					kind : OP_COMP	},
	{	pattern : /^(=)/,					kind : OP_COMP	},
	{	pattern : /^(>)/,					kind : OP_COMP	},
	{	pattern : /^(<)/,					kind : OP_COMP	},
	{	pattern : /^(\+)/,					kind : OP_ADD	},
	{	pattern : /^(\-)/,					kind : OP_ADD	},
	{	pattern : /^(\*)/,					kind : OP_MUL	},
	{	pattern : /^(\/)/,					kind : OP_MUL	},
	{	pattern : /^(;)/,					kind : SEMICOLON},
	{	pattern : /^(,)/,					kind : COMMA	},
	{	pattern : /^(\.)/,					kind : PERIOD	},
	{	pattern : /^(\()/,					kind : LPAREN	},
	{	pattern : /^(\))/,					kind : RPAREN	},
	{	pattern : /^(.)/,					kind : OTHERS	},
	{	pattern : /^/,						kind : EOF	},
];

function Tokenizer(s) {
	this.src = s;
	this.token = 0;
	this.value = "";
}

Tokenizer.prototype.getToken = function() {
	this.src = this.src.replace(/^[ \t\r\n]+/, "");
	this.token = '';
	for (p in pTable) {
		if (this.src.match(pTable[p].pattern)) {
			this.src = this.src.replace(pTable[p].pattern, "");
			this.value = RegExp.$1;
			this.token = pTable[p].kind;
			break;
		}
	}
	if (this.token == IDENT && keywords[this.value]) {
		this.token = keywords[this.value];
	}
	return this.token;
}

function Identifier(name, addr, kind, level) {
	this.name = name;
	this.addr = addr;
	this.kind = kind;
	this.level = level;
}

function PCode(f, L, a) {
	this.func  = f;			// function code
	this.level = L;			// block level
	this.addr  = a;			// displacement address
}

function VM() {
	this.codeAlloc = 0;
	this.code = [];			// Instruction Code table
}

VM.prototype.gen = function(op, level, addr) {
	var cp = this.codeAlloc++;
	this.code[cp] = new PCode(op, level, addr);
	return cp;
}

// static link をたどる
function base(bp, level, stack) {
	for (var b1 = bp; level > 0; level--)
		b1 = stack[b1];
	return b1;
}

VM.prototype.interprete = function(op, level, addr) {
	var sp = 0;				// stack pointer
	var bp = 1;				// base pointer
	var pc = 0;				// program counter
	var stack = [0, 0, 0, 0];
	pl0Println("*** START PL/0 ***");
	do {
		var pcode = this.code[pc++];
		switch (pcode.func) {
		case F_LIT:
			stack[++sp] = pcode.addr;
			break;
		case F_OPR:
			switch(pcode.addr) {
			case V_NEG: stack[sp] = -stack[sp]; break;
			case V_ODD: stack[sp] &= 1; break;
			case V_ADD: sp--; stack[sp] += stack[sp + 1]; break;
			case V_SUB: sp--; stack[sp] -= stack[sp + 1]; break;
			case V_MUL: sp--; stack[sp] *= stack[sp + 1]; break;
			case V_DIV: sp--; stack[sp] /= stack[sp + 1]; break;
			case V_LES: sp--; stack[sp] = stack[sp] < stack[sp + 1]; break;
			case V_GTR: sp--; stack[sp] = stack[sp] > stack[sp + 1]; break;
			case V_EQL: sp--; stack[sp] = stack[sp] == stack[sp + 1]; break;
			case V_NEQ: sp--; stack[sp] = stack[sp] != stack[sp + 1]; break;
			case V_LEQ: sp--; stack[sp] = stack[sp] <= stack[sp + 1]; break;
			case V_GEQ: sp--; stack[sp] = stack[sp] >= stack[sp + 1]; break;
			case V_PRT: pl0Println(stack[sp--]); break;
			case V_RET:
				sp = bp - 1;
				pc = stack[sp + 3];
				bp = stack[sp + 2];
				break;
			default:
				throw "interpret-1";
			}
			break;
		case F_LOD:
			stack[++sp] = stack[base(bp, pcode.level, stack) + pcode.addr];
			break;
		case F_STO:
			stack[base(bp, pcode.level, stack) + pcode.addr] = stack[sp--];
			break;
		case F_CAL:
			stack[sp + 1] = base(bp, pcode.level, stack);
			stack[sp + 2] = bp;
			stack[sp + 3] = pc;
			bp = sp + 1;
			pc = pcode.addr;
			break;
		case F_INT:
			sp += pcode.addr;
			break;
		case F_JMP:
			pc = pcode.addr;
			break;
		case F_JPC:
			if (stack[sp--] == 0)
				pc = pcode.addr;
			break;
		default:
			throw "interpret-2";
		}
	} while (pc != 0);
	pl0Println("***  END PL/0  ***");
}

VM.prototype.dump = function() {
	for (var i = 0; i < this.codeAlloc; i++) {
		if (this.code[i].func == F_OPR) {
			pl0Println(i + ":" + codeName[this.code[i].func] + "\t" + opName[this.code[i].addr]);
		} else {
			pl0Println(i + ":" + codeName[this.code[i].func] + "\t" + this.code[i].level + ":" + this.code[i].addr);
		}
	}
}

VM.prototype.backPatch = function(x) {
	this.code[x].addr = this.codeAlloc;
}

VM.prototype.currentCx = function() {
	return this.codeAlloc;
}

function Scope(parent, level) {
	this.parent = parent;
	this.table = {};
	this.level     = level;
	this.dataAlloc = 3;
}

Scope.prototype.add = function(name, addr, kind) {
	var p = new Identifier(name, addr, kind, this.level);
	this.table[name] = p;
	return p;
}

Scope.prototype.addVariable = function(name) {
	return this.add(name, this.dataAlloc++, K_VARIABLE);
}

Scope.prototype.addProcedure = function(name) {
	return this.add(name, 0, K_PROCEDURE);
}

Scope.prototype.push = function() {
	var newScope = new Scope(this, this.level + 1);
	return newScope;
}

Scope.prototype.pop = function() {
	return this.parent;
}

Scope.prototype.search = function(name) {
	var p = this.table[name];
	if (p)
		return p;
	else if (this.parent == null)
		return null;
	else 
		return this.parent.search(name);
}

function Parser(s, vm) {
	this.tokenizer = new Tokenizer(s);
	this.scope = new Scope(null, 0);
	this.vm = vm;
}

Parser.prototype.syntaxError = function(s) {
	alert('Syntax error:' + s + this.tokenizer.src);
	throw "syntax error:" + s + this.tokenizer.src;
}

Parser.prototype.check = function(t) {
	if (t == this.tokenizer.token) {
		this.tokenizer.getToken();
	} else {
		this.syntaxError('check:' + t);
	}
}

Parser.prototype.variableList = function() {
	if (this.tokenizer.token != R_VAR)
		return;
	do {
		this.tokenizer.getToken();	// varの読み飛ばし
		var name = this.tokenizer.value;
		this.check(IDENT);
		this.scope.addVariable(name);
	} while (this.tokenizer.token == COMMA);
	this.check(SEMICOLON);
}

Parser.prototype.procedure = function() {
	this.tokenizer.getToken();	// procedure の読み飛ばし
	var name = this.tokenizer.value;
	this.check(IDENT);
	var p = this.scope.addProcedure(name);
	this.scope = this.scope.push();
	this.check(SEMICOLON);
	this.block(p);
	this.check(SEMICOLON);
	this.scope = this.scope.pop();
}

Parser.prototype.procedureList = function() {
	while (this.tokenizer.token == R_PROCEDURE) {
		this.procedure();
	}
}

Parser.prototype.statementList = function() {
	this.statement();
	while (this.tokenizer.token == SEMICOLON) {
		this.tokenizer.getToken();
		this.statement();
	}
}

Parser.prototype.factor = function() {
	switch (this.tokenizer.token) {
	case IDENT:
		var name = this.tokenizer.value;
		this.tokenizer.getToken();
		var p = this.scope.search(name);
		this.vm.gen(F_LOD, this.scope.level - p.level, p.addr);
		break;
	case NUMBER:
		var num = parseInt(this.tokenizer.value);
		this.tokenizer.getToken();
		this.vm.gen(F_LIT, 0, num);
		break;
	case LPAREN:
		this.tokenizer.getToken();
		this.expression();
		this.check(RPAREN);
		break;
	default:
		this.syntaxError('factor');
	}
}

Parser.prototype.term = function() {
	this.factor();
	while (this.tokenizer.token == OP_MUL) {
		var op = this.tokenizer.value;
		this.tokenizer.getToken();
		this.factor();
		this.vm.gen(F_OPR, 0, optable[op])
	}
}

Parser.prototype.expression = function() {
	this.term();
	while (this.tokenizer.token == OP_ADD) {
		var op = this.tokenizer.value;
		this.tokenizer.getToken();
		this.term();
		this.vm.gen(F_OPR, 0, optable[op])
	}
}

Parser.prototype.condition = function() {
	var tokenizer = this.tokenizer;
	if (tokenizer.token == R_ODD) {
		tokenizer.getToken(); // skip "odd"
		this.expression();
		this.vm.gen(F_OPR, 0, V_ODD)
	} else {
		this.expression();
		while (this.tokenizer.token == OP_COMP) {
			var op = tokenizer.value;
			tokenizer.getToken();
			this.expression();
			this.vm.gen(F_OPR, 0, optable[op])
		}
	}
}

Parser.prototype.assignStatement = function () {
	var name = this.tokenizer.value;
	this.tokenizer.getToken();
	var p = this.scope.search(name);
	this.check(OP_BECOMES);
	this.expression();
	this.vm.gen(F_STO, this.scope.level - p.level, p.addr);
}

Parser.prototype.callProcedure = function () {
	this.tokenizer.getToken();	// skip "call"
	var name = this.tokenizer.value;
	this.check(IDENT);
	var p = this.scope.search(name);
	this.vm.gen(F_CAL, this.scope.level - p.level, p.addr);
	return;
}

Parser.prototype.ifStatement = function () {
	this.tokenizer.getToken();	// skip "if"
	this.condition();
	var x = this.vm.gen(F_JPC, 0, 0);
	this.check(R_THEN);
	this.statement();
	this.vm.backPatch(x);
}

Parser.prototype.whileStatement = function (name) {
	var vm = this.vm;
	this.tokenizer.getToken();	// skip "while"
	var x1 = vm.currentCx();
	this.condition();
	var x2 = vm.gen(F_JPC, 0, 0);
	this.check(R_DO);
	this.statement();
	vm.gen(F_JMP, 0, x1);
	vm.backPatch(x2);
}

Parser.prototype.statement = function() {
	var tokenizer = this.tokenizer;
	switch (tokenizer.token) {
	case IDENT:
		this.assignStatement();
		break;
	case R_CALL:
		this.callProcedure();
		break;
	case R_IF:
		this.ifStatement();
		break;
	case R_WHILE:
		this.whileStatement();
		break;
	case R_PRINT:
		tokenizer.getToken();	// skip "print"
		this.expression();
		this.vm.gen(F_OPR, 0, V_PRT);
		break;
	case R_BEGIN:
		tokenizer.getToken();
		this.statementList();
		this.check(R_END);
		break;
	default:
		// 空文
	}
}

Parser.prototype.block = function(p) {
	var x = this.vm.gen(F_JMP, 0, 0);		// x:手続きの前方参照用
	this.variableList();
	this.procedureList();
	this.vm.backPatch(x);
	if (p)
		p.addr = this.vm.codeAlloc;
	this.vm.gen(F_INT, 0, this.scope.dataAlloc);		// ローカル変数の確保
	this.statement();
	this.vm.gen(F_OPR, 0, V_RET);
}

Parser.prototype.parse = function() {
	this.tokenizer.getToken();
	this.block(null);
	this.check(PERIOD);
	this.vm.dump();
	this.vm.interprete();
}

function compileAndRun(s) {
	var vm = new VM();
	var parser = new Parser(s, vm);
	parser.parse();
}

function main() {
	var scripts = document.getElementsByTagName("script");
	for(var i=0; i < scripts.length; i++) {
		if(scripts[i].type == "text/pl0") {
			var s = scripts[i].text;
			compileAndRun(s);
		}
	}
}
