/*
 * bm_search.c - Boyer-Moore algorithm
 *
 *  Copyright (C) 2005 Seiji Tamura
 *
 *  ignore 0 : case sensitive
 *  ignore 1 : ignore cases
 */
#include "bm_search.h"

int bm_search(unsigned char *text, int text_len,
              unsigned char *pattern, int pat_len,
              int ignore)
{
	int skip[256];
	register int i;
	register int j;

	for (i = 0; i < 256; i++) {
		skip[i] = pat_len;
	}
	if (ignore == 0) {
		for (i = 0; i < pat_len - 1; i++) {
			skip[pattern[i]] = pat_len - i - 1;
		}
	}
	else {
		for (i = 0; i < pat_len - 1; i++) {
			skip[up_case(pattern[i])] = pat_len - i - 1;
		}
	}

	i = pat_len - 1;
	if (ignore == 0) {
		while (i < text_len) {
			j = pat_len - 1;
			while (text[i] == pattern[j]) {
				if (j == 0) {
					return(i);
				}
				i--;
				j--;
			}
			i = i + max(skip[text[i]], pat_len - j);
		}
	}
	else {
		while (i < text_len) {
			j = pat_len - 1;
			while (up_case(text[i]) == up_case(pattern[j])) {
				if (j == 0) {
					return(i);
				}
				i--;
				j--;
			}
			i = i + max(skip[up_case(text[i])], pat_len - j);
		}
	}
	return(-1);
}

int bm_re_search(unsigned char *text, int text_len,
                 unsigned char *pattern, int pat_len,
                 int ignore)
{
	int skip[256];
	register int i;
	register int j;

	for (i = 0; i < 256; i++) {
		skip[i] = pat_len;
	}
	if (ignore == 0) {
		for (i = pat_len - 1; i > 0 ; i--) {
			skip[pattern[i]] = i;
		}
	}
	else {
		for (i = pat_len - 1; i > 0 ; i--) {
			skip[up_case(pattern[i])] = i;
		}
	}

	i = text_len - pat_len;
	if (ignore == 0) {
		while (i >= 0) {
			j = 0;
			while (text[i] == pattern[j]) {
				if (j == pat_len - 1) {
					return(i - (pat_len - 1));
				}
				i++;
				j++;
			}
			i = i - max(skip[text[i]], j + 1);
		}
	}
	else {
		while (i >= 0) {
			j = 0;
			while (up_case(text[i]) == up_case(pattern[j])) {
				if (j == pat_len - 1) {
					return(i - (pat_len - 1));
				}
				i++;
				j++;
			}
			i = i - max(skip[up_case(text[i])], j + 1);
		}
	}
	return(-1);
}
