/*
 *  code.c - Code generation and linking functions
 *
 *  Copyright (C) 2003 Seiji Tamura
 */
#include <stdio.h>
#include "common.h"

extern int tokenline;
extern int labelbuf[];
extern namedlblinf_t namedlblbuf[];

int fargc;
char **fargv;

#define CODE_MAX	16384
static intrcode_t code[CODE_MAX];
static int pc = 0;	/* program counter */

//int add_code(int op, int o1, int o2, int reg, char *inf);
int print_all_code();
int print_code(int pc);

static char errbuf[256];

intrcode_t *get_code()
{
	return code;
}

int get_code_size()
{
	return pc;
}

int pop_code(int count)
{
	int i;

	for (i = 0; i < count; i++)
		pc--;
	return(count);
}

int add_code(int op, int o1, int o2, int reg, char *inf)
{
	if (pc >= CODE_MAX)
		fatal("code buffer overflow");
	code[pc].op = op;
	code[pc].o1 = o1;
	code[pc].o2 = o2;
	code[pc].reg = reg;
	code[pc].line = tokenline;
	code[pc].inf = inf;
	return(pc++);
}

int add_code_d(int op, double d1, int o2, int reg, char *inf)
{
	if (pc >= CODE_MAX)
		fatal("code buffer overflow");
	code[pc].op = op;
	code[pc].d1 = d1;
	code[pc].o2 = o2;
	code[pc].reg = reg;
	code[pc].line = tokenline;
	code[pc].inf = inf;
	return(pc++);
}

int add_code_s(int op, char *s1, int o2, int reg, char *inf)
{
	if (pc >= CODE_MAX)
		fatal("code buffer overflow");
	code[pc].op = op;
	code[pc].s1 = s1;
	code[pc].o2 = o2;
	code[pc].reg = reg;
	code[pc].line = tokenline;
	code[pc].inf = inf;
	return(pc++);
}

int chg_code(int chg_pc, int op, int o1, int o2, int reg, char *inf)
{
	code[chg_pc].op = op;
	code[chg_pc].o1 = o1;
	code[chg_pc].o2 = o2;
	code[chg_pc].reg = reg;
	code[chg_pc].line = tokenline;
	code[chg_pc].inf = inf;
	return(chg_pc);
}

int startup_code()
{
	add_code(FRAME, -1, 0, 0, "main");
	add_code_s(PUSH, (char *)(&fargc), -1, 0, "main");
	add_code(TYPE, TYPE_INT, sizeof(int), 0, 0);
	add_code(ARG, 0, 0, 0, 0);
	add_code_s(PUSH, (char *)(&fargv), -2, 0, "main");
//printf("%p : %s\n", fargv[0], fargv[0]);
//printf("%p : %s\n", fargv[1], fargv[1]);
	add_code(TYPE, TYPE_CHAR, sizeof(int), 2, 0);
	add_code(ARG, 0, 0, 0, 0);
	add_code(CALL, 0, 0, 0, "main");
	add_code(POPUP, 0, 0, 0, 0);
	add_code(EXIT, 0, 0, 0, 0);
	return(0);
}

int popup()
{
	add_code(POPUP, 0, 0, 0, 0);
	return(0);
}

int checklabel(int fent)
{
	int i;

	for (i = fent; i < pc; i++) {
		if (((code[i].op == JMP) || (code[i].op == TJMP) ||
		     (code[i].op == FJMP)) && (code[i].o2 == 0)) {
			if (code[i].o1 >= 0)
				code[i].o2 = labelbuf[code[i].o1];
			else {
				if (namedlblbuf[code[i].reg].pc < 0) {
					char *s, buf[64];
					s = namedlblbuf[code[i].reg].name;
					copystr(buf, "label not found (");
					catstr(buf, s);
					catstr(buf, ")");
					parse_error(buf);
				}
				code[i].o2 = namedlblbuf[code[i].reg].pc;
			}
		}
	}
	return(0);
}

static int setparam(int fent)
{
	int i;
	int idx;
	int mode;
	funcinf_t *finf;
	type_info *tinfo;
	int argcnt;

	i = fent;
	/* ؿ */
	finf = find_function(0, code[i].inf);
	if (finf == NULL)
		return(i);
	if (finf->argcnt == 0)
		return(i);
	argcnt = (finf->argcnt<0)?(-1*finf->argcnt):(finf->argcnt);
	/*printf("funcname %s\n", finf->name);*/
	code[i].reg = 1;
	mode = finf->scope;
	i++;
	for (idx = finf->argpos; code[i].op != CALL && i < pc; i++) {
		if (code[i].op == FRAME)
			 i = setparam(i);
		if (code[i].op != ARG)
			continue;
		if (argcnt <= 0)
			continue;
		tinfo = get_function_param(mode, idx);
		if (tinfo == NULL)
			continue;
		code[i].o1 = tinfo->type;
		code[i].o2 = tinfo->size;
		code[i].reg = tinfo->pflg;
		idx++;
		argcnt--;
	}
	return(i);
}

int checkparam(int fent)
{
	int i;

	for (i = fent; i < pc; i++) {
		if (code[i].op != FRAME)
			continue;
		if (code[i].reg != 0)
			continue;
		i = setparam(i);
	}
	return(i);
}

int checkfunc()
{
	int i;
	varinf_t *vinf;
	funcinf_t *finf;
	int fargcsize, fargvsize;

	finf = find_function(0, "main");
	if (finf == NULL)
		parse_error2("\"main\" not found", 0);
//	if (finf->argcnt == 0) {
//		fargcsize = 0;
//		fargvsize = 0;
//	}
//	else if (finf->argcnt == 1) { /* error ? */
//		fargcsize = sizeof(fargc);
//		fargvsize = 0;
//	}
//	else {
//		fargcsize = sizeof(fargc);
//		fargvsize = sizeof(fargv);
//	}
	fargcsize = sizeof(fargc);
	fargvsize = sizeof(fargv);
	finf->argsize = sizeof(fargc) + sizeof(fargv);
		
	for (i = 0; i < pc; i++) {
//		if (code[i].op == CALL && code[i].o1 == 0) {
		if (code[i].op == CALL && code[i].o1 >= 0) {
			finf = find_function(0, (char *)code[i].inf);
			if (finf == NULL) {
				char *fname = code[i].inf;
				sprintf(errbuf,"\"%s\" undeclared", fname);
				parse_error2(errbuf, code[i].line);
			}
			code[i].o1 = finf->ic;
//			/* Ȥ߹ߴؿǤʤo2ˤϰ򥻥å */
			/* ȹؿǤʤo2ˤϥե졼ॵ򥻥å */
			if (finf->ic >= 0)
//				code[i].o2 = finf->argsize;
				code[i].o2 = finf->fsize;
			else
				code[i].o2 = finf->argcnt;
//			code[i].reg = finf->fsize;
		}
		if (code[i].op == FRAME && code[i].o1 < 0) {
			finf = find_function(0, (char *)code[i].inf);
			if (finf == NULL) {
				char *fname = code[i].inf;
				sprintf(errbuf,"\"%s\" undeclared2", fname);
				parse_error2(errbuf, code[i].line);
			}
			code[i].o1 = finf->fsize;
		}
		if (code[i].op == PUSH) {
			if (code[i].o2 == -1) {
				code[i].o2 = fargcsize;
			}
			else if (code[i].o2 == -2) {
				code[i].o2 = fargvsize;
			}
		}
//		if (((code[i].op == JMP) || (code[i].op == TJMP) ||
//		     (code[i].op == FJMP)) && (code[i].o2 == 0)) {
//			code[i].o2 = labelbuf[code[i].o1];
//		}
		if (code[i].op == ADDR && code[i].o1 < 0) {
			/* ǸΤϥѿʳ */
			vinf = find_var(MODE_GLOBAL, (char *)code[i].inf);
			/*vinf = find_var(1, (char *)code[i].inf);*/
			if (vinf == NULL) {
				char *err = "symbol \"%s\" not found";
				sprintf(errbuf, err, code[i].inf);
				parse_error2(errbuf, code[i].line);
			}
			code[i].o1 = vinf->base;
			code[i].o2 = vinf->offset;
			code[i].reg = 0;
			code[i].inf = 0;
			/* ADDRμɬTYPEǤ */
			code[i + 1].o1 = vinf->type;
			code[i + 1].o2 = vinf->size;
			code[i + 1].reg = vinf->ptflag + 1;
		}
	}
	return(0);
}

int print_all_code()
{
	int i;
	for (i = 0; i < pc; i++) {
		if (code[i].op == FENT)
			printf("%s\n", (char *)code[i].s1);
		printf("%d", i);
		print_code(i);
	}
	return(0);
}

int print_code(int pc)
{
	int o1, o2, reg;
	char *inf;
	char *s1;
	int line;
	double d1;

	o1 = code[pc].o1;
	d1 = code[pc].d1;
	o2 = code[pc].o2;
	s1 = code[pc].s1;
	reg = code[pc].reg;
	inf = code[pc].inf;
	line = code[pc].line;
	switch (code[pc].op) {
	case LOAD:	printf("\tLOAD\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case ADDR:	printf("\tADDR\t%d\t%d\t%s\t%d\n", o1, o2, inf, line);	break;
	case PUSH:	printf("\tPUSH\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case POP:	printf("\tPOP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case ASGN:	printf("\tASGN\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case ADD:	printf("\tADD\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case SUB:	printf("\tSUB\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case MUL:	printf("\tMUL\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case DIV:	printf("\tDIV\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case MOD:	printf("\tMOD\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case INVT:	printf("\tINVT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CMP:	printf("\tCMP\t%d\t%d\t%d\t%d\n",  o1, o2, reg, line);	break;
	case RET:	printf("\tRET\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case JMP:	printf("\tJMP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case FJMP:	printf("\tFJMP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case TJMP:	printf("\tTJMP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CALL:  printf("\tCALL\t%d\t%d\t%s\t%d\n",o1,o2, inf, line);	break;
	case LABEL:	printf("\tLABEL\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case OR:	printf("\tOR\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case XOR:	printf("\tXOR\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case AND:	printf("\tAND\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case NCMP:	printf("\tNCMP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case LT:	printf("\tLT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case LE:	printf("\tLE\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case GT:	printf("\tGT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case GE:	printf("\tGE\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case LSFT:	printf("\tLSFT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case RSFT:	printf("\tRSFT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case FRAME:	printf("\tFRAME\t%d\t%d\t%s\t%d\n", o1, o2, inf, line);	break;
	case ARG:	printf("\tARG\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CNST:	printf("\tCONST\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case EXCHG:	printf("\tEXCHG\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case POPUP:	printf("\tPOPUP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case FENT:	printf("\tFENT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case FRET:	printf("\tFRET\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CPVAL:	printf("\tCPVAL\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case MOV:	printf("\tMOV\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CMPR:	printf("\tCMPR\t%d\t%d\t%d\t%d\n",  o1, o2, reg, line);	break;
	case ADRS:	printf("\tADRS\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case NOT:	printf("\tNOT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CMPL:	printf("\tCMPL\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case TYPE:	printf("\tTYPE\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CAST:	printf("\tCAST\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CNSTD:	printf("\tCNSTD\t%f\t%d\t%d\t%d\n", d1, o2, reg, line);	break;
	case NOP:	printf("\tNOP\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case PCALL:	printf("\tPCALL\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case PFRME:	printf("\tPFRME\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case PARG:	printf("\tPARG\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case EXIT:	printf("\tEXIT\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case ARGEX:	printf("\tARGEX\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case EXCHGF:	printf("\tEXCHGF\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	case CNSTS:	printf("\tCNSTS\t%p\t%d\t%d\t%d\n", s1, o2, reg, line);	break;
	default:	printf("\t\?\?\?\t%d\t%d\t%d\t%d\n", o1, o2, reg, line);	break;
	}
	return(0);
}
