/*
 *  common.h - ̥ǡ
 *
 *  Copyright (C) 2003 Seiji Tamura
 */

#define MINIC_VERSION	"minic version 0.9.34"

#define IDSIZE		30	/* Max length of an identifier		*/
#define STRSIZE		2000	/* String buffer size			*/
#define NEGL		-1
#define ARR_MAX		6	/* Dimention of the array		*/

#define FALSE		0
#define TRUE		1

#define INCLUDE_SYS	0
#define INCLUDE_USR	1

/* struct varinf.base */
#define BASE_GLOBAL	0
#define BASE_LOCAL	1
#define BASE_PARAM	2

/*********************
 ɽ (Symbol Table)
 *********************/
/* ѿ */
typedef struct varinf {
	char	*name;
	int	scope;
	int	type;
	union {
		int	ttype;	/* INT,LONG,CHAR,DOUBLE...		*/
		char	*tname;	/* struct or union tag name		*/
	};
	int	size;
	int	tsize;
	int	base;
	int	offset;
	int	ptflag;
	int	array[ARR_MAX];
	int	confirmed;
	int	funcptr;
	int	argflag;
	struct varinf *next;	/* struct member or function param	*/
	struct varinf *same;	/* varinf that has a same hash value	*/
	struct stuninf *stuni;	/* pointer to struct or union info	*/
}	varinf_t;

/* ؿ */
typedef struct funcinf {
	char	*name;
	int	scope;
	int	argpos;
	int	argcnt;
	int	argsize;
	int	fsize;
	int	ic;
	int	confirmed;
	struct varinf *ret;
	struct varinf *arg;
}	funcinf_t;

/* ¤  */
typedef struct stuninf {
	char	*name;
	int	scope;
	int	type;
	int	size;
	struct varinf *var;
}	stuninf_t;

/* ޥ */
typedef struct macroinf {
	char	name[IDSIZE+1];
	int	type;
	union {
		int	intnum;
		int	ch;
		char	*string;
		double	decnum;
	};
}	macroinf_t;

/* 󥯥롼 */
typedef struct includeinf {
	char	name[IDSIZE+1];
	int	type;		/* INCLUDE_SYS or INCLUDE_USR		*/
}	includeinf_t;

/* typedef */
typedef struct typedefinf {
	char	name[IDSIZE+1];
	int	type;
	union {
		int	ttype;	/* INT,LONG,CHAR,DOUBLE...		*/
		char	*tname;	/* struct or union tag name		*/
	};
	int	ptflag;
}	typedefinf_t;

/* ̾դ٥ */
typedef struct namedlblinf {
	char	name[IDSIZE+1];
	int	pc;
}	namedlblinf_t;

/* switch-caseѥ٥ */
typedef struct switchlblinf {
	int	brk;	/* ̥   */
	int	pc;	/*  */
	int	val;	/* caseʸ */
	int	def;	/* default ?  */
}	switchlblinf_t;

/************
 ȡ
 ************/
typedef struct token {
	int	type;
	union {
		int	ttype;	/* INT,LONG,CHAR,DOUBLE...	*/
		int	intnum;
		int	ch;
		char	*string;
		double	decnum;
		char	ident[IDSIZE+1];
		macroinf_t	*macro;
		includeinf_t	*incinf;
	};
	int	ptflag;
}	token_t;

/**********
 ֥
 **********/
typedef struct intrcode {
	int op;
	union {
		int o1;
		double d1;
		char *s1;
	};
	int o2;
	int reg;
	int line;
	char *inf;
}	intrcode_t;

/**************
 ֥̿
 **************/
/* operation codes */
enum
{
	LOAD,	ADDR,	PUSH,	POP,	ASGN,	ADD,
	SUB,	MUL,	DIV,	MOD,	CMP,  	RET,
	JMP,	FJMP,	TJMP,	CALL,	LABEL,	OR,
	XOR,	AND,	NCMP,	LT,	LE,	GT,
	GE,	LSFT,	RSFT,	FRAME,	ARG,	CNST,
	EXCHG,	POPUP,	FENT,	FRET,	CPVAL,	MOV,
	ADRS,	INVT,	NOT,	CMPL,	TYPE,	CMPR,
	CAST,	CNSTD,	NOP,	PCALL,	PFRME,	PARG,
	EXIT,	ARGEX,	EXCHGF,	CNSTS
};

#if 0
op   | o1,d1,s1        | o2         | reg        | inf
-----+-----------------+------------+---------------------------
FENT   function name     0            0            0
ADDR   base address      offset       0            var name or 0

#endif

/**************************
 פʤɤ˴ؤ⡼
 **************************/
#define MODE_LOCAL	0x0001
#define MODE_GLOBAL	0x0002
#define MODE_STATIC	0x0004
#define MODE_PARAM	0x0008
#define MODE_EXTERN	0x0010
#define MODE_TYPEDEF	0x0020
#define MODE_DECLARE	0x0040

/* token type */

#define IDENT		257
#define LITERAL		258
#define COMMA		259
#define SEMICOLON	260
#define LBRACE		261
#define RBRACE		262
#define LPAR		263
#define RPAR		264
#define LBRACK		265
#define RBRACK		266
#define TYPENAME	267
#define BREAK		268
#define ELSE		269
#define EXTERN		270
#define IF		271
#define RETURN		272
#define DO		273
#define WHILE		274
#define EQUAL		275
#define OROR		276
#define ANDAND		277
#define OR		278
#define UPARROW		279
#define AND		280
#define EQEQ		281
#define NOTEQ		282
#define LE		283
#define LT		284
#define GE		285
#define GT		286
#define LSHIFT		287
#define RSHIFT		288
#define PLUS		289
#define MINUS		290
#define STAR		291
#define SLASH		292
#define PERCENT		293
#define PLUSPLUS	294
#define MINUSMINUS	295
#define EXCLAM		296
#define TILDE		297
#define FOR		298
#define GOTO		299
#define CASE		300
#define DEFAULT		301
#define SWITCH		302
#define COLON		303
#define CONTINUE	304
#define QUESTION	305
#define NUMBER		306
#define CHARACTER	307
#define STRING		308
#define STAREQ		309
#define ANDEQ		310
#define OREQ		311
#define PLUSEQ		312
#define MINUSEQ		313
#define SLASHEQ		314
#define PERCENTEQ	315
#define LSHIFTEQ	316
#define RSHIFTEQ	317
#define UPARROWEQ	318
#define DECIMAL		319
#define STATIC		320
#define STRUCT		321
#define UNION		322
#define ARROW		323
#define DOT		324
#define SIZEOF		325
#define DEFINE		326
#define INCLUDE		327
#define TYPEDEF		328
#define IFDEF		329
#define IFNDEF		330
#define ENDIF		331
#define TRIPLEDOT	332
#define TYPECAST	333
#define TYPE_CHAR	334
#define TYPE_DOUBLE	335
#define TYPE_FLOAT	336
#define TYPE_INT	337
#define TYPE_LONG	338
#define TYPE_SHORT	339
#define TYPE_VOID	340
#define UNSIGNED	341
#define TYPE_UCHAR	342
#define TYPE_UINT	343
#define TYPE_ULONG	344
#define TYPE_USHORT	345

#define TYPE_POINTER	346

/* register index */

#define AX		0
#define BX		1
#define CX		2
#define DX		3
#define SI		4
#define DI		5
#define BP		6
#define SP		7
#define IP		8
#define CS		9
#define DS		10
#define ES		11
#define SS		12
#define REGSIZE		13

/* code buffer */

typedef struct {
	char	*stdname;
	int	scode;
	int	argcnt;
	varinf_t	*args;
}	intrfunc;

/* type infomation */

typedef struct {
	int type;
	int size;
	int pflg;
}	type_info;

/* external functions */

/* main.c */
extern int charget();
extern int include_main(includeinf_t *);

/* misc.c */
extern int compstr(char *, char *);
extern int copystr(char *, char *);
extern char *catstr(char *, char *);
extern char *chrstr(char *, int);
extern void *copymem(void *, void *, int);
extern int lenstr(char *);

/* table.c */
extern void init_table(int);
extern int init_label();
extern int label_next();
extern int regist_label(int, int);
extern int regist_sw_label(int, int, int, int);
extern switchlblinf_t *pop_sw_label(int);
extern int init_named_label();
extern int regist_named_label(char *, int);
extern int regist_macro(macroinf_t *);
extern macroinf_t *search_macro(char *);
extern int regist_typedef(typedefinf_t *);
extern typedefinf_t *search_typedef(char *);
extern char *regist_name(int, char *);
extern int correct_var_offset(int mode, varinf_t *pvinf);
extern varinf_t *regist_var(int, varinf_t *);
extern varinf_t *find_var(int, char *);
extern stuninf_t *regist_struct(int, token_t *);
extern int end_regist_struct(int, stuninf_t *);
extern stuninf_t *find_struct(int, char *);
extern varinf_t *find_struct_member(varinf_t *, char *);
extern funcinf_t *regist_function(int, varinf_t *, int);
extern int end_function_param(int, funcinf_t *);
extern int end_regist_function(int, funcinf_t *, int);
extern int end_prototype();
extern funcinf_t *find_function(int, char *);
extern type_info *get_function_param(int, int);
extern int inc_local_block();
extern int dec_local_block();
extern int init_file_scope_table();

/* hash.c */
extern int hash_value(char *, int);
extern void init_hash(int *, int, int);

/* code.c */
extern int get_code_size();
extern intrcode_t *get_code();
extern int pop_code(int count);
extern int add_code(int, int, int, int, char *);
extern int add_code_d(int, double, int, int, char *);
extern int add_code_s(int, char *, int, int, char *);
extern int chg_code(int, int, int, int, int, char *);
extern int startup_code();
extern int popup();
extern int checklabel(int);
extern int checkparam(int);
extern int checkfunc();
extern int print_all_code();
extern int print_code(int);

/* error.c */
extern void yyerror(char *);
extern void yyerror2(char *, int);
extern void fatal(char *);
extern void parse_error(char *);
extern void parse_error2(char *, int);
extern void runerror(char *);
extern void runerror2(char *, int);

/* expr.c */
extern int expression(int);
extern int expr_sizeof();
extern int expr_typecast();

/* lex.c */
extern void initchar();
extern void add_token_que(token_t *);
extern token_t *get_token_que();

/* parse.c */
extern void parse(int, int);
extern void parse_include();
extern void chktoken(int, char *);

/* intprt.c */
extern int interpret(intrcode_t *, int, int);
//extern int push(char *addr, int size);
//extern int pop(char *addr, int size);
extern int push_v(char *addr, int size);
extern int pop_v(char *addr, int size);
extern int push_t(type_info *tinfo);
extern int pop_t(type_info *tinfo);
extern int cast2int(type_info *ptinfo, char *buf);

/* funcs.c */
extern void if_fprintf(intrcode_t *code);
extern void if_printf(intrcode_t *code);
extern void if_malloc(intrcode_t *code);
extern void if_free(intrcode_t *code);
extern void if_memcpy(intrcode_t *code);
extern void if_memset(intrcode_t *code);
extern void if_strcpy(intrcode_t *code);
extern void if_sprintf(intrcode_t *code);
extern void if_fopen(intrcode_t *code);
extern void if_fclose(intrcode_t *code);
extern void if_fgetc(intrcode_t *code);
extern void if_fgets(intrcode_t *code);
extern void if_fputc(intrcode_t *code);
extern void if_fputs(intrcode_t *code);
extern void if_getc(intrcode_t *code);
extern void if_getchar(intrcode_t *code);
extern void if_putc(intrcode_t *code);
extern void if_putchar(intrcode_t *code);
extern void if_strlen(intrcode_t *code);
extern void if_dynfunc(intrcode_t *code);
extern void if_rx_search(intrcode_t *code);
extern void if_amatch(intrcode_t *code);
extern void if_bf_search(intrcode_t *code);
extern void if_bf_re_search(intrcode_t *code);
extern void if_bm_search(intrcode_t *code);
extern void if_bm_re_search(intrcode_t *code);
extern void if_atoi(intrcode_t *code);
extern void if_realloc(intrcode_t *code);
extern void if_qsort(intrcode_t *code);
extern void if_abort(intrcode_t *code);
extern void if_exit(intrcode_t *code);
extern void if_fscanf(intrcode_t *code);
extern void if_scanf(intrcode_t *code);
extern void if_sscanf(intrcode_t *code);
extern void if_strcmp(intrcode_t *code);
extern void if_strncmp(intrcode_t *code);
extern void if_clock(intrcode_t *code);
extern void if_time(intrcode_t *code);
extern void if_difftime(intrcode_t *code);
extern void if_mktime(intrcode_t *code);
extern void if_asctime(intrcode_t *code);
extern void if_ctime(intrcode_t *code);
extern void if_gmtime(intrcode_t *code);
extern void if_localtime(intrcode_t *code);
extern void if_strftime(intrcode_t *code);

/* eval.c */
extern int eval_expr_int();

/* macro (push, pup, push_t, pop_t) */
#if 1
#define push(addr, size) push_v((char *)(addr), (size))
#define pop(addr, size)  pop_v((char *)(addr), (size))
#else
#if 1
#define push(addr,size)                                         \
	(((size)==sizeof(int))?                                 \
		(*((int *)&stack[sp])=*((int *)(addr))):        \
		(((size)==sizeof(char))?                        \
			(stack[sp]=*((char *)(addr)),0):        \
			(memcpy(&stack[sp],(addr),(size)),0)),  \
	sp+=(size),                                             \
	sp)
#define pop(addr,size)                                          \
	(sp-=(size),                                            \
	((size)==sizeof(int))?                                  \
		(*((int *)(addr))=*((int *)&stack[sp])):        \
		(((size)==sizeof(char))?                        \
			(*((char *)(addr))=stack[sp],0):        \
			(memcpy((addr),&stack[sp],(size)),0)),  \
	sp)
#define push_t(tinfo)                                           \
	(*((type_info *)&stack[sp])=*(tinfo),                   \
	sp+=sizeof(type_info),                                  \
	sp)
#define pop_t(tinfo)                                            \
	(sp-=sizeof(type_info),                                 \
	*(tinfo)=*((type_info *)&stack[sp]),                    \
	sp)
#else
#define push(addr, size)                                        \
{                                                               \
	if ((size) == sizeof(int))                              \
		*((int *)&stack[sp]) = *((int *)(addr));        \
	else if ((size) == sizeof(char))                        \
		stack[sp] = *((char *)(addr));                  \
	else                                                    \
		memcpy(&stack[sp], (addr), (size));             \
	sp += (size);                                           \
}
#define pop(addr, size)                                         \
{                                                               \
	sp -= (size);                                           \
	if ((size) == sizeof(int))                              \
		*((int *)(addr)) = *((int *)&stack[sp]);        \
	else if ((size) == sizeof(char))                        \
		*((char *)(addr)) = stack[sp];                  \
	else                                                    \
		memcpy((addr), &stack[sp], (size));             \
}
#define push_t(tinfo)                                           \
{                                                               \
	*((type_info *)&stack[sp]) = *(tinfo);                  \
	sp += sizeof(type_info);                                \
}
#define pop_t(tinfo)                                            \
{                                                               \
	sp -= sizeof(type_info);                                \
	*(tinfo) = *((type_info *)&stack[sp]);                  \
}
#endif
#endif
