/*
 *  eval.c - Evaluation of constant type
 *
 *  Copyright (C) 2005 Seiji Tamura
 */
#include <stdio.h>
#include "common.h"

/* memo */
/* 礭caseʸͤʤɤ򥳥ѥ˷׻ */
/* 顼åؤɤʤΤǴ */

#define FALSE	0
#define TRUE	1

#define GetToken() tok=gettoken()

extern token_t *tok;
extern token_t *gettoken();

#define END_SIGN	-1

#define OP_EQUAL	0	/* = */
#define OP_PLUS		1	/* + */
#define OP_MINUS	2	/* - */
#define OP_STAR		3	/* * */
#define OP_SLASH	4	/* / */
#define OP_PLUS_S	5	/* single+ */
#define OP_MINUS_S	6	/* single- */
#define OP_LPAR		7	/* ( */
#define OP_RPAR		8	/* ) */
#define OP_END		9	/* $ */
#define OP_VALUE	10	/* ͤޤϥޥ */
#define OP_INVALID	11	/* ʱ黻 */

/* local variables */
static int opr_stack[64];
static int out_stack[64];
static int opr_sp;
static int out_sp;
static int last_operator;

typedef struct order {
	int opr;
	int l;
	int r;
}	order_t;

static order_t order_tbl[] = {
	{OP_EQUAL,	1,	2 },
	{OP_PLUS,	4,	3 },
	{OP_MINUS,	4,	3 },
	{OP_STAR,	6,	5 },
	{OP_SLASH,	6,	5 },
	{OP_PLUS_S,	7,	8 },
	{OP_MINUS_S,	7,	8 },
	{OP_LPAR,	0,	10},
	{OP_RPAR,	9,	0 },
	{OP_END,	0,	0 },
};
static int order_tbl_size = sizeof(order_tbl)/sizeof(order_tbl[0]);

/* local functions */
static void push_opr(int);
static int pop_opr();
static int top_opr();
static void push_out(int);
static int pop_out();
static int top_out();
static int parse_expr();
static int chk_eval_type(int);
static int chk_out_val(token_t *);
static void do_calc(token_t *);
static int get_operator(int);
static int chk_monomial();
static int l_order(int);
static int r_order(int);
static order_t *get_order_struct(int);
static int monomial(int, int);
static int binomial(int, int, int);
static int push_operator(token_t *);

/* ɾ׻̤֤ */
int eval_expr_int()
{
	int ret_val;

	opr_sp = 0;
	out_sp = 0;
	last_operator = OP_END;

	/* λɽ黻Ҥ򥹥åѤ */
	push_opr(OP_END);

	/* ϥåˣѤ */
	push_out(0);

	/* βϽ򥳡 */
	ret_val = parse_expr();

	return(ret_val);
}

static void push_opr(int opr)
{
	last_operator = opr;
	opr_stack[opr_sp++] = opr;
}

static int pop_opr()
{
	return(opr_stack[--opr_sp]);
}

static int top_opr()
{
	return(opr_stack[opr_sp - 1]);
}

static void push_out(int out)
{
	last_operator = OP_VALUE;
	out_stack[out_sp++] = out;
}

static int pop_out()
{
	return(out_stack[--out_sp]);
}

static int top_out()
{
	return(out_stack[out_sp - 1]);
}

static int parse_expr()
{
	token_t last_tok;

	while (chk_eval_type(tok->type)) {
		/* ޥϽϥåѤ */
		if (chk_out_val(tok)) {
			continue;
		}
		/* ̤ͥ㤤黻Ҥ줿ν */
		do_calc(tok);
		/* 黻Ҥ򥹥åѤ */
		if (push_operator(tok)) {
			continue;
		}
		/* '(' */
		if (top_opr() == OP_LPAR) {
			pop_opr();
			GetToken();
			continue;
		}
		/* 黻Ҥ̵ʤäǽλ */
		if (top_opr() == OP_END) {
			break;
		}
	}
	/* ٤Ʒ׻ */
	last_tok.type = END_SIGN;
	do_calc(&last_tok);
	return(top_out());
}

static int chk_eval_type(int type)
{
	switch (type) {
	case EQUAL:	/* = */
	case PLUS:	/* + */
	case MINUS:	/* - */
	case STAR:	/* * */
	case SLASH:	/* / */
	case LPAR:	/* ( */
	case RPAR:	/* ) */
	case IDENT:	/* ̻(ޥ) */
	case NUMBER:	/*  */
	case CHARACTER:	/* ʸ */
	case SIZEOF:	/* sizeof */
		return(TRUE);
	default:
		return(FALSE);
	}
}

static int chk_out_val(token_t *t)
{
	int val = 0;

	if (t->type == NUMBER) {
		val = (int)tok->intnum;
		GetToken();
	}
	else if (t->type == CHARACTER) {
		val = (int)tok->ch;
		GetToken();
	}
	else if (t->type == SIZEOF) {
		GetToken();
		val = expr_sizeof();
	}
	else {
		return(FALSE);
	}
	push_out(val);
	return(TRUE);
}

static void do_calc(token_t *t)
{
	int cur_opr;
	int do_opr;
	int l_val, r_val;

	cur_opr = get_operator(t->type);
	if (cur_opr == OP_INVALID)
		parse_error("invalid operator");
	while (l_order(top_opr()) > r_order(cur_opr)) {
		do_opr = pop_opr();
		if (do_opr == OP_PLUS_S || do_opr == OP_MINUS_S) {
			r_val = pop_out();
			push_out(monomial(r_val, do_opr));
		}
		else {
			r_val = pop_out();
			l_val = pop_out();
			push_out(binomial(l_val, r_val, do_opr));
		}
	}
}

static int get_operator(int type)
{
	int ret_type = OP_END;

	switch (type) {
	case EQUAL:	/* = */
		ret_type = OP_EQUAL;
		break;
	case PLUS:	/* + */
		if (chk_monomial())
			ret_type = OP_PLUS_S;
		else
			ret_type = OP_PLUS;
		break;
	case MINUS:	/* - */
		if (chk_monomial())
			ret_type = OP_MINUS_S;
		else
			ret_type = OP_MINUS;
		break;
	case STAR:	/* * */
		ret_type = OP_STAR;
		break;
	case SLASH:	/* / */
		ret_type = OP_SLASH;
		break;
	case LPAR:	/* ( */
		ret_type = OP_LPAR;
		break;
	case RPAR:	/* ) */
		ret_type = OP_RPAR;
		break;
	case END_SIGN:	/* END */
		ret_type = OP_END;
		break;
	case IDENT:	/* ̻(ޥ) */
	case NUMBER:	/*  */
	default:
		ret_type = OP_INVALID;
		break;
	}
	return(ret_type);
}

static int chk_monomial()
{
	if (last_operator == OP_VALUE || last_operator == OP_RPAR)
		return(FALSE);
	else
		return(TRUE);
}

static int l_order(int opr)
{
	order_t *ret;

	ret = get_order_struct(opr);
	if (ret) {
		return(ret->l);
	}
	return(0);
}

static int r_order(int opr)
{
	order_t *ret;

	ret = get_order_struct(opr);
	if (ret) {
		return(ret->r);
	}
	return(0);
}

static order_t *get_order_struct(int opr)
{
	int i;
	order_t *ret = NULL;

	for (i = 0; i < order_tbl_size; i++) {
		if (order_tbl[i].opr == opr) {
			ret = &order_tbl[i];
			break;
		}
	}
	return(ret);
}

static int monomial(int val, int opr)
{
	int ret_val;
	if (opr == OP_MINUS_S)
		ret_val = -val;
	else	/* (opr == OP_PLUS_S) */
		ret_val = val;
	return(ret_val);
}

static int binomial(int l_val, int r_val, int opr)
{
	int ret_val = 0;

	switch (opr) {
	case OP_EQUAL:
		ret_val = l_val = r_val;
		break;
	case OP_PLUS:
		ret_val = l_val + r_val;
		break;
	case OP_MINUS:
		ret_val = l_val - r_val;
		break;
	case OP_STAR:
		ret_val = l_val * r_val;
		break;
	case OP_SLASH:
		if (r_val == 0)
			parse_error("division by zero");
		ret_val = l_val / r_val;
		break;
	default:
		parse_error("invalid operator");
		break;
	}
	return(ret_val);
}

static int push_operator(token_t *t)
{
	int cur_opr;

	cur_opr = get_operator(t->type);
	if (cur_opr == OP_INVALID)
		parse_error("invalid operator");
	if (l_order(top_opr()) < r_order(cur_opr)) {
		push_opr(cur_opr);
		GetToken();
		return(TRUE);
	}
	return(FALSE);
}
