/*
 *  intprt.c - minic virtual machine
 *
 *  Copyright (C) 2003 Seiji Tamura
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"

/* memo */
/* ۥޥ */

//typedef struct {
//	int type;
//	int size;
//	int pflg;
//}	type_info;

#define IS_STRUCT(a)	((a) == STRUCT || (a) == UNION)

typedef unsigned int uint;
typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;

/* local variables */
#define STACKSIZE	100000
static char vstack[STACKSIZE];
//static char *stack = vstack;
char *stack = vstack;
#define GLOBALSIZE	100000
static char vgarea[GLOBALSIZE];
//static char *garea = vgarea;
char *garea = vgarea;

//static int sp = 0;	/* stack pointer */
int sp = 0;	/* stack pointer */
//static int fp = 0;	/* frame pointer */
int fp = 0;	/* frame pointer */
//static int ic = 0;	/* instruction counter */
int ic = 0;	/* instruction counter */
//static int chksp = 0;
int chksp = 0;
//static int argsize = 0;
int argsize = 0;
static int curargsize = 0;

/* local functions */
//static int push(void *addr, int size);
//static int pop(void *addr, int size);
//static int push_t(type_info *tinfo);
//static int pop_t(type_info *tinfo);
static int assign(intrcode_t *code);
static int get_type_size(int type);
static int chkpointer(intrcode_t *, type_info *, char *, type_info *, char *);
static int getoptype(int op, type_info *ptinfo1, type_info *ptinfo2);
//static int cast2int(type_info *ptinfo, char *buf);
static uint cast2uint(type_info *ptinfo, char *buf);
static double cast2dbl(type_info *ptinfo, char *buf);
static char cast2chr(type_info *ptinfo, char *buf);
static char cast2uchr(type_info *ptinfo, char *buf);
static char *cast2ptr(type_info *ptinfo, char *buf);
static int binomial_int(int, type_info *, char *, type_info *, char *);
static int binomial_dbl(int, type_info *, char *, type_info *, char *);
static int binomial_chr(int, type_info *, char *, type_info *, char *);
static int binomial_ptr(int, type_info *, char *, type_info *, char *);
static int binomial(intrcode_t *code);
static int monomial(int op);
static int typecast(intrcode_t *code);
static void intrinsic_function(intrcode_t *code);

#ifdef DEBUG
static int debugASGN(type_info *tinfo, int *ptmp);
static int debugCAST(type_info *tinfo, int *ptmp);
#endif
#if 1
//static int push(void *addr, int size)
int push_v(char *addr, int size)
{
	if (size == sizeof(int))
		*((int *)&stack[sp]) = *((int *)addr);
	else if (size == sizeof(char *))
		*((char **)&stack[sp]) = *((char **)addr);
	else if (size == sizeof(char))
		stack[sp] = *((char *)addr);
	else
		memcpy(&stack[sp], addr, size);
	sp += size;
#ifdef DEBUG
//	printf("push sp = %d\tsize = %d\n", sp, size);
#endif
	return(sp);
}

//static int pop(void *addr, int size)
int pop_v(char *addr, int size)
{
	sp -= size;
	if (size == sizeof(int))
		*((int *)addr) = *((int *)&stack[sp]);
	else if (size == sizeof(char *))
		*((char **)addr) = *((char **)&stack[sp]);
	else if (size == sizeof(char))
		*((char *)addr) = stack[sp];
	else
		memcpy(addr, &stack[sp], size);
#ifdef DEBUG
//	printf("pop  sp = %d\tsize = %d\n", sp, size);
#endif
	return(sp);
}

//static int push_t(type_info *tinfo)
int push_t(type_info *tinfo)
{
	*((type_info *)&stack[sp]) = *tinfo;
	sp += sizeof(type_info);
	return(sp);
}

//static int pop_t(type_info *tinfo)
int pop_t(type_info *tinfo)
{
	sp -= sizeof(type_info);
	*tinfo = *((type_info *)&stack[sp]);
	return(sp);
}
#endif
int interpret(intrcode_t *code, int size, int trace)
{
	char *addr;
	type_info tinfo1;
	type_info tinfo2;
	char **ppc_buf1;
	int *pi_buf1;
	int *pi_buf2;
	char buf1[256];
	char buf2[256];
	int size1;
	int size2;
//	int tmpsp = 0;
//	int fsize = 0;
//	int argsize = 0;
//	int curargsize = 0;
	int value;
	static int reg[REGSIZE];
	int tmpargsize = 0;
	int tmpcurargsize = 0;

//	ic = 0;
//	fp = 0;
//	sp = 0;

	while (1) {
		if (ic >= size) {
#ifdef DEBUG
//			printf("(ic, fp, sp) = (%d, %d, %d)\n", ic, fp, sp);
			printf("(ic, size) = (%d, %d)\n", ic, size);
#endif
			runerror("error: instruction counter overflow");
		}
		if (trace) {
			printf("%d", ic);
			print_code(ic);
		}
		switch (code[ic].op) {
		case LOAD:
			pop_t(&tinfo1);
			if (tinfo1.pflg <= 0)
				runerror2("System error LOAD", code[ic].line);
			pop(buf1, sizeof(void *));
			tinfo1.pflg--;
//			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			if (tinfo1.pflg > 0)
				size1 = sizeof(void *);
			else {
				size1 = get_type_size(tinfo1.type);
				if (size1 < 0)
					size1 = tinfo1.size;
			}
//printf("LOAD buf %p\n", *((char **)buf1));
			ppc_buf1 = (char **)buf1;
			push(*ppc_buf1, size1);
//			push(*((char **)buf1), size1);
			if (!(tinfo1.type == STRUCT || tinfo1.type == UNION))
				tinfo1.size = size1;
			push_t(&tinfo1);
			break;
		case ADDR:
			switch (code[ic].o1) {
			case BASE_LOCAL:
				addr = stack + fp + code[ic].o2;
				break;
			case BASE_PARAM:
				addr = stack + fp - curargsize + code[ic].o2;
				addr -= sizeof(ic)+sizeof(fp)+sizeof(argsize);
				break;
			case BASE_GLOBAL:
				addr = garea + code[ic].o2;
				break;
			default:
				addr = 0;
				runerror("error: unknown base type");
				break;
			}
			push(&addr, sizeof(addr));
			break;
		case PUSH:
			/* ̿᤬ľܸƤӽФ뤳Ȥ¿ʬʤ */
			push((void *)code[ic].s1, code[ic].o2);
			break;
		case POP:
			/* ̿᤬ľܸƤӽФ뤳Ȥ¿ʬʤ */
			pop((void *)code[ic].s1, code[ic].o2);
			break;
		case ASGN:
			assign(&code[ic]);
			break;
		case ADD:
		case SUB:
		case MUL:
		case DIV:
		case MOD:
		case OR:
		case XOR:
		case AND:
		case CMP:
		case NCMP:
		case LT:
		case LE:
		case GT:
		case GE:
		case LSFT:
		case RSFT:
			binomial(&code[ic]);
			break;
		case RET:
			/* åΥȥåפͤ */
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
//{
//int ii;
//for (ii = 0; ii < size1; ii++) {
//  printf("%02x ", (int)((unsigned char)buf1[ii]));
//}
//printf("\n");
//}
			/* spfpic򸵤᤹ */
			sp = fp;
			pop(&ic, sizeof(ic));
//			pop(&fsize, sizeof(fsize));
			pop(&fp, sizeof(fp));
			pop(&argsize, sizeof(argsize));
			sp -= argsize;
			pop(&argsize, sizeof(argsize));
			pop(&curargsize, sizeof(curargsize)); // 20090421
			//curargsize = argsize; // 20090421
			/* ͤ򥹥åѤ */
			push(buf1, size1);
			push_t(&tinfo1);
			break;
		case JMP:
			ic = code[ic].o2;
			break;
		case FJMP:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			if (size1 == sizeof(int)) {
			pi_buf1 = (int *)buf1;
			value = *pi_buf1;
//				value = *((int *)buf1);
			}
			else
				value = cast2int(&tinfo1, buf1);
//			if (*((int *)buf1) == 0) {
			if (value == 0) {
				if (code[ic].reg == TRUE) {
					pi_buf1 = (int *)buf1;
					*pi_buf1 = FALSE;
//					*((int *)buf1) = FALSE;
					tinfo1.type = TYPE_INT;
					tinfo1.size = sizeof(int);
					tinfo1.pflg = 0;
					push(buf1, size1);
					push_t(&tinfo1);
				}
				ic = code[ic].o2;
			}
			break;
		case TJMP:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			if (size1 == sizeof(int)) {
				pi_buf1 = (int *)buf1;
				value = *pi_buf1;
//				value = *((int *)buf1);
			}
			else
				value = cast2int(&tinfo1, buf1);
//			if (*((int *)buf1)) {
			if (value) {
				if (code[ic].reg == TRUE) {
					pi_buf1 = (int *)buf1;
					*pi_buf1 = TRUE;
//					*((int *)buf1) = TRUE;
					tinfo1.type = TYPE_INT;
					tinfo1.size = sizeof(int);
					tinfo1.pflg = 0;
					push(buf1, size1);
					push_t(&tinfo1);
				}
				ic = code[ic].o2;
			}
			break;
		case CALL:
			if (code[ic].o1 >= 0) {
				/* ϸƤӽФ褦ѹ */
				/* ΤǡߤΥåݥ󥿤Ƥ */
				/* Ф줿ؿΥե졼ݥ󥿤Ȥʤ   */
//				fp = sp - argsize;
//				fp = sp;
//				if (argsize != code[ic].o2)
//					printf("invalid argument\n");
//				tmpsp = fp - sizeof(ic);
//				tmpsp = fp - argsize - sizeof(ic);
//				memcpy(&stack[tmpsp], &ic, sizeof(ic));
//				tmpsp -= sizeof(fsize);
//				memcpy(&fsize, &stack[tmpsp], sizeof(fsize));
//				fsize = code[ic].o2;
				curargsize = argsize;
				push(&argsize, sizeof(argsize));
				push(&fp, sizeof(fp));
//				push(&fsize, sizeof(fsize));
				push(&ic, sizeof(ic));
				fp = sp;
				sp += code[ic].o2;
				ic = code[ic].o1;
			}
			else {
				intrinsic_function(&code[ic]);
				/* åΥȥåפͤ */
				pop_t(&tinfo1);
				size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
				pop(buf1, size1);
//				/* spfpic򸵤᤹ */
//				sp = fp;
//				pop(&ic, sizeof(ic));
//				pop(&fsize, sizeof(fsize));
//				pop(&fp, sizeof(fp));
//				pop(&argsize, sizeof(argsize));
//				sp -= argsize;
				pop(&argsize, sizeof(argsize));
				pop(&curargsize, sizeof(curargsize)); // 20090421
				//curargsize = argsize; // 20090421
				/* ͤ򥹥åѤ */
				push(buf1, size1);
				push_t(&tinfo1);
			}
			break;
		case PCALL:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			pop_t(&tinfo2);
			size2 = (tinfo2.pflg>0)?(sizeof(void *)):(tinfo2.size);
			pop(buf2, size2);
			pi_buf1 = (int *)buf1;
			pi_buf2 = (int *)buf2;
			if (*pi_buf1 >= 0) {
//			if (*((int *)buf1) >= 0) {
				/* ϸƤӽФ褦ѹ */
				/* ΤǡߤΥåݥ󥿤Ƥ */
				/* Ф줿ؿΥե졼ݥ󥿤Ȥʤ   */
				curargsize = argsize;
				push(&argsize, sizeof(argsize));
				push(&fp, sizeof(fp));
				push(&ic, sizeof(ic));
				fp = sp;
				/* 16ӥåȤic */
				/* 15ӥåȤե졼ॵ */
				sp += (((*pi_buf1) & 0x7fff0000) >> 16) + curargsize;
				ic = (*pi_buf1) & 0x0000ffff;
//				sp += (((*((int *)buf1)) & 0x7fff0000) >> 16) + curargsize;
//				ic = (*((int *)buf1)) & 0x0000ffff;
//printf("%d \n", (((*((int *)buf1)) & 0x7fff0000) >> 16));
//printf("%d \n", curargsize);
			}
			else {
				code[ic].o1 = *pi_buf1;
				code[ic].o2 = *pi_buf2;
//				code[ic].o1 = *((int *)buf1);
//				code[ic].o2 = *((int *)buf2);
				intrinsic_function(&code[ic]);
				/* åΥȥåפͤ */
				pop_t(&tinfo1);
				size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
				pop(buf1, size1);
				pop(&argsize, sizeof(argsize));
				pop(&curargsize, sizeof(curargsize)); // 20090421
				//curargsize = argsize; // 20090421
				/* ͤ򥹥åѤ */
				push(buf1, size1);
				push_t(&tinfo1);
			}
			break;
		case LABEL:
			/* do nothing */
			break;
		case FRAME:
//			fsize = code[ic].o1;
			push(&curargsize, sizeof(curargsize)); // 20090421
			push(&argsize, sizeof(argsize));
//			push(&fp, sizeof(fp));
//			push(&fsize, sizeof(fsize));
//			push(&ic, sizeof(ic));
			argsize = 0;
			break;
		case ARG:
			if (code[ic].o1)
				typecast(&code[ic]);
			pop_t(&tinfo1);
			argsize += (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			break;
		case ARGEX:
			if (code[ic].o1)
				typecast(&code[ic]);
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			pop_t(&tinfo2);
			size2 = (tinfo2.pflg>0)?(sizeof(void *)):(tinfo2.size);
			pop(buf2, size2);
			push(buf1, size1);
			push_t(&tinfo1);
			pop_t(&tinfo1);
			argsize += (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			push(buf2, size2);
			push_t(&tinfo2);
			break;
		case PFRME:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			push(&curargsize, sizeof(curargsize)); // 20090421
			push(&argsize, sizeof(argsize));
			push(buf1, size1);
			push_t(&tinfo1);
			argsize = 0;
			break;
		case PARG:
			if (code[ic].o1)
				typecast(&code[ic]);
			pop_t(&tinfo1);
			argsize += (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			pop_t(&tinfo2);
			size2 = (tinfo2.pflg>0)?(sizeof(void *)):(tinfo2.size);
			pop(buf2, size2);
			push(buf1, size1);
			push(buf2, size2);
			push_t(&tinfo2);
			break;
		case CNST:
			push(&code[ic].o1, code[ic].o2);
			break;
		case CNSTD:
			push(&code[ic].d1, code[ic].o2);
			break;
		case CNSTS:
			push(&code[ic].s1, code[ic].o2);
			break;
		case EXCHG:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			pop_t(&tinfo2);
			size2 = (tinfo2.pflg>0)?(sizeof(void *)):(tinfo2.size);
			pop(buf2, size2);
			push(buf1, size1);
			push_t(&tinfo1);
			push(buf2, size2);
			push_t(&tinfo2);
			break;
		case EXCHGF:
			pop(&tmpargsize, sizeof(tmpargsize));
			pop(&tmpcurargsize, sizeof(tmpcurargsize)); // 20090421
			pop_t(&tinfo2);
			size2 = (tinfo2.pflg>0)?(sizeof(void *)):(tinfo2.size);
			pop(buf2, size2);
			push(&tmpcurargsize, sizeof(tmpcurargsize)); // 20090421
			push(&tmpargsize, sizeof(tmpargsize));
			push(buf2, size2);
			push_t(&tinfo2);
			break;
		case POPUP:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			break;
		case FENT:
			/* do nothing */
			break;
		case FRET:
			sp = fp;
			pop(&ic, sizeof(ic));
//			pop(&fsize, sizeof(fsize));
			pop(&fp, sizeof(fp));
			pop(&argsize, sizeof(argsize));
			sp -= argsize;
			pop(&argsize, sizeof(argsize));
			pop(&curargsize, sizeof(curargsize)); // 20090421
			//curargsize = argsize; // 20090421
			/* ͤ򥹥åѤ  */ ////////////////////////////
			tinfo1.type = TYPE_INT;
			tinfo1.size = sizeof(int);
			tinfo1.pflg = 0;
			pi_buf1 = (int *)buf1;
			*pi_buf1 = 0;
//			(*((int *)buf1)) = 0;
			push(buf1, sizeof(int));
			push_t(&tinfo1);
			break;
		case CPVAL:
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
			pop(buf1, size1);
			push(buf1, size1);
			push_t(&tinfo1);
			push(buf1, size1);
			push_t(&tinfo1);
			break;
		case MOV:	/* 쥸ޥˤݤ˼??? */
			{
				/* intǤȤƽ٤˷㥹Ȥ */
				intrcode_t code;
				code.op = CAST;
				code.o1 = TYPE_INT;
				code.o2 = 4;
				code.reg = 0;
				code.line = 0;
				typecast(&code);
			}
			pop_t(&tinfo1);
			size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
//			memset(buf1, 0, sizeof(buf1));
			pop(buf1, size1);
			/* intǤȤƽ */
			pi_buf1 = (int *)buf1;
			reg[code[ic].reg] = *pi_buf1;
//			reg[code[ic].reg] = *((int *)buf1);
			break;
		case CMPR:	/* ꤵ줿쥸ͤ */
			pi_buf1 = (int *)buf1;
			if (reg[code[ic].reg] == code[ic].o1)
//				(*((int *)buf1)) = 1;
				*pi_buf1 = 1;
			else
//				(*((int *)buf1)) = 0;
				*pi_buf1 = 0;
			push(buf1, sizeof(int));
			tinfo1.type = TYPE_INT;
			tinfo1.size = sizeof(int);
			tinfo1.pflg = 0;
			push_t(&tinfo1);
			break;
		case ADRS:	/* ¤ϲ⤷Ƥʤ... */
//			pop_t(&tinfo1);
//			if (tinfo1.type == STRUCT || tinfo1.type == UNION)
//				pop(buf1, sizeof(void *));
//			else if (tinfo1.pflg)
//				pop(buf1, sizeof(void *));
//			else	pop(buf1, tinfo1.size); /* error ? */
////			printf("_ADRRRRS %p\n", (int*)(*(int *)buf1));
//			push(&buf1, sizeof(int));
//			tinfo.type = tinfo1.type;
//			tinfo.size = tinfo.size;
//			tinfo.pflg = tinfo1.pflg;
//			push_t(&tinfo);
			break;
		case INVT:
		case NOT:	/* ΤߤǤα黻? */
		case CMPL:
			monomial(code[ic].op);
			break;
		case TYPE:
			tinfo1.type = code[ic].o1;
			tinfo1.size = code[ic].o2;
			tinfo1.pflg = code[ic].reg;
			push_t(&tinfo1);
			break;
		case CAST:
			typecast(&code[ic]);
			break;
		case EXIT:
//			if (sp)
//				runerror("...stack pointer is not zero");
			if (sp != chksp) {
				printf("sp %d, chksp %d\n", sp, chksp); // 20090421
				runerror("...stack pointer is broken");
			}
			return(0);
		case NOP:
			/* do nothing */
			break;
		default:
			printf("undefined op code: %d\n", code[ic].op);
			runerror("system error: interpret");
		}
		ic++;
	}
	return(0);
}

static int assign(intrcode_t *code)
{
	type_info tinfo1;
	type_info tinfo2;
	int *pi_buf1;
//	int *pi_buf2;
//	int **ppi_buf1;
	int **ppi_buf2;
	uint *pui_buf1;
//	uint *pui_buf2;
//	uint **ppui_buf1;
	uint **ppui_buf2;
//	double *pd_buf1;
//	double *pd_buf2;
//	double **ppd_buf1;
	double **ppd_buf2;
//	char *pc_buf1;
//	char *pc_buf2;
//	char **ppc_buf1;
	char **ppc_buf2;
//	uchar *puc_buf1;
//	uchar *puc_buf2;
//	uchar **ppuc_buf1;
	uchar **ppuc_buf2;
	char buf1[256];
	char buf2[256];
	int size1;

	pop_t(&tinfo1);
	size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
	pop(buf1, size1);

	/* ¦ϥɥ쥹ǼƤ */
	pop_t(&tinfo2);
	if (tinfo2.pflg <= 0)
		runerror2("System error ASGN", code->line);
	pop(buf2, sizeof(void *));

	if (tinfo2.pflg == 1) {
		switch (tinfo2.type) {
		case TYPE_INT:
			pi_buf1 = (int *)buf1;
			ppi_buf2 = (int **)buf2;
			if (size1 == sizeof(int))
//				*(*((int **)buf2)) = *((int *)buf1);
				**ppi_buf2 = *pi_buf1;
			else
//				*(*((int **)buf2)) = cast2int(&tinfo1, buf1);
				**ppi_buf2 = cast2int(&tinfo1, buf1);
			break;
		case TYPE_UINT:
			pui_buf1 = (uint *)buf1;
			ppui_buf2 = (uint **)buf2;
			if (size1 == sizeof(int))
//				*(*((uint **)buf2)) = *((uint *)buf1);
				**ppui_buf2 = *pui_buf1;
			else
//				*(*((uint **)buf2)) = cast2uint(&tinfo1, buf1);
				**ppui_buf2 = cast2uint(&tinfo1, buf1);
			break;
		case TYPE_DOUBLE:
			ppd_buf2 = (double **)buf2;
//			*(*((double **)buf2)) = cast2dbl(&tinfo1, buf1);
			**ppd_buf2 = cast2dbl(&tinfo1, buf1);
			break;
		case TYPE_CHAR:
			ppc_buf2 = (char **)buf2;
//			*(*((char **)buf2)) = cast2chr(&tinfo1, buf1);
			**ppc_buf2 = cast2chr(&tinfo1, buf1);
			break;
		case TYPE_UCHAR:
			ppuc_buf2 = (uchar **)buf2;
//			*(*((uchar **)buf2)) = cast2uchr(&tinfo1, buf1);
			**ppuc_buf2 = cast2uchr(&tinfo1, buf1);
			break;
		default:
			if (size1 == sizeof(int)) {
				pi_buf1 = (int *)buf1;
				ppi_buf2 = (int **)buf2;
//				*(*((int **)buf2)) = *((int *)buf1);
				**ppi_buf2 = *pi_buf1;
			}
			else {
				ppc_buf2 = (char **)buf2;
//				memcpy(*((char **)buf2), buf1, size1);
				memcpy(*ppc_buf2, buf1, size1);
			}
		}
	}
//	else if (size1 == sizeof(int)) {
//		*(*((int **)buf2)) = *((int *)buf1);
//	}
	else {
//{
//int ii;
//for (ii = 0; ii < size1; ii++) {
//  printf("%02x ", (int)((unsigned char)buf1[ii]));
//}
//printf("\n");
//for (ii = 0; ii < sizeof(void *); ii++) {
//  printf("%02x ", (int)((unsigned char)buf2[ii]));
//}
//printf("\n");
//}
		ppc_buf2 = (char **)buf2;
//		memcpy(*((char **)buf2), buf1, size1);
//		memcpy(*ppc_buf2, buf1, size1);
		*((char **)(*ppc_buf2)) = (char *)cast2ptr(&tinfo1, buf1);

	}

	/* 줿¦ѿ򥹥åѤ */
	push(buf2, sizeof(void *));
	push_t(&tinfo2);
#ifdef DEBUG
	debugASGN(&tinfo2, (int *)(*((int *)buf2)));
#endif
	return(0);
}

static int get_type_size(int type)
{
	int size;

	switch (type) {
	case TYPE_CHAR:
	case TYPE_UCHAR:
		size = sizeof(char);
		break;
	case TYPE_DOUBLE:
	case TYPE_FLOAT:
		size = sizeof(double);
		break;
	case TYPE_INT:
	case TYPE_LONG:
	case TYPE_SHORT:
	case TYPE_UINT:
	case TYPE_ULONG:
	case TYPE_USHORT:
		size = sizeof(int);
		break;
	default:
		return(-1);
	}
	return(size);
}

static int chkpointer(intrcode_t *code,
                      type_info *ptinfo1,
                      char *buf1,
                      type_info *ptinfo2,
                      char *buf2)
{
	int pflg1, pflg2;
	int val1, val2;
	int size;

	pflg1 = ptinfo1->pflg;
	pflg2 = ptinfo2->pflg;
	if (pflg1 == 0 && pflg2 == 0)
		return(0);

	if (code->op == ADD) {
		if (pflg1 > 0 && pflg2 > 0) {
#if DEBUG
//			printf("(%d, %d)\n", pflg1, pflg2);
#endif
			return(-1);
		}
	}
	else if (code->op == SUB) {
		if (pflg1 > 0 && pflg2 > 0) {
			if (ptinfo1->type != ptinfo2->type)
				return(-1);
			else
				return(0);
		}
		if (pflg1 != 0 && pflg2 != 0)
			return(-1);
	}

	if (pflg2 > 0) {
#if DEBUG
//			printf("pflg2: %d\n", pflg2);
#endif
		val1 = *((int *)buf1);
		if (pflg2 == 1) {
			size = get_type_size(ptinfo2->type);
			if (size < 0)
				*((int *)buf1) = val1 * ptinfo2->size;
			else
				*((int *)buf1) = val1 * size;
		}
		else	*((int *)buf1) = val1 * sizeof(void *);
	}
	else { // (pflg1 > 0)
#if DEBUG
//			printf("pflg1: %d\n", pflg1);
#endif
		val2 = *((int *)buf2);
		if (pflg1 == 1) {
			size = get_type_size(ptinfo1->type);
			if (size < 0)
				*((int *)buf2) = val2 * ptinfo1->size;
			else
				*((int *)buf2) = val2 * size;
		}
		else	*((int *)buf2) = val2 * sizeof(void *);
	}

	return(0);
}

static int getoptype(int op, type_info *ptinfo1, type_info *ptinfo2)
{
	int type1, type2;

	type1 = ptinfo1->type;
	type2 = ptinfo2->type;

	if (ptinfo1->pflg > 0 || ptinfo2->pflg > 0) {
//		/* ݥ󥿱黻ϤȤꤢintΥǷ׻ */
//		return(TYPE_INT);
		return(TYPE_POINTER);
	}
	else if (type1 == TYPE_DOUBLE || type2 == TYPE_DOUBLE) {
		return(TYPE_DOUBLE);
	}
	else if (type1 == TYPE_FLOAT || type2 == TYPE_FLOAT) {
		return(TYPE_DOUBLE);
	}
	else if (type1 == NUMBER || type2 == NUMBER) {
		return(TYPE_INT);
	}
	else if (type1 == TYPE_INT || type2 == TYPE_INT) {
		return(TYPE_INT);
	}
	else if (type1 == TYPE_LONG || type2 == TYPE_LONG) {
		return(TYPE_INT);
	}
	else if (type1 == TYPE_SHORT || type2 == TYPE_SHORT) {
		return(TYPE_INT);
	}
	else if (type1 == TYPE_CHAR || type2 == TYPE_CHAR) {
		return(TYPE_CHAR);
	}
#if DEBUG
//	printf("(type1, type2) = (%d, %d)\n", type1, type2);
#endif
	return(-1);
}

//static int cast2int(type_info *ptinfo, char *buf)
int cast2int(type_info *ptinfo, char *buf)
{
	if (ptinfo->pflg > 0) {
		/* ݥ󥿱黻ϤȤꤢintΥǷ׻ */
		return(*((int *)buf));
	}
	switch (ptinfo->type) {
	case TYPE_DOUBLE:	return((int)(*((double *)buf)));
	case TYPE_FLOAT:	return((int)(*((float *)buf)));
	case NUMBER:		return(*((int *)buf));
	case TYPE_INT:		return(*((int *)buf));
	case TYPE_LONG:		return((int)(*((long *)buf)));
	case TYPE_SHORT:	return((int)(*((short *)buf)));
	case TYPE_CHAR:		return((int)(*((char *)buf)));
	case CHARACTER:		return((int)(*((char *)buf)));
	case TYPE_UCHAR:	return((int)(*((uchar *)buf)));
	case TYPE_UINT:		return((int)(*((uint *)buf)));
	case TYPE_ULONG:	return((int)(*((ulong *)buf)));
	case TYPE_USHORT:	return((int)(*((ushort *)buf)));
	default:		return(*((int *)buf));
	}
}

static uint cast2uint(type_info *ptinfo, char *buf)
{
	if (ptinfo->pflg > 0) {
		/* ݥ󥿱黻ϤȤꤢintΥǷ׻ */
		return(*((uint *)buf));
	}
	switch (ptinfo->type) {
	case TYPE_DOUBLE:	return((uint)(*((double *)buf)));
	case TYPE_FLOAT:	return((uint)(*((float *)buf)));
	case NUMBER:		return((uint)(*((int *)buf)));
	case TYPE_INT:		return((uint)(*((int *)buf)));
	case TYPE_LONG:		return((uint)(*((long *)buf)));
	case TYPE_SHORT:	return((uint)(*((short *)buf)));
	case TYPE_CHAR:		return((uint)(*((char *)buf)));
	case CHARACTER:		return((uint)(*((char *)buf)));
	case TYPE_UCHAR:	return((uint)(*((uchar *)buf)));
	case TYPE_UINT:		return(*((uint *)buf));
	case TYPE_ULONG:	return((uint)(*((ulong *)buf)));
	case TYPE_USHORT:	return((uint)(*((ushort *)buf)));
	default:		return(*((uint *)buf));
	}
}

static double cast2dbl(type_info *ptinfo, char *buf)
{
	if (ptinfo->pflg > 0) {
		/* error ? */
		return((double)(*((int *)buf)));
	}
	switch (ptinfo->type) {
	case TYPE_DOUBLE:	return(*((double *)buf));
	case TYPE_FLOAT:	return((double)(*((float *)buf)));
	case NUMBER:		return((double)(*((int *)buf)));
	case TYPE_INT:		return((double)(*((int *)buf)));
	case TYPE_LONG:		return((double)(*((long *)buf)));
	case TYPE_SHORT:	return((double)(*((short *)buf)));
	case TYPE_CHAR:		return((double)(*((char *)buf)));
	case TYPE_UCHAR:	return((double)(*((uchar *)buf)));
	case TYPE_UINT:		return((double)(*((uint *)buf)));
	case TYPE_ULONG:	return((double)(*((ulong *)buf)));
	case TYPE_USHORT:	return((double)(*((ushort *)buf)));
	default:		return(*((double *)buf));
	}
}

static char cast2chr(type_info *ptinfo, char *buf)
{
	if (ptinfo->pflg > 0) {
		/* error ? */
		return((char)(*((int *)buf)));
	}
	switch (ptinfo->type) {
	case TYPE_DOUBLE:	return((char)(*((double *)buf)));
	case TYPE_FLOAT:	return((char)(*((float *)buf)));
	case NUMBER:		return((char)(*((int *)buf)));
	case TYPE_INT:		return((char)(*((int *)buf)));
	case TYPE_LONG:		return((char)(*((long *)buf)));
	case TYPE_SHORT:	return((char)(*((short *)buf)));
	case TYPE_CHAR:		return(*((char *)buf));
	case TYPE_UCHAR:	return((char)(*((uchar *)buf)));
	case TYPE_UINT:		return((char)(*((uint *)buf)));
	case TYPE_ULONG:	return((char)(*((ulong *)buf)));
	case TYPE_USHORT:	return((char)(*((ushort *)buf)));
	default:		return(*((char *)buf));
	}
}

static char cast2uchr(type_info *ptinfo, char *buf)
{
	if (ptinfo->pflg > 0) {
		/* error ? */
		return((uchar)(*((int *)buf)));
	}
	switch (ptinfo->type) {
	case TYPE_DOUBLE:	return((uchar)(*((double *)buf)));
	case TYPE_FLOAT:	return((uchar)(*((float *)buf)));
	case NUMBER:		return((uchar)(*((int *)buf)));
	case TYPE_INT:		return((uchar)(*((int *)buf)));
	case TYPE_LONG:		return((uchar)(*((long *)buf)));
	case TYPE_SHORT:	return((uchar)(*((short *)buf)));
	case TYPE_CHAR:		return((uchar)(*((char *)buf)));
	case TYPE_UCHAR:	return(*((uchar *)buf));
	case TYPE_UINT:		return((uchar)(*((uint *)buf)));
	case TYPE_ULONG:	return((uchar)(*((ulong *)buf)));
	case TYPE_USHORT:	return((uchar)(*((ushort *)buf)));
	default:		return(*((uchar *)buf));
	}
}

static char *cast2ptr(type_info *ptinfo, char *buf)
{
	if (ptinfo->pflg > 0) {
		/* ݥ󥿱黻ϤȤꤢintΥǷ׻ */
		return(*((char **)buf));
	}
	switch (ptinfo->type) {
	case TYPE_DOUBLE:	return((char *)((long)(*((double *)buf))));
	case TYPE_FLOAT:	return((char *)((long)(*((float *)buf))));
	case NUMBER:		return((char *)(*((int *)buf)));
	case TYPE_INT:		return((char *)(*((int *)buf)));
	case TYPE_LONG:		return((char *)(*((long *)buf)));
	case TYPE_SHORT:	return((char *)((int)(*((short *)buf))));
	case TYPE_CHAR:		return((char *)((int)(*((char *)buf))));
	case CHARACTER:		return((char *)((int)(*((char *)buf))));
	case TYPE_UCHAR:	return((char *)((int)(*((uchar *)buf))));
	case TYPE_UINT:		return((char *)(*((uint *)buf)));
	case TYPE_ULONG:	return((char *)(*((ulong *)buf)));
	case TYPE_USHORT:	return((char *)((int)(*((ushort *)buf))));
	default:		return(*((char **)buf));
	}
}

static int
binomial_int(int op, type_info *pif1, char *b1, type_info *pif2, char *b2)
{
	int val1, val2;

	val1 = cast2int(pif1, b1);
	val2 = cast2int(pif2, b2);

	switch (op) {
	case ADD:	val1 = val2 + val1;	break;
	case SUB:	val1 = val2 - val1;	break;
	case MUL:	val1 = val2 * val1;	break;
	case DIV:	val1 = val2 / val1;	break;
	case MOD:	val1 = val2 % val1;	break;
	case OR:	val1 = val2 | val1;	break;
	case XOR:	val1 = val2 ^ val1;	break;
	case AND:	val1 = val2 & val1;	break;
	case CMP:	val1 = (val2 == val1);	break;
	case NCMP:	val1 = (val2 != val1);	break;
	case LT:	val1 = val2 < val1;	break;
	case LE:	val1 = val2 <= val1;	break;
	case GT:	val1 = val2 > val1;	break;
	case GE:	val1 = val2 >= val1;	break;
	case LSFT:	val1 = val2 << val1;	break;
	case RSFT:	val1 = val2 >> val1;	break;
	default:
//		runerror2("System error binomial", code->line);
		runerror("System error binomial_int");
	}
	push(&val1, sizeof(int));
	if (pif1->pflg == 0 &&  pif2->pflg == 0) {
		pif1->type = TYPE_INT;
		pif1->size = sizeof(int);
		pif1->pflg = 0;
		push_t(pif1);
		
	}
	else if (pif1->pflg >= pif2->pflg) {
		push_t(pif1);
	}
	else {
		push_t(pif2);
	}

	return(0);
}

static int
binomial_dbl(int op, type_info *pif1, char *b1, type_info *pif2, char *b2)
{
	double val1, val2;
	int ival;

	val1 = cast2dbl(pif1, b1);
	val2 = cast2dbl(pif2, b2);

	switch (op) {
	case ADD:	val1 = val2 + val1;	break;
	case SUB:	val1 = val2 - val1;	break;
	case MUL:	val1 = val2 * val1;	break;
	case DIV:	val1 = val2 / val1;	break;
//	case MOD:	val1 = val2 % val1;	break;
//	case OR:	val1 = val2 | val1;	break;
//	case XOR:	val1 = val2 ^ val1;	break;
//	case AND:	val1 = val2 & val1;	break;
	case CMP:	val1 = (val2 == val1);	break;
	case NCMP:	val1 = (val2 != val1);	break;
	case LT:	val1 = val2 < val1;	break;
	case LE:	val1 = val2 <= val1;	break;
	case GT:	val1 = val2 > val1;	break;
	case GE:	val1 = val2 >= val1;	break;
//	case LSFT:	val1 = val2 << val1;	break;
//	case RSFT:	val1 = val2 >> val1;	break;
	default:
//		runerror2("System error binomial", code->line);
		runerror("System error binomial_int");
	}

	if (pif1->pflg == 0 &&  pif2->pflg == 0) {
		switch (op) {
		case CMP:
		case NCMP:
		case LT:
		case LE:
		case GT:
		case GE:
			ival = val1;
			push(&ival, sizeof(int));
			pif1->type = TYPE_INT;
			pif1->size = sizeof(int);
			pif1->pflg = 0;
			push_t(pif1);
			break;
		default:
			push(&val1, sizeof(double));
			pif1->type = TYPE_DOUBLE;
			pif1->size = sizeof(double);
			pif1->pflg = 0;
			push_t(pif1);
			break;
		}
	}
	else if (pif1->pflg >= pif2->pflg) {
		push(&val1, sizeof(double));
		push_t(pif1);
	}
	else {
		push(&val1, sizeof(double));
		push_t(pif2);
	}

	return(0);
}

static int
binomial_chr(int op, type_info *pif1, char *b1, type_info *pif2, char *b2)
{
	char val1, val2;
	int ival;

	val1 = cast2chr(pif1, b1);
	val2 = cast2chr(pif2, b2);

	switch (op) {
	case ADD:	val1 = val2 + val1;	break;
	case SUB:	val1 = val2 - val1;	break;
	case MUL:	val1 = val2 * val1;	break;
	case DIV:	val1 = val2 / val1;	break;
	case MOD:	val1 = val2 % val1;	break;
	case OR:	val1 = val2 | val1;	break;
	case XOR:	val1 = val2 ^ val1;	break;
	case AND:	val1 = val2 & val1;	break;
	case CMP:	val1 = (val2 == val1);	break;
	case NCMP:	val1 = (val2 != val1);	break;
	case LT:	val1 = val2 < val1;	break;
	case LE:	val1 = val2 <= val1;	break;
	case GT:	val1 = val2 > val1;	break;
	case GE:	val1 = val2 >= val1;	break;
	case LSFT:	val1 = val2 << val1;	break;
	case RSFT:	val1 = val2 >> val1;	break;
	default:
//		runerror2("System error binomial", code->line);
		runerror("System error binomial_int");
	}

	if (pif1->pflg == 0 &&  pif2->pflg == 0) {
		switch (op) {
		case CMP:
		case NCMP:
		case LT:
		case LE:
		case GT:
		case GE:
			ival = val1;
			push(&ival, sizeof(int));
			pif1->type = TYPE_INT;
			pif1->size = sizeof(int);
			pif1->pflg = 0;
			push_t(pif1);
			break;
		default:
			push(&val1, sizeof(char));
			pif1->type = TYPE_CHAR;
			pif1->size = sizeof(char);
			pif1->pflg = 0;
			push_t(pif1);
			break;
		}
	}
	else if (pif1->pflg >= pif2->pflg) {
		push(&val1, sizeof(char));
		push_t(pif1);
	}
	else {
		push(&val1, sizeof(char));
		push_t(pif2);
	}

	return(0);
}

static int
binomial_ptr(int op, type_info *pif1, char *b1, type_info *pif2, char *b2)
{
	char *val1 = 0;
	char *val2 = 0;
	int ival = 0;

	if (pif1->pflg > 0 && pif2->pflg <= 0) {
		if (op == SUB) {
			runerror("System error binomial_ptr (1)");
		}
		if (op == ADD) {
			val1 = cast2ptr(pif1, b1);
			ival = cast2int(pif2, b2);
		}
		else {
			val1 = cast2ptr(pif1, b1);
			val2 = cast2ptr(pif2, b2);
		}
//printf("1 : %p, %d\n", val1, ival);
	}
	else if (pif1->pflg <= 0 && pif2->pflg > 0) {
		if (op == ADD || op == SUB) {
			ival = cast2int(pif1, b1);
			val1 = cast2ptr(pif2, b2);
//printf("2 : %d, %p\n", ival, val1);
		}
		else {
			val1 = cast2ptr(pif1, b1);
			val2 = cast2ptr(pif2, b2);
		}
	}
	else if (pif1->pflg > 0 && pif2->pflg > 0) {
		if (op == ADD || op == SUB) {
			ival = cast2int(pif1, b1);
			val1 = cast2ptr(pif2, b2);
//printf("3 : %d, %p\n", ival, val1);
		}
		else {
			val1 = cast2ptr(pif1, b1);
			val2 = cast2ptr(pif2, b2);
//printf("4 : %p, %p\n", val1, val2);
		}
	}
	else {
		runerror("System error binomial_ptr (2)");
	}

	switch (op) {
	case ADD:	val1 = val1 + ival;	break;
	case SUB:	val1 = val1 - ival;	break;
//	case MUL:	val1 = val2 * val1;	break;
//	case DIV:	val1 = val2 / val1;	break;
//	case MOD:	val1 = val2 % val1;	break;
//	case OR:	val1 = val2 | val1;	break;
//	case XOR:	val1 = val2 ^ val1;	break;
//	case AND:	val1 = val2 & val1;	break;

	case CMP:	ival = (val2 == val1);	break;
	case NCMP:	ival = (val2 != val1);	break;
	case LT:	ival = val2 < val1;	break;
	case LE:	ival = val2 <= val1;	break;
	case GT:	ival = val2 > val1;	break;
	case GE:	ival = val2 >= val1;	break;

//	case LSFT:	val1 = val2 << val1;	break;
//	case RSFT:	val1 = val2 >> val1;	break;
	default:
		runerror("invalid operands to binary");
	}
//	switch (op) {
//	case ADD:	printf("ADD\n");	break;
//	case SUB:	printf("SUBn");	break;
//	case CMP:	printf("CMP\n");	break;
//	case NCMP:	printf("NCMP\n");	break;
//	case LT:	printf("LT\n");	break;
//	case LE:	printf("LE\n");	break;
//	case GT:	printf("GT\n");	break;
//	case GE:	printf("GE\n");	break;
//	default:
//		runerror("System error binomial_ptr (3)");
//	}
//printf("val1 : %p\n", val1);
//printf("ival : %d\n", ival);
	switch (op) {
	case ADD:
	case SUB:
		push(&val1, sizeof(char *));
		if (pif1->pflg >= pif2->pflg) {
			push_t(pif1);
		}
		else {
			push_t(pif2);
		}
		break;
	case CMP:
	case NCMP:
	case LT:
	case LE:
	case GT:
	case GE:
		push(&ival, sizeof(int));
		pif1->type = TYPE_INT;
		pif1->size = sizeof(int);
		pif1->pflg = 0;
		push_t(pif1);
		break;
	default:
		runerror("System error binomial_ptr (4)");
	}

	return(0);
}

/* a binomial expression formula */
static int binomial(intrcode_t *code)
{
	type_info tinfo1;
	type_info tinfo2;
	char buf1[256];
	char buf2[256];
	int size1;
	int size2;
	int type;

	pop_t(&tinfo1);
	size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
	pop(buf1, size1);

	pop_t(&tinfo2);
	size2 = (tinfo2.pflg>0)?(sizeof(void *)):(tinfo2.size);
	pop(buf2, size2);

	if ((code->op == ADD || code->op == SUB) && code->reg > 0) {
		if (chkpointer(code, &tinfo1, buf1, &tinfo2, buf2) < 0)
			runerror("binomial...chkpointer");
	}

	if ((type = getoptype(code->op, &tinfo1, &tinfo2)) < 0) {
		type = TYPE_INT;
	}

	switch (type) {
	case TYPE_INT:
//printf("call binomial_int\n");
		binomial_int(code->op, &tinfo1, buf1, &tinfo2, buf2);
		break;
	case TYPE_DOUBLE:
//printf("call binomial_dbl\n");
		binomial_dbl(code->op, &tinfo1, buf1, &tinfo2, buf2);
		break;
	case TYPE_CHAR:
//printf("call binomial_chr\n");
		binomial_chr(code->op, &tinfo1, buf1, &tinfo2, buf2);
		break;
	case TYPE_POINTER:
//printf("call binomial_ptr\n");
//printf("buf1 %p : \n", *(char **)buf1);//, *(char **)buf1);
//printf("buf2 %p : \n", *(char **)buf2);//, *(char **)buf2);
		binomial_ptr(code->op, &tinfo1, buf1, &tinfo2, buf2);
		break;
	default:
		runerror("binomial...type");
	}

	return(0);
}

/* a monomial expression */
static int monomial(int op)
{
	type_info tinfo1;
	int *pi_buf1;
	int **ppi_buf1;
	double *pd_buf1;
	double **ppd_buf1;
	char buf1[256];

	pop_t(&tinfo1);
	if (tinfo1.type == STRUCT || tinfo1.type == UNION) {
		pop(buf1, sizeof(void *)); /* error */
	}
	else if (tinfo1.pflg) {
		pop(buf1, sizeof(void *));
	}
	else {
		pop(buf1, tinfo1.size);
	}

	/* ȤꤢintȤƱ黻 -> doubleб  */
	if (tinfo1.pflg > 0) { /* ݥ󥿤ξͤ */
		if (tinfo1.type == TYPE_DOUBLE) {
			pd_buf1 = (double *)buf1;
			ppd_buf1 = (double **)buf1;
//			(*((double *)buf1)) = *((double *)(*((double **)buf1)));
			*pd_buf1 = **ppd_buf1;
		}
		else {
			pi_buf1 = (int *)buf1;
			ppi_buf1 = (int **)buf1;
//			(*((int *)buf1)) = *((int *)(*((int **)buf1)));
			*pi_buf1 = **ppi_buf1;
		}
		tinfo1.pflg--;
	}

	// λǥݥ󥿤ξϥ顼
	if (tinfo1.pflg > 0) {
		runerror("System error monomial 1");
	}

	switch (op) {
	case INVT:
		/* -黻 */
		if (tinfo1.type == TYPE_DOUBLE) {
			pd_buf1 = (double *)buf1;
//			(*((double *)buf1)) = -(*((double *)buf1));
			*pd_buf1 = (-1) * (*pd_buf1);
		}
		else {
			pi_buf1 = (int *)buf1;
//			(*((int *)buf1)) = -(*((int *)buf1));
			*pi_buf1 = (-1) * (*pi_buf1);
		}
		break;
	case NOT: /* ΤߤǤα黻? */
		/* !黻 */
		pi_buf1 = (int *)buf1;
//		(*((int *)buf1)) = !(*((int *)buf1));
		*pi_buf1 = !(*pi_buf1);
		break;
	case CMPL:
		/* ~黻 */
		pi_buf1 = (int *)buf1;
//		(*((int *)buf1)) = ~(*((int *)buf1));
		*pi_buf1 = ~(*pi_buf1);
		break;
	default:
//		runerror2("System error binomial", code->line);
		runerror("System error monomial 2");
	}

	if (tinfo1.pflg > 0) {
		push(buf1, sizeof(void *));
	}
	else {
		push(buf1, tinfo1.size);
	}
	push_t(&tinfo1);

	return(0);
}

static int typecast(intrcode_t *code)
{
	type_info tinfo1;
	type_info tinfo2;
	int *pi_buf1;
	int *pi_buf2;
//	int **ppi_buf1;
//	int **ppi_buf2;
	uint *pui_buf1;
	uint *pui_buf2;
//	uint **ppui_buf1;
//	uint **ppui_buf2;
//	double *pd_buf1;
	double *pd_buf2;
//	double **ppd_buf1;
//	double **ppd_buf2;
//	char *pc_buf1;
//	char *pc_buf2;
//	char **ppc_buf1;
	char **ppc_buf2;
//	uchar *puc_buf1;
//	uchar *puc_buf2;
//	uchar **ppuc_buf1;
//	uchar **ppuc_buf2;
	char buf1[256];
	char buf2[256];
	int size1;
	int size2;

	pop_t(&tinfo1);
	size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
	pop(buf1, size1);

	/* ݥ󥿷ؤΥ㥹ȤǤϤʤ */
	if (code->reg == 0) {
		switch (code->o1) {
		case TYPE_INT:
			pi_buf1 = (int *)buf1;
			pi_buf2 = (int *)buf2;
			if (size1 == sizeof(int))
//				*((int *)buf2) = *((int *)buf1);
				*pi_buf2 = *pi_buf1;
			else
//				*((int *)buf2) = cast2int(&tinfo1, buf1);
				*pi_buf2 = cast2int(&tinfo1, buf1);
			break;
		case TYPE_UINT:
			pui_buf1 = (uint *)buf1;
			pui_buf2 = (uint *)buf2;
			if (size1 == sizeof(uint))
//				*((uint *)buf2) = *((uint *)buf1);
				*pui_buf2 = *pui_buf1;
			else
//				*((uint *)buf2) = cast2uint(&tinfo1, buf1);
				*pui_buf2 = cast2uint(&tinfo1, buf1);
			break;
		case TYPE_DOUBLE:
			pd_buf2 = (double *)buf2;
//			*((double *)buf2) = cast2dbl(&tinfo1, buf1);
			*pd_buf2 = cast2dbl(&tinfo1, buf1);
			break;
		case TYPE_CHAR:
			*((char *)buf2) = cast2chr(&tinfo1, buf1);
			break;
		case TYPE_UCHAR:
			*((uchar *)buf2) = cast2uchr(&tinfo1, buf1);
			break;
		default:
			if (size1 == sizeof(int)) {
				pi_buf1 = (int *)buf1;
				pi_buf2 = (int *)buf2;
//				*((int *)buf2) = *((int *)buf1);
				*pi_buf2 = *pi_buf1;
			}
			else {
				memcpy(buf2, buf1, size1);
			}
		}
	}
//	else if (size1 == sizeof(int)) {
//		*((int *)buf2) = *((int *)buf1);
//	}
	else {
//		*((int **)buf2) = (int *)(*((int **)buf1));
		/* (sizeof(int *)==sizeof(int))ȤߤʤƤ... */
//		*((int **)buf2) = (int *)cast2int(&tinfo1, buf1);
		ppc_buf2 = (char **)buf2;
//		*((int **)buf2) = (int *)cast2ptr(&tinfo1, buf1);
		*ppc_buf2 = (char *)cast2ptr(&tinfo1, buf1);
	}

	/* 㥹ȸѿ */
	tinfo2.type = code->o1;
	tinfo2.size = code->o2;
	tinfo2.pflg = code->reg;
	size2 = (tinfo2.pflg) ? (sizeof(void *)) : (tinfo2.size);

	/* 줿¦ѿ򥹥åѤ */
	push(buf2, size2);
	push_t(&tinfo2);
#ifdef DEBUG
	printf("tinfo2.type : %d\n", tinfo2.type);
	printf("tinfo2.size : %d\n", tinfo2.size);
	printf("tinfo2.pflg : %d\n", tinfo2.pflg);
	debugCAST(&tinfo2, (int *)buf2);
#endif
	return(0);
}

#define GETCHAR         -1
#define PUTCHAR         -2
#define PRINTF          -3
#define MALLOC          -4
#define FREE            -5
#define MEMCPY          -6
#define MEMSET          -7
#define STRCPY          -8
#define SPRINTF         -9
#define FOPEN           -10
#define FCLOSE          -11
#define FGETC           -12
#define FPUTC           -13
#define STRLEN          -14
#define RX_SEARCH       -15
#define FGETS           -16
#define AMATCH          -17
#define BF_SEARCH       -18
#define BF_RE_SEARCH    -19
#define BM_SEARCH       -20
#define BM_RE_SEARCH    -21
#define ATOI            -22
#define REALLOC         -23
#define QSORT           -24
#define EXIT            -25
#define SCANF           -26
#define STRCMP          -27
#define STRNCMP         -28
#define FPRINTF         -29
#define FPUTS           -30
#define GETC            -31
#define PUTC            -32
#define ABORT           -33
#define CLOCK           -34
#define TIME            -35
#define DIFFTIME        -36
#define MKTIME          -37
#define ASCTIME         -38
#define CTIME           -39
#define GMTIME          -40
#define LOCALTIME       -41
#define STRFTIME        -42
#define FSCANF          -43
#define SSCANF          -44

#define MOD (MODE_LOCAL|MODE_PARAM)

varinf_t putchar_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fprintf_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t printf_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t sprintf_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t malloc_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t free_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t memcpy_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t memset_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t strcpy_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fopen_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fclose_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fgetc_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fputc_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fputs_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t getc_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t putc_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t strlen_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t rx_search_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fgets_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t amatch_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t bf_search_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a4",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a5",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t atoi_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t realloc_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t qsort_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a4",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t exit_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t fscanf_arg[] = {
	{"f0",MOD,TYPENAME,{TYPE_CHAR},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"f1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a4",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a5",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a6",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a7",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a8",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a9",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b4",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b5",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b6",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b7",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b8",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b9",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t scanf_arg[] = {
	{"f0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a4",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a5",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a6",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a7",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a8",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a9",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b4",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b5",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b6",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b7",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b8",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b9",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t sscanf_arg[] = {
	{"f0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"f1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a4",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a5",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a6",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a7",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a8",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a9",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b4",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b5",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b6",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b7",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b8",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b9",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"b0",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t strcmp_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t strncmp_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t time_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_LONG},sizeof(long *),sizeof(long),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t difftime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_LONG},sizeof(long *),sizeof(long),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_LONG},sizeof(long *),sizeof(long),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t mktime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t asctime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t ctime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_LONG},sizeof(long *),sizeof(long),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t gmtime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_LONG},sizeof(long *),sizeof(long),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t localtime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_LONG},sizeof(long *),sizeof(long),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

varinf_t strftime_arg[] = {
	{"a1",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a2",MOD,TYPENAME,{TYPE_INT },sizeof(int),sizeof(int),0,0,0,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a3",MOD,TYPENAME,{TYPE_CHAR},sizeof(char *),sizeof(char),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
	{"a4",MOD,TYPENAME,{TYPE_VOID},sizeof(void *),sizeof(void),0,0,1,{-1,-1,-1,-1,-1,-1},1,0,0,0,0},
};

/* Ȥ߹ߴؿ(Intrinsic function)  */
intrfunc intrfuncs[] = {
	{"getchar",		GETCHAR,	0,	NULL		},
	{"putchar",		PUTCHAR,	1,	putchar_arg	},
	{"printf",		PRINTF ,	-1,	printf_arg	},
	{"malloc",		MALLOC ,	1,	malloc_arg	},
	{"free",		FREE   ,	1,	free_arg	},
	{"memcpy",		MEMCPY ,	3,	memcpy_arg	},
	{"memset",		MEMSET ,	3,	memset_arg	},
	{"strcpy",		STRCPY ,	2,	strcpy_arg	},
	{"sprintf",		SPRINTF,	-2,	sprintf_arg	},
	{"fopen",		FOPEN  ,	2,	fopen_arg	},
	{"fclose",		FCLOSE ,	1,	fclose_arg	},
	{"fgetc",		FGETC  ,	1,	fgetc_arg	},
	{"fputc",		FPUTC  ,	2,	fputc_arg	},
	{"strlen",		STRLEN ,	1,	strlen_arg	},
	{"rx_search",		RX_SEARCH,	3,	rx_search_arg	},
	{"fgets",		FGETS,		3,	fgets_arg	},
	{"amatch",		AMATCH,		2,	amatch_arg	},
	{"bf_search",		BF_SEARCH,	5,	bf_search_arg	},
	{"bf_re_search",	BF_RE_SEARCH,	5,	bf_search_arg	},
	{"bm_search",		BM_SEARCH,	5,	bf_search_arg	},
	{"bm_re_search",	BM_RE_SEARCH,	5,	bf_search_arg	},
	{"atoi",		ATOI,		1,	atoi_arg	},
	{"realloc",		REALLOC,	2,	realloc_arg	},
	{"qsort",		QSORT,		4,	qsort_arg	},
	{"exit",		EXIT,		1,	exit_arg	},
	{"scanf",		SCANF,		-21,	scanf_arg	},
	{"strcmp",		STRCMP,		2,	strcmp_arg	},
	{"strncmp",		STRNCMP,	3,	strncmp_arg	},
	{"fprintf",		FPRINTF,	-2,	fprintf_arg	},
	{"fputs",		FPUTS,		2,	fputs_arg	},
	{"getc",		GETC,		1,	getc_arg	},
	{"putc",		PUTC,		2,	putc_arg	},
	{"abort",		ABORT,		0,	NULL		},
	{"clock",		CLOCK,		0,	NULL		},
	{"time",		TIME,		1,	time_arg	},
	{"difftime",		DIFFTIME,	2,	difftime_arg	},
	{"mktime",		MKTIME,		1,	mktime_arg	},
	{"asctime",		ASCTIME,	1,	asctime_arg	},
	{"ctime",		CTIME,		1,	ctime_arg	},
	{"gmtime",		GMTIME,		1,	gmtime_arg	},
	{"localtime",		LOCALTIME,	1,	localtime_arg	},
	{"strftime",		STRFTIME,	4,	strftime_arg	},
	{"fscanf",		FSCANF,		-22,	fscanf_arg	},
	{"sscanf",		SSCANF,		-22,	sscanf_arg	},
};
int intrfuncs_cnt = sizeof(intrfuncs)/sizeof(intrfuncs[0]);

static void intrinsic_function(intrcode_t *code)
{
	switch (code->o1) {
	case GETCHAR:      if_getchar(code);      break;
	case PUTCHAR:      if_putchar(code);      break;
	case PRINTF:       if_printf(code);       break;
	case MALLOC:       if_malloc(code);       break;
	case FREE:         if_free(code);         break;
	case MEMCPY:       if_memcpy(code);       break;
	case MEMSET:       if_memset(code);       break;
	case STRCPY:       if_strcpy(code);       break;
	case FOPEN:        if_fopen(code);        break;
	case FCLOSE:       if_fclose(code);       break;
	case FGETC:        if_fgetc(code);        break;
	case FPUTC:        if_fputc(code);        break;
	case STRLEN:       if_strlen(code);       break;
	case SPRINTF:      if_sprintf(code);      break;
	case RX_SEARCH:    if_rx_search(code);    break;
	case FGETS:        if_fgets(code);        break;
	case AMATCH:       if_amatch(code);       break;
	case BF_SEARCH:    if_bf_search(code);    break;
	case BF_RE_SEARCH: if_bf_re_search(code); break;
	case BM_SEARCH:    if_bm_search(code);    break;
	case BM_RE_SEARCH: if_bm_re_search(code); break;
	case ATOI:         if_atoi(code);         break;
	case REALLOC:      if_realloc(code);      break;
	case QSORT:        if_qsort(code);        break;
	case EXIT:         if_exit(code);         break;
	case SCANF:        if_scanf(code);        break;
	case STRCMP:       if_strcmp(code);       break;
	case STRNCMP:      if_strncmp(code);      break;
	case FPRINTF:      if_fprintf(code);      break;
	case FPUTS:        if_fputs(code);        break;
	case GETC:         if_getc(code);         break;
	case PUTC:         if_putc(code);         break;
	case ABORT:        if_abort(code);        break;
	case CLOCK:        if_clock(code);        break;
	case TIME:         if_time(code);         break;
	case DIFFTIME:     if_difftime(code);     break;
	case MKTIME:       if_mktime(code);       break;
	case ASCTIME:      if_asctime(code);      break;
	case CTIME:        if_ctime(code);        break;
	case GMTIME:       if_gmtime(code);       break;
	case LOCALTIME:    if_localtime(code);    break;
	case STRFTIME:     if_strftime(code);     break;
	case FSCANF:       if_fscanf(code);       break;
	case SSCANF:       if_sscanf(code);       break;
	default:           if_dynfunc(code);      break;
	}
}

int vm_fp = 0;
int cb_i_pv_pv(const void *pv1, const void *pv2)
{
	char *value;
	type_info tinfo1;
	intrcode_t code;
	char buf1[256];
	int size1;
	int lfp, lsp, lic;
	int tmp_sp;
	int ret;

	tmp_sp = chksp;
	chksp = sp;
	lfp = fp;
	lsp = sp;
	lic = ic;
//printf("---- cb_i_pv_pv called ----\n");
//printf("1 : %d %d\n", sp, chksp);
//case FRAME:
	push(&curargsize, sizeof(curargsize)); // 20090421
	push(&argsize, sizeof(argsize));
	argsize = 0;

	/* Ĥͤ򥹥åѤ */
	value = (char *)pv1;
	push(&value, sizeof(value));
	/* 򥹥åѤ */
	tinfo1.type = TYPE_CHAR;
	tinfo1.size = sizeof(value);
	tinfo1.pflg = 1;
	push_t(&tinfo1);
	/* ɬפʤз㥹ȤԤå鷿 */
//	if (code[ic].o1)
//		typecast(&code[ic]);
	pop_t(&tinfo1);
	argsize += (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);

	/* Ĥͤ򥹥åѤ */
	value = (char *)pv2;
	push(&value, sizeof(value));
	/* 򥹥åѤ */
	tinfo1.type = TYPE_CHAR;
	tinfo1.size = sizeof(value);
	tinfo1.pflg = 1;
	push_t(&tinfo1);
	/* ɬפʤз㥹ȤԤå鷿 */
//	if (code[ic].o1)
//		typecast(&code[ic]);
	pop_t(&tinfo1);
	argsize += (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);

//case CALL:
	if (vm_fp >= 0) {
		/* ϸƤӽФ褦ѹ */
		/* ΤǡߤΥåݥ󥿤Ƥ */
		/* Ф줿ؿΥե졼ݥ󥿤Ȥʤ   */
		ic = 7; /* 9ܤEXITʤΤ8Ѥ */
		curargsize = argsize;
		push(&argsize, sizeof(argsize));
		push(&fp, sizeof(fp));
		push(&ic, sizeof(ic));
		fp = sp;
//		sp += 64; /* ѿΰΥ狼ʤ */
//		ic = vm_fp;
		sp += ((vm_fp & 0x7fff0000) >> 16);
		ic = vm_fp & 0x0000ffff;
//printf("fsize : %d \n", ((vm_fp & 0x7fff0000) >> 16));
//printf("ic    : %d \n", vm_fp & 0x0000ffff);
//printf("2 : %d %d\n", sp, chksp);
		interpret(get_code(), get_code_size(), 0);
//printf("3 : %d %d\n", sp, chksp);
		/* å̵ͤɤ߹ࡪ */
		ret = (*((int *)&stack[sp]));
	}
	else {
		int *pi_buf1;
		/* Υ֥å̤ƥ 20070516 */
		code.op = CALL;
		code.o1 = vm_fp;
		code.o2 = 4;
		code.reg = (int)"callback";
		code.line = 0;
		intrinsic_function(&code);
		/* åΥȥåפͤ */
		pop_t(&tinfo1);
		size1 = (tinfo1.pflg>0)?(sizeof(void *)):(tinfo1.size);
		pop(buf1, size1);
		pop(&argsize, sizeof(argsize));
		pop(&curargsize, sizeof(curargsize)); // 20090421
		//curargsize = argsize; // 20090421
//		/* ͤ򥹥åѤ */
//		push(buf1, size1);
//		push_t(&tinfo1);
		/* ͤ򥻥å */
		pi_buf1 = (int *)buf1;
//		ret = (*((int *)buf1));
		ret = *pi_buf1;
	}
	fp = lfp;
	sp = lsp;
	ic = lic;
//printf("4 : %d %d\n", sp, chksp);
	chksp = tmp_sp;
	return(ret);
}

/*****************************************************************************/
/* 鲼ϥǥХåѤδؿǤ                                        */
/*****************************************************************************/
/* functions for debugging */
#ifdef DEBUG
static int debugASGN(type_info *tinfo, int *ptmp)
{
	switch (tinfo->type) {
	case TYPE_INT:
		printf("ASGN (int) %d\n", *ptmp);
		break;
	case TYPE_UINT:
		printf("ASGN (unsigned int) %u\n", *ptmp);
		break;
	case TYPE_CHAR:
	case TYPE_UCHAR:
		if (tinfo->pflg == 1 && tinfo->type == TYPE_CHAR) {
			printf("ASGN (char) %c\n", *((char *)ptmp));
			break;
		}
		else if (tinfo->pflg == 1 && tinfo->type == TYPE_UCHAR) {
			printf("ASGN (unsigned char) %c\n", *((char *)ptmp));
			break;
		}
		/* falling to next... */
	case STRING:
		printf("ASGN (string) %s\n", (char *)(*ptmp));
		break;
	default:
		printf("ASGN (unknown) %d\n", *ptmp);
//		printf("ASGN (struct)\n");
		break;
	}
	return(0);
}

static int debugCAST(type_info *tinfo, int *ptmp)
{
	switch (tinfo->type) {
	case TYPE_INT:
		printf("CAST (int) %d\n", *ptmp);
		break;
	case TYPE_UINT:
		printf("CAST (unsigned int) %u\n", *ptmp);
		break;
	case TYPE_CHAR:
	case TYPE_UCHAR:
		if (tinfo->pflg == 0 && tinfo->type == TYPE_CHAR) {
			printf("CAST (char) %c\n", *((char *)ptmp));
			break;
		}
		else if (tinfo->pflg == 0 && tinfo->type == TYPE_UCHAR) {
			printf("CAST (unsigned char) %c\n", *((char *)ptmp));
			break;
		}
		/* falling to next... */
	case STRING:
		printf("CAST (string) %s\n", (char *)(*ptmp));
		break;
	case TYPE_DOUBLE:
		printf("CAST (double) %f\n", *((double *)ptmp));
		break;
	default:
		printf("CAST (unknown) %d\n", *ptmp);
//		printf("CAST (struct)\n");
		break;
	}
	return(0);
}
#endif
