/*
 *  lex.c - Lexical analysis
 *
 *  Copyright (C) 2003 Seiji Tamura
 */
#include <stdio.h>
#include "common.h"

/* memo */
/* øλϤ˲äƥץץåεǽΰ */
/* Ƥޤ  (#define #include #ifdef #ifndef #endif) */

#define cget() c=charget()

extern int errcnt;

/* global variables */
char id[IDSIZE + 1];
int value;
int tokenline;

/* global functions */
void initchar();
int yylex();

/* local variables */
static token_t token;
static int c = ' ';
static char string[STRSIZE];
static int strtop = 0;
#define QUE_TOKEN_TBL_SIZE 64
static token_t que_token_tbl[QUE_TOKEN_TBL_SIZE];
static int que_token_cnt = 0;

/* local functions */
static int getident(int);
static int getnumber();
static int getcharacter();
static int getstring();
static void enterstr(int v);
static int charval();
static int chkpreproc();
static int getdefine();
static int get_macro_value(macroinf_t *);
static int chkdefine();
static int getinclude();
static void settypedef_typename(typedefinf_t *);
static void settypedef_struct(typedefinf_t *);
static int chktypedef();
static int getifdef();

typedef struct reserved_word {
	char *ident;
	int type;
	int ttype;
}	reserved_word_t;

static reserved_word_t keywords[] = {
	{"auto",	NEGL,		NEGL		},
	{"break",	BREAK,		NEGL		},
	{"case",	CASE,		NEGL		},
	{"char",	TYPENAME,	TYPE_CHAR	},
	{"continue",	CONTINUE,	NEGL		},
	{"default",	DEFAULT,	NEGL		},
	{"define",	DEFINE,		NEGL		},
	{"do",		DO,		NEGL		},
	{"double",	TYPENAME,	TYPE_DOUBLE	},
	{"else",	ELSE,		NEGL		},
	{"endif",	ENDIF,		NEGL		},
	{"extern",	EXTERN,		NEGL		},
	{"float",	TYPENAME,	TYPE_FLOAT	},
	{"for",		FOR,		NEGL		},
	{"goto",	GOTO,		NEGL		},
	{"if",		IF,		NEGL		},
	{"ifdef",	IFDEF,		NEGL		},
	{"ifndef",	IFNDEF,		NEGL		},
	{"include",	INCLUDE,	NEGL		},
	{"int",		TYPENAME,	TYPE_INT	},
	{"long",	TYPENAME,	TYPE_LONG	},
	{"register",	NEGL,		NEGL		},
	{"return",	RETURN,		NEGL		},
	{"short",	TYPENAME,	TYPE_SHORT	},
	{"sizeof",	SIZEOF,		NEGL		},
	{"static",	STATIC,		NEGL		},
	{"struct",	STRUCT,		NEGL		},
	{"switch",	SWITCH,		NEGL		},
	{"typedef",	TYPEDEF,	NEGL		},
	{"union",	UNION,		NEGL		},
	{"unsigned",	UNSIGNED,	NEGL		},
	{"void",	TYPENAME,	TYPE_VOID	},
	{"while",	WHILE,		NEGL		}
};
#define keywords_cnt sizeof(keywords)/sizeof(keywords[0])

#define chkdigit(c) ((c) >= '0' && (c) <= '9')
#define chkhex(c)   (((c) >= '0' && (c) <= '9') || \
                     ((c) >= 'A' && (c) <= 'F') || ((c) >= 'a' && (c) <= 'f'))
#define chkalpha(c) (((c) >= 'a' && (c) <= 'z') || ((c) >= 'A' && (c) <= 'Z'))

#ifdef DEBUG
static int printToken(token_t *tok);
static char *get_type_name(int ttype);
#endif

void initchar()
{
	c = ' ';
}

void add_token_que(token_t *que_token)
{
	if (que_token_cnt >= QUE_TOKEN_TBL_SIZE)
		fatal("error: token que buffer overflow");
	que_token_tbl[que_token_cnt++] = *que_token;
}

token_t *get_token_que()
{
	static token_t cur_que_token;
	int i;

	if (que_token_cnt == 0)
		return(NULL);
	cur_que_token = que_token_tbl[0];
	for (i = 1; i < que_token_cnt; i++)
		que_token_tbl[i - 1] = que_token_tbl[i];
	que_token_cnt--;
	return(&cur_que_token);
}

token_t *gettoken()
{
	token_t *que_token;
	static int need_endif = 0;
	static int endif_cnt = 0;
	int tmp_need_endif = 0;

	if ((que_token = get_token_que()) != NULL) {
		token = *que_token;
		goto out;
	}

	rescan:
	token.ptflag = 0;
	while (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
		cget();
	}
	if (c == EOF) {
		if (endif_cnt != 0) {
			yyerror("#ifdef (#ifndef) and #endif count not match");
			yyerror2("#endif count  ", endif_cnt);
		}
		token.type = EOF;
		goto out;
	}
	if (chkalpha(c) || (c == '_')) {
		if (getident(0) == -1)
			goto rescan;
		chkdefine();
		if (token.type == IDENT)
			chktypedef();
		goto out;
	}
	else if (chkdigit(c)) {
		if (getnumber() == -1)
			goto rescan;
		goto out;
	}
	else if (c == '\'') {
		if (getcharacter() == -1)
			goto rescan;
		goto out;
	}
	else if (c == '\"') {
		if (getstring() == -1)
			goto rescan;
		goto out;
	}
	switch (c) {
	case ',':
		cget();
		token.type = COMMA;
		break;
	case '.':
		cget();
		if (c == '.') {
			cget();
			if (c == '.') {
				cget();
				token.type = TRIPLEDOT;
				break;
			}
			else {
				yyerror("unknown token");
				cget();
				goto rescan;
			}
		}
		token.type = DOT;
		break;
	case ';':
		cget();
		token.type = SEMICOLON;
		break;
	case ':':
		cget();
		token.type = COLON;
		break;
	case '{':
		cget();
		token.type = LBRACE;
		break;
	case '}':
		cget();
		token.type = RBRACE;
		break;
	case '(':
		cget();
		token.type = LPAR;
		break;
	case ')':
		cget();
		token.type = RPAR;
		break;
	case '[':
		cget();
		token.type = LBRACK;
		break;
	case ']':
		cget();
		token.type = RBRACK;
		break;
	case '*':
		cget();
		if (c == '=') {
			cget();
			token.type = STAREQ;
			break;
		}
		token.type = STAR;
		break;
	case '&':
		cget();
		if (c == '&') {
			cget();
			token.type = ANDAND;
			break;
		}
		else if (c == '=') {
			cget();
			token.type = ANDEQ;
			break;
		}
		token.type = AND;
		break;
	case '|':
		cget();
		if (c == '|') {
			cget();
			token.type = OROR;
			break;
		}
		else if (c == '=') {
			cget();
			token.type = OREQ;
			break;
		}
		token.type = OR;
		break;
	case '+':
		cget();
		if (c == '+') {
			cget();
			token.type = PLUSPLUS;
			break;
		}
		else if (c == '=') {
			cget();
			token.type = PLUSEQ;
			break;
		}
		token.type = PLUS;
		break;
	case '-':
		cget();
		if (c == '-') {
			cget();
			token.type = MINUSMINUS;
			break;
		}
		else if (c == '=') {
			cget();
			token.type = MINUSEQ;
			break;
		}
		else if (c == '>') {
			cget();
			token.type = ARROW;
			break;
		}
		token.type = MINUS;
		break;
	case '/':
		cget();
		if (c == '*') {
			cget();
			do {
				while (c != '*' && c != EOF) {
					cget();
				}
				cget();
			} while (c != '/' && c != EOF);
			if (c == EOF) {
				yyerror("comment not terminated");
				token.type = EOF;
				break;
			}
			cget();
			goto rescan;
		}
		else if (c == '/') {
			do {
				cget();
			} while (c != '\n' && c != EOF);
			if (c == EOF) {
				token.type = EOF;
				break;
			}
			cget();
			goto rescan;
		}
		else if (c == '=') {
			cget();
			token.type = SLASHEQ;
			break;
		}
		token.type = SLASH;
		break;
	case '!':
		cget();
		if (c == '=') {
			cget();
			token.type = NOTEQ;
			break;
		}
		token.type = EXCLAM;
		break;
	case '~':
		cget();
		token.type = TILDE;
		break;
	case '%':
		cget();
		if (c == '=') {
			cget();
			token.type = PERCENTEQ;
			break;
		}
		token.type = PERCENT;
		break;
	case '<':
		cget();
		if (c == '<') {
			cget();
			if (c == '=') {
				cget();
				token.type = LSHIFTEQ;
				break;
			}
			token.type = LSHIFT;
			break;
		}
		else if (c == '=') {
			cget();
			token.type = LE;
			break;
		}
		token.type = LT;
		break;
	case '>':
		cget();
		if (c == '>') {
			cget();
			if (c == '=') {
				cget();
				token.type = RSHIFTEQ;
				break;
			}
			token.type = RSHIFT;
			break;
		}
		else if (c == '=') {
			cget();
			token.type = GE;
			break;
		}
		token.type = GT;
		break;
	case '=':
		cget();
		if (c == '=') {
			cget();
			token.type = EQEQ;
			break;
		}
		token.type = EQUAL;
		break;
	case '^':
		cget();
		if (c == '=') {
			cget();
			token.type = UPARROWEQ;
			break;
		}
		token.type = UPARROW;
		break;
	case '#':
		if (chkpreproc() == 0) {
			char *err_endif = "#endif expected";
			if (token.type == DEFINE) {
				if (getdefine() == 0) {
					while (c != '\n' && c != EOF)
						cget();
					break;
				}
			}
			else if (token.type == INCLUDE) {
				if (getinclude() == 0) {
					while (c != '\n' && c != EOF)
						cget();
					break;
				}
			}
#define chk_endif() ((token.type == ENDIF) && (token.intnum == tmp_cnt))
#define chk_else() ((!need_endif) && (tmp_cnt == (endif_cnt - 1)))
			else if (token.type == IFDEF) {
				int tmp_cnt = endif_cnt;
				endif_cnt++;
				if (getifdef() != 1) {
					while (c != '\n' && c != EOF)
						cget();
					tmp_need_endif = need_endif;
					need_endif = 1;
					while (!chk_endif()) {
						gettoken();
						if (token.type == EOF) {
							yyerror(err_endif);
							break;
						}
						if (chk_else())
							break;
						need_endif = 1;
					}
					need_endif = tmp_need_endif;
					goto rescan;
				}
			}
			else if (token.type == IFNDEF) {
				int tmp_cnt = endif_cnt;
				endif_cnt++;
				if (getifdef() == 1) {
					while (c != '\n' && c != EOF)
						cget();
					tmp_need_endif = need_endif;
					need_endif = 1;
					while (!chk_endif()) {
						gettoken();
						if (token.type == EOF) {
							yyerror(err_endif);
							break;
						}
						if (chk_else())
							break;
						need_endif = 1;
					}
					need_endif = tmp_need_endif;
					goto rescan;
				}
			}
			else if (token.type == ELSE) {
				int tmp_cnt = endif_cnt - 1;
				if (need_endif) {
					need_endif = 0;
					goto out;
				}
				else {
					while (c != '\n' && c != EOF)
						cget();
					tmp_need_endif = need_endif;
					need_endif = 1;
					while (!chk_endif()) {
						gettoken();
						if (token.type == EOF) {
							yyerror(err_endif);
							break;
						}
						need_endif = 1;
					}
					need_endif = tmp_need_endif;
					goto rescan;
				}
			}
			else if (token.type == ENDIF) {
				endif_cnt--;
				token.intnum = endif_cnt;
				if (need_endif) {
					while (c != '\n' && c != EOF)
						cget();
					break;
				}
			}
		}
		while (c != '\n' && c != EOF) {
			cget();
		}
		if (c == EOF) {
			token.type = EOF;
			break;
		}
		cget();
		goto rescan;
	case '\?':
		cget();
		token.type = QUESTION;
		break;
	default:
		yyerror("invalid character");
		cget();
		goto rescan;
	}
out:
#ifdef DEBUG
	printToken(&token);
#endif
	return(&token);
}

static int getident(int preproc)
{
	int low, high, middle, x;

	low = 0;
	while (chkalpha(c) || (c == '_') || chkdigit(c)) {
		if (low < IDSIZE) {
			token.ident[low++] = c;
		}
		cget();
	}
	token.ident[low] = 0;
	low = 0;
	high = keywords_cnt - 1;
	while (low <= high) {
		middle = (low + high) / 2;
		x = compstr(token.ident, keywords[middle].ident);
		if (x == 0) {
			if (keywords[middle].type == NEGL) {
				yyerror("unsupported keyword");
				return(-1);
			}
			if (keywords[middle].type == TYPENAME) {
				token.ttype = keywords[middle].ttype;
				token.type = keywords[middle].type;
			}
			else if (keywords[middle].type == UNSIGNED) {
				gettoken();
				switch (token.type) {
				case TYPENAME:
					switch (token.ttype) {
					case TYPE_CHAR:
						token.ttype = TYPE_UCHAR;
						break;
					case TYPE_INT:
						token.ttype = TYPE_UINT;
						break;
					case TYPE_LONG:
						token.ttype = TYPE_ULONG;
						break;
					case TYPE_SHORT:
						token.ttype = TYPE_USHORT;
						break;
					default:
						yyerror("unsupported type");
						return(-1);
					}
					break;
				case IDENT:
					add_token_que(&token);
					token.type = TYPENAME;
					token.ttype = TYPE_UINT;
					token.ptflag = 0;
					break;
				default:
					yyerror("type or ident expected");
					return(-1);
				}
			}
			else {
				token.type = keywords[middle].type;
				if (!preproc) {
					switch (token.type) {
					case DEFINE:
					case INCLUDE:
					case IFDEF:
					case IFNDEF:
					/*case ELSE:*/
					case ENDIF:
						token.type = IDENT;
						break;
					default:
						break;
					}
				}
			}
			return(0);
		}
		else if (x < 0) {
			high = middle - 1;
		}
		else {
			low = middle + 1;
		}
	}
	token.type = IDENT;
	return(0);
}

static int getnumber()
{
	int base;
	//int dbase;
	double dbase;
	int num;
	double dvalue;

	if (c == '0') {
		cget();
		if (c == 'x' || c == 'X') {
			base = 16;
			cget();
		}
		else {
			base = 8;
		}
	}
	else {
		base = 10;
	}
	value = 0;
	dbase = 1;
	dvalue = 0.0;
	switch (base) {
	case 8:
	case 10:
		while (chkdigit(c)) {
			if (base == 8 && (c == '8' || c == '9')){
				yyerror("8 or 9 in octal number");
			}
			value = value * base + (c - '0');
			cget();
		}
		if (c != '.') {
			token.intnum = value;
			token.type = NUMBER;
			break;
		}
		if (base == 8 && value == 0) {
			base = 10;
		}
		if (base != 10) {
			yyerror("decimal of octal number not supported");
			cget();
			return(-1);
		}
		cget();
		while (chkdigit(c)) {
			dvalue = dvalue * base + (c - '0');
			dbase *= base;
			cget();
		}
		token.decnum = (double)value + (double)dvalue / dbase;
		token.type = DECIMAL;
		break;
	case 16:
		while (chkhex(c)) {
			if (c >= 'A' && c <= 'F')
				num = 10 + (c - 'A');
			else if (c >= 'a' && c <= 'f')
				num = 10 + (c - 'a');
			else
				num = (c - '0');
			value = value * base + num;
			cget();
		}
		if (c != '.') {
			token.intnum = value;
			token.type = NUMBER;
			break;
		}
		yyerror("decimal of hexical number not supported");
		cget();
		return(-1);
#if 0
		while (chkhex(c)) {
			if (c >= 'A' && c <= 'F')
				num = 10 + (c - 'A');
			else if (c >= 'a' && c <= 'f')
				num = 10 + (c - 'a');
			else
				num = (c - '0');
			dbase *= base;
			dvalue = dvalue + (double)num / dbase;
			cget();
		}
		dvalue += value;
		token.decnum = dvalue;
		token.type = DECIMAL;
		break;
#endif
	default:
		yyerror("system error!");
		return(-1);
	}
	return(0);
}

static int getcharacter()
{
	cget();
	if (c == '\'') {
		yyerror("null character literal");
	}
	else {
		value = charval();
	}
	if (c == '\'') {
		cget();
	}
	else {
		yyerror("character literal invalid format");
	}

	token.ch = value;

	token.type = CHARACTER;
	return(0);
}

static int getstring()
{
//	value = (int)(&string[strtop]);
	token.string = &string[strtop];
	cget();
	while (c != '\"' && c != '\n') {
		enterstr(charval());
	}
	if (c == '\"') {
		cget();
	}
	else {
		yyerror("string literal invalid format");
	}
	enterstr(0);
	token.type = STRING;
	return(0);
}

static void enterstr(int v)
{
	if (strtop >= STRSIZE) {
		fatal("string table overflow");
	}
	string[strtop++] = v;
}

static int charval()
{
	int v;
	int l;

//	if (c < 040 || c >= 0177) {
//		yyerror("control character in string");
//		cget();
//		return(0);
//	}
	if (c != '\\') {
		v = c;
		cget();
		return(v);
	}
	cget();
	if (c >= '0' && c <= '7') {
		v = 0;
		l = 0;
		while (l < 3 &&  c >= '0' && c <= '7') {
			v = (v << 3) + (c - '0');
			l++;
			cget();
		}
		return(v);
	}
	switch (c) {
	case 'n':
		v = '\n';
		break;
	case 't':
		v = '\t';
		break;
	case 'b':
		v = '\b';
		break;
	case 'r':
		v = '\r';
		break;
	case 'f':
		v = '\f';
		break;
	default:
		v = c;
	}
	cget();
	return(v);
}

static int chkpreproc()
{
	do {
		cget();
	} while (c == ' ' || c == '\t');

	if (chkalpha(c)) {
		if (getident(1) == -1)
			return(-1);
		switch (token.type) {
		case DEFINE:
		case INCLUDE:
		case IFDEF:
		case IFNDEF:
		case ELSE:
		case ENDIF:
			return(0);
		default:
			break;
		}
	}
	return(-1);
}

static int getdefine()
{
	static macroinf_t macro;

	while (c == ' ' || c == '\t') {
		cget();
	}
	if (chkalpha(c) || (c == '_')) {
		if (getident(0) == -1)
			return(-1);
		if (token.type != IDENT)
			return(-1);
	}
	else {
		return(-1);
	}
	copystr(macro.name, token.ident);
	while (c == ' ' || c == '\t') {
		cget();
	}
	if (get_macro_value(&macro) == -1) {
		return(-1);
	}
	token.type = DEFINE;
	token.macro = &macro;
	return(0);
}

static int get_macro_value(macroinf_t *pmacro)
{
	int num_sign = 1;
	if (c == '(') {
		cget();
		if (get_macro_value(pmacro) == -1)
			return(-1);
		if (c != ')') {
			yyerror("right parenthesis expected");
			return(-1);
		}
		cget();
		return(0);
	}
	if (c == '-') {
		num_sign = -1;
		cget();
	}

	if (chkdigit(c)) {
		if (getnumber() == -1)
			return(-1);
		pmacro->type = token.type;
		if (pmacro->type == NUMBER) {
			pmacro->intnum = token.intnum * num_sign;
		}
		else if (pmacro->type == DECIMAL) {
			pmacro->decnum = token.decnum * num_sign;
		}
	}
	else if (num_sign == -1) {
		yyerror("unsupported definition");
		return(-1);
	}
	else if (c == '\'') {
		if (getcharacter() == -1)
			return(-1);
		pmacro->type = token.type;
		pmacro->ch = token.ch;
	}
	else if (c == '\"') {
		if (getstring() == -1)
			return(-1);
		pmacro->type = token.type;
		pmacro->string = token.string;
	}
	else if (c == '\n' || c == '\r' || c == EOF) {
		pmacro->type = NUMBER;
		pmacro->intnum = 1;
	}
	else {
		yyerror("unsupported definition");
		return(-1);
	}
	return(0);
}

static int chkdefine()
{
	macroinf_t *pmacro;

	pmacro = search_macro(token.ident);
	if (pmacro) {
		switch (pmacro->type) {
		case NUMBER:
			token.intnum = pmacro->intnum;
			break;
		case DECIMAL:
			token.decnum = pmacro->decnum;
			break;
		case CHARACTER:
			token.ch = pmacro->ch;
			break;
		case STRING:
			token.string = pmacro->string;
			break;
		default:
			yyerror("system error!");
			return(-1);
		}
		token.type = pmacro->type;
	}
	return(0);
}

static int getinclude()
{
	static includeinf_t incinf;
	int idx;

	while (c == ' ' || c == '\t') {
		cget();
	}
	if (c == '\"') {
		cget();
		idx = 0;
		while (c != '\"' && c != '\n' && c != EOF) {
			if (idx < IDSIZE)
				incinf.name[idx++] = c;
			cget();
		}
		if (c != '\"')
			return(-1);
		incinf.name[idx] = 0;
		incinf.type = INCLUDE_USR;
	}
	else if (c == '<') {
		cget();
		idx = 0;
		while (c != '>' && c != '\n' && c != EOF) {
			if (idx < IDSIZE)
				incinf.name[idx++] = c;
			cget();
		}
		if (c != '>')
			return(-1);
		incinf.name[idx] = 0;
		incinf.type = INCLUDE_SYS;
	}
	else {
		return(-1);
	}
	token.type = INCLUDE;
	token.incinf = &incinf;
	return(0);
}

static void settypedef_typename(typedefinf_t *ptdefinf)
{
//	int i;
//	token_t star_tok;

	token.type = ptdefinf->type;
	token.ttype = ptdefinf->ttype;
//	star_tok.type = STAR;
//	for (i = 0; i < ptdefinf->ptflag; i++)
//		add_token_que(&star_tok);
	token.ptflag = ptdefinf->ptflag;
}

static void settypedef_struct(typedefinf_t *ptdefinf)
{
//	int i;
//	token_t star_tok;
	token_t struct_tok;

	token.type = ptdefinf->type;
	struct_tok.type = IDENT;
	copystr(struct_tok.ident, ptdefinf->tname);
	add_token_que(&struct_tok);
//	star_tok.type = STAR;
//	for (i = 0; i < ptdefinf->ptflag; i++)
//		add_token_que(&star_tok);
	token.ptflag = ptdefinf->ptflag;
}

static int chktypedef()
{
	typedefinf_t *ptdefinf;

	ptdefinf = search_typedef(token.ident);
	if (ptdefinf) {
		switch (ptdefinf->type) {
		case TYPENAME:
			settypedef_typename(ptdefinf);
			break;
		case STRUCT:
		case UNION:
			settypedef_struct(ptdefinf);
			break;
		default:
			yyerror("invalid typedef");
			return(-1);
		}
	}
	return(0);
}

/* #ifdefǻꤵƤå */
/* Ƥ                   :  1    */
/* Ƥʤä             :  0    */
/* оݤʸ#ifdefʹߤˤʤ : -1    */
static int getifdef()
{
	macroinf_t *pmacro;

	while (c == ' ' || c == '\t') {
		cget();
	}
	if (chkalpha(c) || (c == '_')) {
		if (getident(0) == -1)
			return(-1);
		if (token.type != IDENT)
			return(-1);
	}
	else {
		return(-1);
	}
	pmacro = search_macro(token.ident);
	if (pmacro)
		return(1);
	else
		return(0);
}

/*****************************************************************************/
/* 鲼ϥǥХåѤδؿǤ                                        */
/*****************************************************************************/
/* functions for debugging */
#ifdef DEBUG
static int printToken(token_t *tok)
{
	switch (tok->type) {
	case IDENT:
		printf("IDENT           : %s\n", tok->ident);
		break;
	case LITERAL:
		printf("LITERAL         :\n");
		break;
	case COMMA:
		printf("COMMA           :\n");
		break;
	case SEMICOLON:
		printf("SEMICOLON       :\n");
		break;
	case LBRACE:
		printf("LBRACE          :\n");
		break;
	case RBRACE:
		printf("RBRACE          :\n");
		break;
	case LPAR:
		printf("LPAR            :\n");
		break;
	case RPAR:
		printf("RPAR            :\n");
		break;
	case LBRACK:
		printf("LBRACK          :\n");
		break;
	case RBRACK:
		printf("RBRACK          :\n");
		break;
	case TYPENAME:
		printf("TYPENAME        : %s\n", get_type_name(tok->ttype));
		break;
	case BREAK:
		printf("BREAK           :\n");
		break;
	case ELSE:
		printf("ELSE            :\n");
		break;
	case IF:
		printf("IF              :\n");
		break;
	case RETURN:
		printf("RETURN          :\n");
		break;
	case WHILE:
		printf("WHILE           :\n");
		break;
	case EQUAL:
		printf("EQUAL           :\n");
		break;
	case OROR:
		printf("OROR            :\n");
		break;
	case ANDAND:
		printf("ANDAND          :\n");
		break;
	case OR:
		printf("OR              :\n");
		break;
	case UPARROW:
		printf("UPARROW         :\n");
		break;
	case AND:
		printf("AND             :\n");
		break;
	case EQEQ:
		printf("EQEQ            :\n");
		break;
	case NOTEQ:
		printf("NOTEQ           :\n");
		break;
	case LE:
		printf("LE              :\n");
		break;
	case LT:
		printf("LT              :\n");
		break;
	case GE:
		printf("GE              :\n");
		break;
	case GT:
		printf("GT              :\n");
		break;
	case LSHIFT:
		printf("LSHIFT          :\n");
		break;
	case RSHIFT:
		printf("RSHIFT          :\n");
		break;
	case PLUS:
		printf("PLUS            :\n");
		break;
	case MINUS:
		printf("MINUS           :\n");
		break;
	case STAR:
		printf("STAR            :\n");
		break;
	case SLASH:
		printf("SLASH           :\n");
		break;
	case PERCENT:
		printf("PERCENT         :\n");
		break;
	case PLUSPLUS:
		printf("PLUSPLUS        :\n");
		break;
	case MINUSMINUS:
		printf("MINUSMINUS      :\n");
		break;
	case EXCLAM:
		printf("EXCLAM          :\n");
		break;
	case TILDE:
		printf("TILDE           :\n");
		break;
	case FOR:
		printf("FOR             :\n");
		break;
	case CASE:
		printf("CASE            :\n");
		break;
	case DEFAULT:
		printf("DEFAULT         :\n");
		break;
	case SWITCH:
		printf("SWITCH          :\n");
		break;
	case COLON:
		printf("COLON           :\n");
		break;
	case CONTINUE:
		printf("CONTINUE        :\n");
		break;
	case QUESTION:
		printf("QUESTION        :\n");
		break;
	case NUMBER:
		printf("NUMBER          : %d\n", tok->intnum);
		break;
	case CHARACTER:
		printf("CHARACTER       : %c\n", tok->ch);
		break;
	case STRING:
		printf("STRING          : %s\n", tok->string);
		break;
	case STAREQ:
		printf("STAREQ          :\n");
		break;
	case ANDEQ:
		printf("ANDEQ           :\n");
		break;
	case OREQ:
		printf("OREQ            :\n");
		break;
	case PLUSEQ:
		printf("PLUSEQ          :\n");
		break;
	case MINUSEQ:
		printf("MINUSEQ         :\n");
		break;
	case SLASHEQ:
		printf("SLASHEQ         :\n");
		break;
	case PERCENTEQ:
		printf("PERCENTEQ       :\n");
		break;
	case LSHIFTEQ:
		printf("LSHIFTEQ        :\n");
		break;
	case RSHIFTEQ:
		printf("RSHIFTEQ        :\n");
		break;
	case UPARROWEQ:
		printf("UPARROWEQ       :\n");
		break;
	case DECIMAL:
		printf("DECIMAL         : %f\n", tok->decnum);
		break;
	case ARROW:
		printf("ARROW           :\n");
		break;
	case DOT:
		printf("DOT             :\n");
		break;
	case EOF:
		printf("[EOF]\n\n");
		break;
	case STRUCT:
		printf("STRUCT          :\n");
		break;
	case STATIC:
		printf("STATIC          :\n");
		break;
	case UNION:
		printf("UNION           :\n");
		break;
	case DEFINE:
		printf("DEFINE          :\n");
		break;
	case IFDEF:
		printf("IFDEF           :\n");
		break;
	case IFNDEF:
		printf("IFNDEF          :\n");
		break;
	case INCLUDE:
		printf("INCLUDE         :\n");
		break;
	case ENDIF:
		printf("ENDIF           :\n");
		break;
	case PCALL:
		printf("PCALL           :\n");
		break;
	case PFRME:
		printf("PFRME           :\n");
		break;
	case PARG:
		printf("PARG            :\n");
		break;
	default:
		printf("unknown token   : tok->type = %d\n", tok->type);
		return(-1);
	}
	return(0);
}

/* int char ʤɤȤ߹߷̾Τ֤ޤ */
static char *get_type_name(int ttype)
{
	char *ch;

	switch (ttype) {
	case TYPE_CHAR:
		ch = "char";
		break;
	case TYPE_UCHAR:
		ch = "unsigned char";
		break;
	case TYPE_DOUBLE:
		ch = "double";
		break;
	case TYPE_FLOAT:
		ch = "float";
		break;
	case TYPE_INT:
		ch = "int";
		break;
	case TYPE_LONG:
		ch = "long";
		break;
	case TYPE_SHORT:
		ch = "short";
		break;
	case TYPE_UINT:
		ch = "unsigned int";
		break;
	case TYPE_ULONG:
		ch = "unsigned long";
		break;
	case TYPE_USHORT:
		ch = "unsigned short";
		break;
	case TYPE_VOID:
		ch = "void";
		break;
	default:
		ch = "unknown type";
		break;
	}
	return(ch);
}
#endif
