/*
 *  main.c - The main processing
 *
 *  Copyright (C) 2003 Seiji Tamura
 */
#ifdef __WIN32__
#include <windows.h>
#else /* __WIN32__ */
#include <unistd.h>
#endif /* __WIN32__ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "common.h"

#define PATH_LEN	4096

#define OPT_DEBUG	0x0001
#define OPT_TRACE	0x0002
#define OPT_NOGO	0x0004
#define OPT_ARGS	0x0008
#define OPT_VERSION	0x0010

extern int fargc;
extern char **fargv;
extern int tokenline;

/* global variables */
int errcnt = 0;
char *filename = "(stdin)";

/* local variables */
static FILE *source;
static char *cmdname;

/* local functions */
static int get_option(char *s);
static int parse_arg(int *argc, char *argv[], int *filecnt);
static char *exe_path(char *name, char *path, int len, char **fname);
static char *real_path(char *name, char *path, int len, char **fname);
static char *get_include_path(includeinf_t *incinf);

/* get one char from stream */
int charget()
{
	int c = getc(source);
	if (c == '\n')
		tokenline++;
	return(c);
}

/* getting options */
static int get_option(char *s)
{
	int c;
	int option;

	option = 0;
	while (c = *s++, c) {
		switch (c) {
		case 'd':
			option |= OPT_DEBUG;
			break;
		case 't':
			option |= OPT_TRACE;
			break;
		case 'n':
			option |= OPT_NOGO;
			break;
		case 'a':
			option |= OPT_ARGS;
			break;
		case 'v':
			option |= OPT_VERSION;
			break;
		default:
			printf("unknown option (%c)\n", c);
			break;
		}
	}
	return(option);
}

/* parsing arguments */
static int parse_arg(int *argc, char *argv[], int *filecnt)
{
	int i;
	int args;
	int option;
	int flag;
	char *s;

	*filecnt = 0;
	args = 1;
	flag = 1;
	option = 0;
	for (i = 1; i < *argc; i++) {
		s = argv[i];
		if ((s[0] == '-') && (flag)) {
			option |= get_option(&s[1]);
			if (option & OPT_ARGS)
				flag = 0;
		}
		else {
			if (flag)
				(*filecnt)++;
			argv[args++] = s;
		}
	}
	*argc = args;
	return(option);
}

static char *exe_path(char *name, char *path, int len, char **fname)
{
	int pos;

#ifdef __WIN32__
	pos = GetModuleFileName(NULL, path, len);
#else /* __WIN32__ */
	if (*name == '/') {
		copystr(path, name);
	}
	else if (chrstr(name, '/') || chrstr(name, '\\')) {
		real_path(name, path, len, fname);
	}
	else {
		char *env, *sp, *ep;
		int sepa;
		char buf[PATH_LEN];
		int exists = 0;
		env = getenv("PATH");
		sepa = (chrstr(env, ';'))?(';'):(':');
		sp = env;
		do {
			struct stat stbuf;
			ep = chrstr(sp, sepa);
			if (ep) {
				copymem(buf, sp, ep - sp);
				buf[ep - sp] = '\0';
			}
			else {
				copystr(buf, sp);
			}
			catstr(buf, "/");
			catstr(buf, name);
			real_path(buf, path, len, fname);
			exists = (stat(path,&stbuf)==(-1))?(0):(1);
			if (exists)
				break;
			sp = ep + 1;
		} while (ep);
		if (!exists)
			copystr(path, name);
#ifdef DEBUG
		printf("%s\n", path);
#endif
	}
	pos = 0;
	while (path[pos] != '\0')
		pos++;
#endif /* __WIN32__ */
	for (pos--; pos >= 0; pos--)
		if (path[pos] == '\\' || path[pos] == '/')
			break;
	*fname = &path[(pos>0)?(pos+1):(0)];
	return(path);
}

static char *real_path(char *name, char *path, int len, char **fname)
{
	char *ret;
	int pos;

#ifdef __WIN32__
	GetFullPathName(name, len, path, fname);
	ret = path;
#else /* __WIN32__ */
	if (*name == '/') {
		copystr(path, name);
	}
	else if (chrstr(name, ':')) {		/* for windows drive name */
		copystr(path, name);
	}
	else {
		char cdir[PATH_LEN];
		if (!getcwd(cdir, sizeof(cdir))) {
			copystr(path, name);
		}
		else {
			copystr(path, cdir);
			catstr(path, "/");
			catstr(path, name);
		}
	}
	ret = path;
#endif /* __WIN32__ */
	pos = 0;
	while (path[pos] != '\0')
		pos++;
	for (pos--; pos >= 0; pos--)
		if (path[pos] == '\\' || path[pos] == '/')
			break;
	*fname = &path[(pos>0)?(pos+1):(0)];
	return(ret);
}

static char *get_include_path(includeinf_t *incinf)
{
	static char path[PATH_LEN];
	char *fname;

	copystr(path, "");
	if (incinf->type == INCLUDE_SYS) {
		exe_path(cmdname, path, sizeof(path), &fname);
		copystr(fname, "../include/minic/");
		catstr(path, incinf->name);
#ifdef DEBUG
		printf("%s\n", path);
#endif
	}
	else {	/* (incinf->type == INCLUDE_USR) */
		real_path(filename, path, sizeof(path), &fname);
		copystr(fname, incinf->name);
#ifdef DEBUG
		printf("%s\n", path);
#endif
	}
	return(path);
}

/* 󥯥롼ɥեν */
int include_main(includeinf_t *incinf)
{
	char *fname;
	FILE *tmpsource = source;

	fname = get_include_path(incinf);
	if ((source = fopen(fname, "r")) != NULL) {
		int tmpline = tokenline;
		char *tmpfilename = filename;
		tokenline = 1;
		filename = incinf->name;
		parse_include();
		fclose(source);
		tokenline = tmpline;
		filename = tmpfilename;
	}
	else {
		printf("include file %s not found\n", fname);
		fclose(tmpsource);
		exit(1);
	}
	source = tmpsource;
	return(0);
}

/* main function */
int main(int argc, char *argv[])
{
	int option;
	int filecnt = 0;
	int i;
	int trace;
	int size = 0;
	intrcode_t *code = NULL;

	cmdname = argv[0];
	option = parse_arg(&argc, argv, &filecnt);

	if (option & OPT_VERSION) {
		printf("%s\n", MINIC_VERSION);
		return(0);
	}

	fargc = argc - filecnt;
	fargv = &argv[filecnt];

	/* Ƽơ֥ */
	init_table(1);

//	filecnt = argc - 1;
	if (filecnt == 0) {
		source = stdin;
		parse(1, 1);
	}
	else {
		for (i = 0; i < filecnt; i++) {
			filename = argv[i + 1];
			if ((source = fopen(filename, "r")) != NULL) {
				parse(i + 1, filecnt);
				fclose(source);
			}
			else {
				printf("%s: open failed\n", filename);
				exit(1);
			}
		}
	}

	if (option & OPT_DEBUG) {
//		printcode();
		print_all_code();
	}
	if (errcnt) {
		exit(1);
	}
	if (!(option & OPT_NOGO)) {
		trace = (option & OPT_TRACE) ? (1) : (0);
		if (code == NULL) {
			code = get_code();
			size = get_code_size();
		}
		interpret(code, size, trace);
	}
	return(0);
}
