/*
 *  parse.c - Parsing and semantic analysis
 *
 *  Copyright (C) 2003 Seiji Tamura
 */
#include <stdio.h>
#include "common.h"

#define GetToken() tok=gettoken()

extern token_t *gettoken();
extern void parse_error(char *s);
extern int tokenline;
extern char *garea;

/* global variables */
token_t *tok;

/* global functions */
void parse(int no, int total);
void chktoken(int symbol, char *message);

/* local variables */
/* none */

typedef struct type_list {
	int type;
	struct type_list *next;
}	type_list_t;
#define TYPE_LISTS_SIZE 64
type_list_t type_lists[TYPE_LISTS_SIZE];

/* local functions */
static void declare(int mode);
static void var_decl(int mode);
static int set_array_value(char **pos, type_list_t **list);
static void resize_array(int mode, varinf_t *pvinf, int size);
static int typedef_decl(int mode, varinf_t *pvinf);
static int declarator(int mode, varinf_t *pvinf);
static void struct_decl(int mode, int chksemi);
static int set_struct_value(int mode, char **ppos, varinf_t *pvinf);
static void func_decl(int mode, varinf_t *pvinf);
static int parse_arg(int mode, int *varargs);
static void decl_list(int mode);
static void statement_list(int mode, int brk, int cnt);
static void statement(int mode, int brk, int cnt);
static void resist_intrinsic_constant();
static void resist_intrinsic_function();

#ifdef DEBUG
extern void print_global_struct();
extern void print_global_variable();
extern void print_all_function();
//static void debug_skip(int type1, int eq, int type2);
#endif

/* all starts with this function */
void parse(int no, int total)
{
	int pc;

	tokenline = 1;
	/* ߤΥץ५󥿤 */
	pc = get_code_size();
	/* ĤΥեξ */
	if (no == 1) {
		/* ֥ɤνԤ */
//		initcode();
		/* Ȥ߹ϿԤ */
		resist_intrinsic_constant();
		/* Ȥ߹ߴؿϿԤ */
		resist_intrinsic_function();
		/* ȥݥȤΥɽ */
		startup_code();
	}
	/* ʸǼѿʤɤ */
	initchar();
	/* ե륹פΥơ֥ */
	init_file_scope_table();
	/* ȡɤ߹(Ƚ˾ɤ) */
	GetToken();
	while (tok->type != EOF) {
		/* ؿޤѿ */
		declare(MODE_GLOBAL);
	}
	/* ǸΥեϡؿΥ󥯤Ԥ */
	if (no == total) {
		checkparam(0);
		checkfunc();
	}
	else {
		checkparam(pc);
	}
#ifdef DEBUG
	printf("----- print_global_struct -----\n");
	print_global_struct();
	printf("----- print_global_variable -----\n");
	print_global_variable();
	printf("----- print_all_function -----\n");
	print_all_function();
#endif
}

void parse_include()
{
	/* ʸǼѿʤɤ */
	initchar();
	/* ȡɤ߹(Ƚ˾ɤ) */
	GetToken();
	while (tok->type != EOF) {
		/* ؿޤѿ */
		declare(MODE_GLOBAL);
	}
	/* ʸǼѿʤɤ */
	initchar();
}

static void declare(int mode)
{
	varinf_t vinf;

	/* staticꤵƤ */
	if (tok->type == STATIC) {
		mode |= MODE_STATIC;
		GetToken();
	}
	/* externꤵƤ */
	else if (tok->type == EXTERN) {
		mode |= MODE_EXTERN;
		GetToken();
	}
	/* typedefꤵƤ */
	else if (tok->type == TYPEDEF) {
		mode |= MODE_TYPEDEF;
		GetToken();
	}

	if (tok->type == TYPENAME) {
		var_decl(mode);
	}
	else if (tok->type == STRUCT) {
		struct_decl(mode, 1);
	}
	else if (tok->type == UNION) {
		/*union_decl();*/
		struct_decl(mode, 1);
	}
	else if (tok->type == IDENT) {
		funcinf_t *finf;
		varinf_t *pvinf;
		char *pos;
		int i;

		/* ͤλ꤬ʤintȤư */
		vinf.scope = mode;
		vinf.type = TYPENAME;
		vinf.ttype = TYPE_INT;
		vinf.size = sizeof(int);
		vinf.base = 0;
		vinf.offset = 0;
		vinf.ptflag = 0;
		vinf.argflag = 0;
		vinf.name = regist_name(mode, tok->ident);
		i = 0;
		while (i < ARR_MAX) {
			vinf.array[i++] = -1;
		}
		vinf.next = NULL;
		func_decl(mode, &vinf);

		finf = find_function(0, vinf.name);
		pvinf = find_var(MODE_LOCAL, vinf.name);
		pos = garea + pvinf->offset;
		(*(int *)pos) = finf->ic;
	}
	else if (tok->type == DEFINE) {
#ifdef DEBUG
		macroinf_t *macro;
		macro = tok->macro;
		switch (macro->type) {
		case NUMBER:
			printf("%s %d (NUMBER)\n",macro->name,macro->intnum);
			break;
		case DECIMAL:
			printf("%s %f (DECIMAL)\n",macro->name,macro->decnum);
			break;
		case CHARACTER:
			printf("%s %c (CHARACTER)\n",macro->name,macro->ch);
			break;
		case STRING:
			printf("%s %s (STRING)\n",macro->name,macro->string);
			break;
		default:
			printf("define error\n");
			break;
		}
#endif
		regist_macro(tok->macro);
		GetToken();
	}
	else if (tok->type == INCLUDE) {
#ifdef DEBUG
		includeinf_t *incinf;
		incinf = tok->incinf;
		switch (incinf->type) {
		case INCLUDE_SYS:
			printf("#include <%s>\n", incinf->name);
			break;
		case INCLUDE_USR:
			printf("#include \"%s\"\n", incinf->name);
			break;
		default:
			printf("include error\n");
			break;
		}
#endif
		include_main(tok->incinf);
		GetToken();
	}
	else {
		parse_error("unrecognized external definition");
	}
}

static token_t var_tok;
static void var_decl(int mode)
{
//	token_t var_tok;
	varinf_t vinf;
	int tmpptflag = tok->ptflag;

	/* ѿη */
//	var_tok = *tok;

	vinf.scope = mode;
	vinf.type = tok->type;
	vinf.ttype = tok->ttype;
	do {
		int i;
		vinf.size = 0;
		vinf.base = 0;
		vinf.offset = 0;
		vinf.stuni = 0;
//		vinf.ptflag = 0;
		vinf.ptflag = tmpptflag;
		vinf.argflag = 0;
//		for (i = 0; i < ARR_MAX; i++) {
//			vinf.array[i] = -1;
//		}
		i = 0;
		while (i < ARR_MAX) {
			vinf.array[i++] = -1;
		}
		vinf.funcptr = 0;
		vinf.next = NULL;
		GetToken();
		if (mode & MODE_TYPEDEF) {
			typedefinf_t tdefinf;
			typedef_decl(mode, &vinf);
			copystr(tdefinf.name, vinf.name);
			tdefinf.type = vinf.type;
			tdefinf.ttype = vinf.ttype;
			tdefinf.ptflag = vinf.ptflag;
			regist_typedef(&tdefinf);
		}
		else {
			if (declarator(mode, &vinf)) {
				return;
			}
			regist_var(mode, &vinf);
		}
		if (tok->type == EQUAL) {
			if (mode & MODE_TYPEDEF) {
				parse_error("invalid typedef decl");
			}
			else if (mode == MODE_LOCAL) {
				add_token_que(&var_tok);
				add_token_que(tok);
				GetToken();
				expression(mode | MODE_DECLARE);
				popup();
			}
			else {
				varinf_t *vinf;
				char *pos;
				int size;
				type_list_t *list;
				vinf = find_var(MODE_LOCAL, var_tok.ident);
				pos = garea + vinf->offset;
				GetToken();
				switch (tok->type) {
				case CHARACTER:
					(*(char *)pos) = (char)tok->ch;
					GetToken();
					break;
				case STRING:
					(*(char **)pos) = tok->string;
					GetToken();
					break;
				case DECIMAL:
					(*(double *)pos) = tok->decnum;
					GetToken();
					break;
				case LBRACE:
					list = type_lists;
					if (vinf->ptflag)
						list->type = TYPE_VOID;
					else
						list->type = vinf->ttype;
					list->next = list;
					GetToken();
					size = set_array_value(&pos, &list);
					resize_array(mode, vinf, size);
					chktoken(RBRACE, "right brace");
					break;
				default:
					(*(int *)pos) = eval_expr_int();
					break;
				}
			}
		}
	} while (tok->type == COMMA);
	chktoken(SEMICOLON, "semicolon");
}

static int set_array_value(char **ppos, type_list_t **list)
{
	int size = 0;
	do {
		switch (tok->type) {
		case CHARACTER:
			switch ((*list)->type) {
			case TYPE_CHAR:
				(*(char *)(*ppos)) = (char)tok->ch;
				(*ppos) += sizeof(char);
				break;
//			case STRING:
//				(*(char **)(*ppos)) = (char *)tok->ch;
//				(*ppos) += sizeof(char *);
//				break;
			case TYPE_DOUBLE:
			case TYPE_FLOAT:
				(*(double *)(*ppos)) = (double)tok->ch;
				(*ppos) += sizeof(double);
				break;
			default:
				(*(int *)(*ppos)) = (int)tok->ch;
				(*ppos) += sizeof(int);
				break;
			}
			GetToken();
			size++;
			(*list) = (*list)->next;
			break;
		case STRING:
			switch ((*list)->type) {
			case TYPE_CHAR:
				yyerror("different size cast");
//				(*(char *)(*ppos)) = (char)tok->string;
				(*ppos) += sizeof(char);
				break;
//			case STRING:
//				(*(char **)(*ppos)) = (char *)tok->string;
//				(*ppos) += sizeof(char *);
//				break;
			case TYPE_DOUBLE:
			case TYPE_FLOAT:
				yyerror("cannot convert to a pointer type");
//				(*(double *)(*ppos)) = (double)tok->string;
				(*ppos) += sizeof(double);
				break;
			default:
//				(*(int *)(*ppos)) = (int)tok->string;
				(*((char **)(*ppos))) = (char *)tok->string;
//				(*ppos) += sizeof(int);
				(*ppos) += sizeof(char *);
				break;
			}
			GetToken();
			size++;
			(*list) = (*list)->next;
			break;
		case DECIMAL:
			switch ((*list)->type) {
			case TYPE_CHAR:
				(*(char *)(*ppos)) = (char)tok->decnum;
				(*ppos) += sizeof(char);
				break;
//			case STRING:
//				yyerror("cannot convert to a pointer type");
////				(*(char **)(*ppos)) = (char *)tok->decnum;
//				(*ppos) += sizeof(char *);
//				break;
			case TYPE_DOUBLE:
			case TYPE_FLOAT:
				(*(double *)(*ppos)) = (double)tok->decnum;
				(*ppos) += sizeof(double);
				break;
			default:
				(*(int *)(*ppos)) = (int)tok->decnum;
				(*ppos) += sizeof(int);
				break;
			}
			GetToken();
			size++;
			(*list) = (*list)->next;
			break;
		case LBRACE:
			GetToken();
			size += set_array_value(ppos, list);
			chktoken(RBRACE, "right brace");
			break;
		case RBRACE:
			return(size);
		default:
			switch ((*list)->type) {
			case TYPE_CHAR:
				(*(char *)(*ppos)) = (char)eval_expr_int();
				(*ppos) += sizeof(char);
				break;
//			case STRING:
//				(*(char **)(*ppos)) = (char *)eval_expr_int();
//				(*ppos) += sizeof(char *);
//				break;
			case TYPE_DOUBLE:
			case TYPE_FLOAT:
				(*(double *)(*ppos)) = (double)eval_expr_int();
				(*ppos) += sizeof(double);
				break;
			default:
				(*(int *)(*ppos)) = (int)eval_expr_int();
				(*ppos) += sizeof(int);
				break;
			}
			/* eval_expr_int()ƤΤǡ */
			/* ǤGetToken()ɬפ̵      */
//			GetToken();
			size++;
			(*list) = (*list)->next;
			break;
		}
		if (tok->type != COMMA)
			break;
		GetToken();
	} while (tok->type != EOF);

	return(size);
}

/* δؿset_array_value()Ƥľ˸Ƥ֤ */
static void resize_array(int mode, varinf_t *pvinf, int size)
{
	int i, cur_size, true_size;

	i = 0;
	cur_size = 1;
	while (i < ARR_MAX) {
		if (pvinf->array[i] > 0)
			cur_size *= pvinf->array[i];
		i++;
	}
	true_size = size / cur_size;
	if (size % cur_size)
		true_size++;
	if (pvinf->array[0] <= 0)
		pvinf->array[0] = true_size;
	else if (pvinf->array[0] < true_size)
		yyerror("excess elements in array or struct initializer");
	correct_var_offset(mode, pvinf);
}

/* typedef */
static int typedef_decl(int mode, varinf_t *pvinf)
{
	switch (tok->type) {
	case IDENT:
		var_tok = *tok;
		pvinf->name = var_tok.ident;
		GetToken();
		break;
	case LPAR:
		GetToken();
		typedef_decl(mode, pvinf);
		chktoken(RPAR, "right parenthesis");
		break;
	case STAR:
		pvinf->ptflag++;
		GetToken();
		typedef_decl(mode, pvinf);
		break;
	default:
		parse_error("invalid typedef_decl");
		break;
	}
	return(0);
}

/* ѿ */
static int declarator(int mode, varinf_t *pvinf)
{
	int dim = 0;
	int arr_size = -1;

	switch (tok->type) {
	case IDENT:
		var_tok = *tok;
		pvinf->name = regist_name(mode, tok->ident);
		GetToken();
		if (tok->type == LPAR) {
			funcinf_t *finf;
			varinf_t *vinf;
			char *pos;

			/* ؿ̾ѿ϶ŪintȤ */
			/* Ĥޤͤηå˻Ȥ櫓ǤϤʤ */
			pvinf->type = TYPENAME;
			pvinf->ttype = TYPE_INT;
			func_decl(mode, pvinf);

			finf = find_function(0, pvinf->name);
			vinf = find_var(MODE_LOCAL, pvinf->name);
			pos = garea + vinf->offset;
			/* 32ӥåȤǤ뤳ȤʤäƤޤ...   */
			/* 16ӥåȤic15ӥåȤե졼ॵ  */
			/* ؿݥ󥿤ȤΥѿΰ32767 */
			/* ХȤ¤뤳Ȥˤʤ롣                    */
			(*(int *)pos)=finf->ic|((finf->fsize<<16)&0x7fff0000);
			return(1);
		}
		break;
	case LPAR:
		GetToken();
		declarator(mode, pvinf);
		chktoken(RPAR, "right parenthesis");
		/* LPAR줿ϴؿݥ */
		if (tok->type == LPAR) {
			pvinf->funcptr = 1;
			GetToken();
			while (tok->type != RPAR) {
				/* ȤꤢɤФǲ⤷ʤ */
//				if (tok->type!=TYPENAME&&tok->type!=TRIPLEDOT)
//					parse_error("typename expected");
				GetToken();
//				while (tok->type == STAR) {
//					GetToken();
//				}
//				if (tok->type != COMMA) {
//					break;
//				}
//				GetToken();
//				if (tok->type == RPAR) {
//					char *e = "error before \')\' token";
//					parse_error(e);
//				}
			}
			chktoken(RPAR, "right parenthesis");
		}
		break;
	case STAR:
		pvinf->ptflag++;
		GetToken();
		if (declarator(mode, pvinf)) {
			return(1);
		}
		break;
	default:
		parse_error("invalid declarator");
		break;
	}
	while (tok->type == LBRACK) {
		GetToken();
		if (tok->type == RBRACK) {
			char *err = "array type has incomplete element type";
			/* ϥݥ󥿤ȸ */
			if (mode & MODE_PARAM) {
				if (dim == 0) {
					//pvinf->array[dim++] = 0;
					//pvinf->ptflag++;
					arr_size = 0;
					GetToken();
					//continue;
				}
				else {
					err = "unknown array element size";
					parse_error(err);
				}
			}
			/* ͤꤵƤʤϤȤꤢ */
			else {
				if (dim > 0)
					yyerror(err);
				arr_size = 0;
				GetToken();
			}
//			GetToken();
//			continue;
		}
		else {
			arr_size = eval_expr_int();
			if (arr_size < 0)
				parse_error("size of array is negative");
			chktoken(RBRACK, "right bracket");
// ʲΥɤ򥳥Ȳ(20071219)
// func(char arr[3][4])Τ褦˰Ĺξ
// ǤʤȤꤢ礦ʤ...
//			if (mode & MODE_PARAM) {
//				pvinf->ptflag++;
//				continue;
//			}
		}
		if (dim < ARR_MAX) {
			pvinf->array[dim++] = arr_size;
		}
		else {
			parse_error("array dimension out of bounds");
		}
		if (mode & MODE_PARAM) {
			pvinf->argflag = 1;
			//printf("pvinf->argflag %s %d\n", pvinf->name, pvinf->argflag);
		}
		else {
			pvinf->argflag = 0;
			//printf("pvinf->argflag %s %d\n", pvinf->name, pvinf->argflag);
		}
	}
	return(0);
}

static void struct_decl(int mode, int chksemi)
{
	stuninf_t *pstuni;
	token_t struct_tok;
	varinf_t vinf;
	int stuni_type;
	int typedefflag = 0;
	int tmpptflag = tok->ptflag;

	if (mode & MODE_TYPEDEF) {
		mode &= ~MODE_TYPEDEF;
		typedefflag = 1;
	}

	/* (¤Τޤ϶) */
	stuni_type = tok->type;

	GetToken();
	if (tok->type != IDENT) {
		parse_error("ident expected");
		return;
	}
	/* ¤Υ̾ξ */
	struct_tok = *tok;

	GetToken();
	if (tok->type == LBRACE) {
		/* ¤ΤξϿ */
		pstuni = regist_struct(mode, &struct_tok);
		GetToken();
		while (tok->type != RBRACE) {
			if (tok->type == TYPENAME) {
				var_decl(mode);
			}
			else if (tok->type == STRUCT) {
				struct_decl(mode, 1);
			}
			else if (tok->type == UNION) {
				/*union_decl();*/
				struct_decl(mode, 1);
			}
//			else if (tok->type == IDENT) {	/* typedef ? */
//				func_decl(mode);
//				parse_error("unexpected token");
//			}
			else {
				parse_error("unrecognized struct definition");
			}
		}
		chktoken(RBRACE, "right brace");
		end_regist_struct(mode, pstuni);
		if (typedefflag) {
			int camma = 0;
			do {
				typedefinf_t tdefinf;
				if (camma)
					GetToken();
				vinf.ptflag = 0;
				vinf.argflag = 0;
				typedef_decl(mode, &vinf);
				copystr(tdefinf.name, vinf.name);
				tdefinf.type = stuni_type;
				tdefinf.tname = pstuni->name;
				tdefinf.ptflag = vinf.ptflag;
				regist_typedef(&tdefinf);
				if (tok->type == COMMA)
					camma = 1;
			} while (tok->type == COMMA);
		}
		chktoken(SEMICOLON, "semicolon");
	}
	else {
		vinf.scope = mode;
		vinf.type = stuni_type;
		pstuni = find_struct(mode, struct_tok.ident);
		if (pstuni == NULL) {
			parse_error("struct not registered\n");
		}
		else {
			vinf.tname = pstuni->name;
			vinf.stuni = pstuni;
		}
		while (1) {
			int i;
			vinf.size = 0;
			vinf.base = 0;
			vinf.offset = 0;
//			vinf.ptflag = 0;
			vinf.ptflag = tmpptflag;
			vinf.argflag = 0;
			i = 0;
			while (i < ARR_MAX) {
				vinf.array[i++] = -1;
			}
			vinf.funcptr = 0;
			vinf.next = NULL;
			if (typedefflag) {
				int camma = 0;
				do {
					typedefinf_t tdefinf;
					if (camma)
						GetToken();
					vinf.ptflag = 0;
					vinf.argflag = 0;
					typedef_decl(mode, &vinf);
					copystr(tdefinf.name, vinf.name);
					tdefinf.type = stuni_type;
					tdefinf.tname = pstuni->name;
					tdefinf.ptflag = vinf.ptflag;
					regist_typedef(&tdefinf);
					if (tok->type == COMMA)
						camma = 1;
				} while (tok->type == COMMA);
				break;
//				typedefinf_t tdefinf;
//				typedef_decl(mode, &vinf);
//				copystr(tdefinf.name, vinf.name);
//				tdefinf.type = stuni_type;
//				tdefinf.tname = pstuni->name;
//				tdefinf.ptflag = vinf.ptflag;
//				regist_typedef(&tdefinf);
			}
			else {
				if (declarator(mode, &vinf)) {
					return;
				}
				regist_var(mode, &vinf);
			}
			if (tok->type == EQUAL) {
				if (typedefflag) {
					parse_error("invalid typedef decl");
				}
				else if (mode == MODE_LOCAL) {
					add_token_que(&var_tok);
					add_token_que(tok);
					GetToken();
					expression(0);
					popup();
				}
				else {
					varinf_t *vinf;
					char *pos;
					vinf = find_var(mode, var_tok.ident);
					pos = garea + vinf->offset;
					GetToken();
					chktoken(LBRACE, "left brace");
					set_struct_value(mode, &pos, vinf);
					chktoken(RBRACE, "right brace");
				}
			}
			/* ξʣ̵ */
			if (tok->type != COMMA || chksemi == 0) {
				break;
			}
			GetToken();
		}
		if (chksemi)
			chktoken(SEMICOLON, "semicolon");
		/* ξϥߥåϹԤʤ */
	}
}

static int set_struct_value(int mode, char **ppos, varinf_t *pvinf)
{
	int size;
	type_list_t *list, *tmplist;
	varinf_t *tmpvinf;

	list = type_lists;
	tmplist = type_lists;
	tmpvinf = pvinf->stuni->var;
	while (tmpvinf) {
		if (tmpvinf->ptflag)
			list->type = TYPE_VOID;
		else
			list->type = tmpvinf->ttype;
		tmplist = list;
		list++;
		tmplist->next = list;
		tmpvinf = tmpvinf->next;
	}
	tmplist->next = type_lists;
	list = type_lists;
	size = set_array_value(ppos, &list);
	if (pvinf->array[0] >= 0) {
		int st_size, true_size;
		st_size = 0;
		tmpvinf = pvinf->stuni->var;
		while (tmpvinf) {
			st_size++;
			tmpvinf = tmpvinf->next;
		}
		if (st_size > 0) {
			true_size = size / st_size;
			if (size % st_size)
				true_size++;
			resize_array(mode, pvinf, true_size);
		}
	}
	return(size);
}

static void func_decl(int mode, varinf_t *retinf)
{
	int allow_kr = 0;
	int varargs = 0;
	token_t func_tok;
	funcinf_t *funcinf;

	/* ѿη */
	func_tok = *tok;

	/* 湽ʸѤΥ٥ */
	init_label();
	/* ̾դ٥ */
	init_named_label();
	/* ؿϿ */
	funcinf = regist_function(mode, retinf, 1);
	if (tok->type != LPAR) {
		GetToken();
	}
	chktoken(LPAR, "left parenthesis");
	if (tok->type == RPAR) {
		GetToken();
	}
	else {
		/* Ϥ */
		allow_kr = parse_arg(MODE_LOCAL | MODE_PARAM, &varargs);
		chktoken(RPAR, "right parenthesis");
	}
	/* K & R η(?) */
	if (allow_kr) {
		decl_list(MODE_LOCAL | MODE_PARAM);
	}
	if (tok->type == SEMICOLON) {
		varinf_t *tmpvinf;

		GetToken();
//		end_regist_function(mode, funcinf, 1);
		tmpvinf = find_var(mode, funcinf->name);
		if (tmpvinf && !funcinf->confirmed)
			tmpvinf->confirmed = 0;
//		funcinf->confirmed = 0;
		// FENTFRET
//		pop_code(2);
		if (!funcinf->confirmed)
			pop_code(1);
		end_prototype();
		return;
	}
	end_function_param(mode, funcinf);
	if (varargs)
		funcinf->argcnt = -1;
	funcinf->confirmed = 1;
	chktoken(LBRACE, "left brace");
#ifdef DEBUG
//	printf("\n--- function \"%s\" start ---\n", func_tok.ident);
#endif
	/* ѿ */
	decl_list(MODE_LOCAL);
	/* ʸ */
	statement_list(MODE_LOCAL, -1, -1);
#ifdef DEBUG
//	printf("--- function \"%s\" end ---\n\n", func_tok.ident);
#endif
	chktoken(RBRACE, "right brace");
	checklabel(funcinf->ic);
	end_regist_function(mode, funcinf, 1);
}

static int parse_arg(int mode, int *varargs)
{
	int allow_kr = 1;
	varinf_t vinf;

#ifdef DEBUG
//	printf("\n--- parse_arg start ---\n");
#endif
	*varargs = 0;
	while (1) {
		if (tok->type == TYPENAME) {
			int i;
			vinf.scope = mode;
			vinf.type = tok->type;
			vinf.ttype = tok->ttype;
			vinf.size = 0;
			vinf.base = 0;
			vinf.offset = 0;
			vinf.ptflag = 0;
			vinf.argflag = 0;
			i = 0;
			while (i < ARR_MAX) {
				vinf.array[i++] = -1;
			}
			vinf.funcptr = 0;
			vinf.next = NULL;
			GetToken();
			if (vinf.ttype == TYPE_VOID && tok->type == RPAR)
				break;
			declarator(mode, &vinf);
			regist_var(mode, &vinf);
			/* η硢K&RĤʤ */
			allow_kr = 0;
		}
		else if (tok->type == STRUCT) {
			struct_decl(mode, 0);
			/* η硢K&RĤʤ */
			allow_kr = 0;
		}
		else if (tok->type == UNION) {
			struct_decl(mode, 0);
			/* η硢K&RĤʤ */
			allow_kr = 0;
		}
		else if (tok->type == IDENT) {
			int i;
			/* ʤϤȤꤢintȸʤ */
			vinf.scope = mode;
			vinf.type = TYPENAME;
			vinf.ttype = TYPE_INT;
			vinf.size = 0;
			vinf.base = 0;
			vinf.offset = 0;
			vinf.ptflag = 0;
			vinf.argflag = 0;
			i = 0;
			while (i < ARR_MAX) {
				vinf.array[i++] = -1;
			}
			vinf.funcptr = 0;
			vinf.next = NULL;
			declarator(mode, &vinf);
			regist_var(mode | MODE_EXTERN, &vinf);
		}
		else if (tok->type == TRIPLEDOT) {
			*varargs = 1;
			GetToken();
			break;
		}
		else {
			parse_error("formal parameter is not identifier");
		}
		if (tok->type != COMMA) {
			break;
		}
		GetToken();
	}
#ifdef DEBUG
//	printf("--- parse_arg end ---\n");
#endif
	return(allow_kr);
}

static void decl_list(int mode)
{
	while (1) {
		switch (tok->type) {
		case STATIC:
		case TYPENAME:
		case STRUCT:
		case UNION:
		case EXTERN:
		case DEFINE:
		case INCLUDE:
			declare(mode);
			break;
		default:
			return;
		}
	}
}

static void statement_list(int mode, int brk, int cnt)
{
	while (tok->type != RBRACE && tok->type != EOF) {
		statement(mode, brk, cnt);
	}
}

static void statement(int mode, int brk, int cnt)
{
	int L1, L2, L3, L4;
	switchlblinf_t *swlblinf;
	int def_ic = 0;

	if (tok->type == SEMICOLON) {
		GetToken();
	}
	else if (tok->type == LBRACE) {
		inc_local_block();
		GetToken();
		decl_list(mode);
		statement_list(mode, brk, cnt);
		dec_local_block();
		chktoken(RBRACE, "right brace");
	}
	else if (tok->type == IF) {
		L1 = label_next();	/* FALSE */
		L2 = label_next();	/* TRUE  */
		GetToken();
		chktoken(LPAR, "left parenthesis");
		if (expression(0))
			add_code(LOAD, 0, 0, 0, 0);
		chktoken(RPAR, "right parenthesis");
		add_code(FJMP, L1, 0, FALSE, 0);
		statement(mode, brk, cnt);
		add_code(JMP, L2, 0, 0, 0);
		regist_label(L1, add_code(LABEL, L1, 0, 0, 0));
		if (tok->type == ELSE) {
			GetToken();
			statement(mode, brk, cnt);
		}
		regist_label(L2, add_code(LABEL, L2, 0, 0, 0));
	}
	else if (tok->type == WHILE) {
		L1 = label_next();	/* break    */
		L2 = label_next();	/* continue */
		GetToken();
		regist_label(L2, add_code(LABEL, L2, 0, 0, 0));
		chktoken(LPAR, "left parenthesis");
		if (expression(0))
			add_code(LOAD, 0, 0, 0, 0);
		chktoken(RPAR, "right parenthesis");
		add_code(FJMP, L1, 0, FALSE, 0);
		statement(mode, L1, L2);
		add_code(JMP, L2, 0, 0, 0);
		regist_label(L1, add_code(LABEL, L1, 0, 0, 0));
	}
	else if (tok->type == DO) {
		L1 = label_next();	/* break    */
		L2 = label_next();	/* continue */
		GetToken();
		regist_label(L2, add_code(LABEL, L2, 0, 0, 0));
		statement(mode, L1, L2);
		chktoken(WHILE, "while");
		chktoken(LPAR, "left parenthesis");
		if (expression(0))
			add_code(LOAD, 0, 0, 0, 0);
		chktoken(RPAR, "right parenthesis");
		chktoken(SEMICOLON, "semicolon");
		add_code(TJMP, L2, 0, FALSE, 0);
		regist_label(L1, add_code(LABEL, L1, 0, 0, 0));
	}
	else if (tok->type == BREAK) {
		add_code(JMP, brk, 0, 0, 0);
		GetToken();
		chktoken(SEMICOLON, "semicolon");
	}
	else if (tok->type == CONTINUE) {
		add_code(JMP, cnt, 0, 0, 0);
		GetToken();
		chktoken(SEMICOLON, "semicolon");
	}
	else if (tok->type == FOR) {
		L1 = label_next();	/* break    */
		L2 = label_next();	/* TRUE     */
		L3 = label_next();	/* continue */
		L4 = label_next();	/* expr     */
		GetToken();
		chktoken(LPAR, "left parenthesis");
		/* ܤμɾȥ */
		if (tok->type == SEMICOLON)
			GetToken();
		else {
			expression(0);
			add_code(POPUP, 0, 0, 0, 0);
			chktoken(SEMICOLON, "semicolon");
		}
		/* 롼׽ƬΥ٥ */
		regist_label(L4, add_code(LABEL, L4, 0, 0, 0));
		/* ܤμɾ */
		if (tok->type == SEMICOLON)
			GetToken();
		else {
			/* ʸȽȥפΥɤ */
			if (expression(0))
				add_code(LOAD, 0, 0, 0, 0);
			chktoken(SEMICOLON, "semicolon");
			add_code(FJMP, L1, 0, FALSE, 0);
		}
		add_code(JMP, L2, 0, 0, 0);
		/* ܤμΥ٥ */
		regist_label(L3, add_code(LABEL, L3, 0, 0, 0));
		/* ܤμɾ */
		if (tok->type == RPAR)
			GetToken();
		else {
			/* 롼׽˼¹Ԥ륳ɤ */
			expression(0);
			add_code(POPUP, 0, 0, 0, 0);
			chktoken(RPAR, "right parenthesis");
		}
		/* 롼Ƭ˥פ */
		add_code(JMP, L4, 0, 0, 0);
		/* for֥åؤΥ٥ */
		regist_label(L2, add_code(LABEL, L2, 0, 0, 0));
		/* for֥å */
		statement(mode, L1, L3);
		/* 롼ν˥פ */
		add_code(JMP, L3, 0, 0, 0);
		/* 롼׽λ֤Υ٥ */
		regist_label(L1, add_code(LABEL, L1, 0, 0, 0));
	}
	else if (tok->type == SWITCH) {
		L1 = label_next();	/* switch eval */
		L2 = label_next();	/* block end   */
		GetToken();
		chktoken(LPAR, "left parenthesis");
		if (expression(0))
			add_code(LOAD, 0, 0, 0, 0);
		/* åΥȥåפͤAX쥸˰ư */
		add_code(MOV, 0, 0, AX, 0);
		chktoken(RPAR, "right parenthesis");
		/* öswitch֥åθȽԤ */
		add_code(JMP, L1, 0, 0, 0);
		/* switch֥å */
		statement(mode, L2, cnt);
		add_code(JMP, L2, 0, 0, 0);
		/* switch-caseɾ򤹤륳ɤؤΥ٥ */
		regist_label(L1, add_code(LABEL, L1, 0, 0, 0));
		while ((swlblinf = pop_sw_label(L2)) != NULL) {
			if (swlblinf->def != 0) {
				def_ic = swlblinf->pc;
				continue;
			}
			/* AX쥸ͤӤ̤򥹥åѤ */
			add_code(CMPR, swlblinf->val, 0, AX, 0);
			/* ̤ʤХ */
			add_code(TJMP, 0, swlblinf->pc, FALSE, 0);
		}
		if (def_ic != 0)
			add_code(JMP, 0, def_ic, 0, 0);
		/* breakˤäswitchγȴݤΥ٥ */
		regist_label(L2, add_code(LABEL, L2, 0, 0, 0));
	}
	else if (tok->type == CASE) {
		int case_val;
		/* caseͤΥɤϿ */
		GetToken();
		case_val = eval_expr_int();
		regist_sw_label(brk, add_code(LABEL, 0, 0, 0, 0), case_val, 0);
		chktoken(COLON, "colon");
	}
	else if (tok->type == DEFAULT) {
		regist_sw_label(brk, add_code(LABEL, 0, 0, 0, 0), 0, 1);
		GetToken();
		chktoken(COLON, "colon");
	}
	else if (tok->type == RETURN) {
		GetToken();
		if (tok->type == SEMICOLON) {
			add_code(FRET, 0, 0, 0, 0);
			GetToken();
		}
		else {
			if (expression(0))
				add_code(LOAD, 0, 0, 0, 0);
			add_code(RET, 0, 0, 0, 0);
			chktoken(SEMICOLON, "semicolon");
		}
	}
	else if (tok->type == GOTO) {
		GetToken();
		add_code(JMP, -1, 0, regist_named_label(tok->ident, -1), 0);
		GetToken();
		chktoken(SEMICOLON, "semicolon");
	}
	else {
		token_t tmp_tok = *tok;

		if (tok->type == IDENT) {
			GetToken();
			if (tok->type == COLON) {
				int tmp_ic = add_code(LABEL,0,0,0,0);
				regist_named_label(tok->ident, tmp_ic);
				GetToken();
			}
			else {
				add_token_que(&tmp_tok);
				add_token_que(tok);
				GetToken();
				expression(0);
				popup();
				chktoken(SEMICOLON, "semicolon");
			}
		}
		else {
			expression(0);
			popup();
			chktoken(SEMICOLON, "semicolon");
		}

//		expression(0);
//		if (tok->type == COLON) {
//			/* ǽäƤϥ٥ʤΤǤΤޤ³ */
//			GetToken();
//		}
//		else {
//			/* ɾΤߤ򤷤ϥå夫ǡ */
//			popup();
//			chktoken(SEMICOLON, "semicolon");
//		}
	}
}

void chktoken(int symbol, char *message)
{
	char *s;
	char *t;
	char buff[256];

	if (tok->type != symbol) {
		s = buff;
		while (*message) {
			*s++ = (*message++);
		}
		t = " expected";
		while (*t) {
			*s++ = (*t++);
		}
		*s = 0;
		parse_error(buff);
	}
	GetToken();
}

#include <time.h>

typedef struct intr_const {
	char *name;
	int value;
}	intr_const_t;

intr_const_t intr_const_tbl[] = {
	{"stdin",	-1	},
	{"stdout",	-2	},
	{"stderr",	-3	},
	{"CLOCKS_PER_SEC",	-4	},
	{"CLK_TCK",	-5	}
};
int intr_const_tbl_cnt = sizeof(intr_const_tbl) / sizeof(intr_const_tbl[0]);

static void resist_intrinsic_constant()
{
	int i, j;
	int mode = MODE_GLOBAL;
	varinf_t vinf, *pvinf;
	char *pos;

	for (i = 0; i < intr_const_tbl_cnt; i++) {
		vinf.scope = mode;
		vinf.type = TYPENAME;
		switch (intr_const_tbl[i].value) {
		case -1:
		case -2:
		case -3:
			vinf.ttype = TYPE_VOID;
			vinf.size = sizeof(void *);
			vinf.ptflag = 1;
			vinf.argflag = 0;
			break;
		case -4:
		case -5:
			vinf.ttype = TYPE_INT;
			vinf.size = sizeof(int);
			vinf.ptflag = 0;
			vinf.argflag = 0;
			break;
		}
		vinf.base = 0;
		vinf.offset = 0;
		vinf.name = regist_name(mode, intr_const_tbl[i].name);
		j = 0;
		while (j < ARR_MAX) {
			vinf.array[j++] = -1;
		}
		vinf.funcptr = 0;
		vinf.next = NULL;
		regist_var(mode, &vinf);
		pvinf = find_var(mode, intr_const_tbl[i].name);
		if (pvinf) {
			pos = garea + pvinf->offset;
//			(*(int *)pos) = intr_const_tbl[i].value;
			switch (intr_const_tbl[i].value) {
			case -1:
				(*(FILE **)pos) = stdin;
				break;
			case -2:
				(*(FILE **)pos) = stdout;
				break;
			case -3:
				(*(FILE **)pos) = stderr;
				break;
			case -4:
				(*((int *)pos)) = CLOCKS_PER_SEC;
				break;
			case -5:
				(*((int *)pos)) = CLOCKS_PER_SEC;
				break;
			default:
				parse_error("fail to resist constant");
				break;
			}
		}
		else {
			parse_error("fail to resist constant");
		}
	}
}

extern intrfunc intrfuncs[];
extern int intrfuncs_cnt;

static void resist_intrinsic_function()
{
	int i, j, k;
	int mode = MODE_GLOBAL;
	int argmode = MODE_LOCAL | MODE_PARAM;
	varinf_t vinf;
	funcinf_t *funcinf;

	for (i = 0; i < intrfuncs_cnt; i++) {
		varinf_t *pvinf;
		char *pos;
		int argcnt = intrfuncs[i].argcnt;

		/* ͤλ꤬ʤintȤư */
		vinf.scope = mode;
		vinf.type = TYPENAME;
		/* ؿ̾ѿ϶ŪintȤ */
		/* Ĥޤͤηå˻Ȥ櫓ǤϤʤ */
		vinf.ttype = TYPE_INT;
		vinf.size = sizeof(int);
		vinf.base = 0;
		vinf.offset = 0;
		vinf.ptflag = 0;
		vinf.argflag = 0;
		vinf.name = regist_name(mode, intrfuncs[i].stdname);
		j = 0;
		while (j < ARR_MAX) {
			vinf.array[j++] = -1;
		}
		vinf.next = NULL;
		funcinf = regist_function(mode, &vinf, intrfuncs[i].scode);
		if (argcnt < 0)
			argcnt = -argcnt;
		for (k = 0; k < argcnt; k++)
			regist_var(argmode, &intrfuncs[i].args[k]);
		end_function_param(mode, funcinf);
		funcinf->confirmed = 1;
		end_regist_function(mode, funcinf, intrfuncs[i].scode);
		funcinf->argcnt = intrfuncs[i].argcnt;

		pvinf = find_var(MODE_LOCAL, vinf.name);
		pos = garea + pvinf->offset;
		(*(int *)pos) = funcinf->ic;
	}
}

/*****************************************************************************/
/* 鲼ϥǥХåѤδؿǤ                                        */
/*****************************************************************************/
/* functions for debugging */
#ifdef DEBUG
#if 0
static void debug_skip(int type1, int eq, int type2)
{
	int flag = 0;

	if (eq) {
		while (tok->type == type1)
			GetToken();
	}
	else {
		while (1) {
			if (tok->type == type2)
				flag++;
			if (tok->type == type1 && flag == 0)
				break;
			if (tok->type == type1 && flag > 0)
				flag--;
			GetToken();
		}
	}
}
#endif
#endif
