/*
 *  table.c - Functions treating table data
 *
 *  Copyright (C) 2003 Seiji Tamura
 */
#include <stdio.h>
#include "common.h"

/* memo */
/* ¤Ǥι¤Բ(̾/ʤԲ)       */
/* ¤Ʊ̾ΥСͭ... */

#define NAMEBUFFSIZE 2048
static char gnbuf[NAMEBUFFSIZE];
static char gsnbuf[NAMEBUFFSIZE];
static char lnbuf[NAMEBUFFSIZE];
static char lsnbuf[NAMEBUFFSIZE];
static int gn_idx;
static int gsn_idx;
static int ln_idx;
static int lsn_idx;

#define VARBUFFSIZE 256
static varinf_t gvbuf[VARBUFFSIZE];
static varinf_t gsvbuf[VARBUFFSIZE];
static varinf_t lvbuf[VARBUFFSIZE];
//static varinf_t lsvbuf[VARBUFFSIZE];
static int gv_idx;
static int gsv_idx;
static int lv_idx;
//static int lsv_idx;
#define LOCAL_BLOCK_MAX 16
static int local_block = 0;
#define VARHASHSIZE 179
static int gvhash[VARHASHSIZE];
static int gsvhash[VARHASHSIZE];
static int lvhash_arr[LOCAL_BLOCK_MAX][VARHASHSIZE];
static int *lvhash;
//static int lsvhash[VARHASHSIZE];

#define STUNIBUFFSIZE 64
static stuninf_t gsubuf[STUNIBUFFSIZE];
static stuninf_t gssubuf[STUNIBUFFSIZE];
static stuninf_t lsubuf[STUNIBUFFSIZE];
static stuninf_t lssubuf[STUNIBUFFSIZE];
static int gsu_idx;
static int gssu_idx;
static int lsu_idx;
static int lssu_idx;
static varinf_t **struct_member;

#define FUNCBUFFSIZE 128
static funcinf_t gfuncbuf[FUNCBUFFSIZE];
static funcinf_t sfuncbuf[FUNCBUFFSIZE];
static int gfunc_idx;
static int sfunc_idx;
static varinf_t **func_args;

static int goffset;
static int poffset;
static int loffset;

#define LABELBUFFSIZE 512
int labelbuf[LABELBUFFSIZE];
static int label;
switchlblinf_t swlabelbuf[LABELBUFFSIZE];
static int swlblcnt;

#define NAMEDLABELBUFFSIZE 32
namedlblinf_t namedlblbuf[NAMEDLABELBUFFSIZE];
static int nlblcnt;

#define FUNCTINFOSIZE 256
static type_info gfunc_tinfo[FUNCTINFOSIZE];
static type_info sfunc_tinfo[FUNCTINFOSIZE];
static int gfunc_tinfo_idx;
static int sfunc_tinfo_idx;

#define MACROINFSIZE 128
static macroinf_t macroinfbuf[MACROINFSIZE];
static int macroinf_idx;

#define TYPEDEFINFSIZE 128
static typedefinf_t typedefinfbuf[TYPEDEFINFSIZE];
static int typedefinf_idx;

/* global functions */
int init_file_scope_table();

/* file scope functions */
static int search_named_label(char *name);
static int chk_var(varinf_t *v1, varinf_t *v2);
static int chk_arr(varinf_t *o, varinf_t *n, int *offset);
static int get_type_size(varinf_t *pvinf);
static int get_struct_size(varinf_t *pvinf);
static varinf_t *search_var(varinf_t *, int, char *, int *);
static int insert_var(varinf_t *, int, char *, int *);
static int regist_function_param(int, varinf_t *);
static int init_local_table();

#ifdef DEBUG
static void print_mode(int mode);
static void print_local_struct();
static void print_local_variable();
#endif

void init_table(int mode)
{
	if (mode) {
		gn_idx = 0;
		gsn_idx = 0;
		ln_idx = 0;
		lsn_idx = 0;
		gsu_idx = 0;
		gssu_idx = 0;
		lsu_idx = 0;
		lssu_idx = 0;
		gv_idx = 0;
		gsv_idx = 0;
		lv_idx = 0;
//		lsv_idx = 0;

		struct_member = 0;

		gfunc_idx = 0;
		sfunc_idx = 0;

		gfunc_tinfo_idx = 0;
		sfunc_tinfo_idx = 0;

		goffset = 0;
		poffset = 0;
		loffset = 0;

		local_block = 0;
		lvhash = lvhash_arr[local_block];
		init_hash(lvhash, VARHASHSIZE, -1);
//		init_hash(lsvhash, VARHASHSIZE, -1);
		init_hash(gvhash, VARHASHSIZE, -1);
		init_hash(gsvhash, VARHASHSIZE, -1);

		macroinf_idx = 0;
		typedefinf_idx = 0;
	}
	swlblcnt = 0;
}

int init_label()
{
	label = 0;
	return(0);
}

int label_next()
{
	return(label++);
}

int regist_label(int idx, int pc)
{
	labelbuf[idx] = pc;
	return(0);
}

int regist_sw_label(int brk, int pc, int val, int def)
{
	if (swlblcnt + 1 >= LABELBUFFSIZE)
		fatal("error: label buffer overflow");
	swlabelbuf[swlblcnt].brk = brk;
	swlabelbuf[swlblcnt].pc = pc;
	swlabelbuf[swlblcnt].val = val;
	swlabelbuf[swlblcnt].def = def;
	return(swlblcnt++);
}

switchlblinf_t *pop_sw_label(int brk)
{
	if (swlblcnt - 1 < 0)
		return(NULL);
//		fatal("System error: pop_sw_label");
#ifdef DEBUG
//	printf("swlabelbuf[%d - 1].brk %d, brk %d\n",
//	       swlblcnt, swlabelbuf[swlblcnt - 1].brk, brk);
#endif
	if (swlabelbuf[swlblcnt - 1].brk != brk)
		return(NULL);
	return(&swlabelbuf[--swlblcnt]);
}

int init_named_label()
{
	nlblcnt = 0;
	return(0);
}

int regist_named_label(char *name, int pc)
{
	int idx;

	if (nlblcnt >= NAMEDLABELBUFFSIZE)
		fatal("error: named label buffer overflow");

	idx = search_named_label(name);
	if (idx >= 0 && namedlblbuf[idx].pc >= 0 && pc >= 0)
		yyerror("label redeclaration found");

	if (idx >= 0) {
		if (pc >= 0)
			namedlblbuf[idx].pc = pc;
	}
	else {
		/* ʸnameĹIDSIZEʲ */
		copystr(namedlblbuf[nlblcnt].name, name);
		namedlblbuf[nlblcnt].pc = pc;
		idx = nlblcnt;
		nlblcnt++;
	}

	return(idx);
}

static int search_named_label(char *name)
{
	int i;

	for (i = 0; i < nlblcnt; i++)
		if (compstr(namedlblbuf[i].name, name) == 0)
			return(i);

	return(-1);
}

int regist_macro(macroinf_t *macro)
{
	if (macroinf_idx >= MACROINFSIZE)
		fatal("error: macro buffer overflow");
	macroinfbuf[macroinf_idx] = *macro;
	return(macroinf_idx++);
}

macroinf_t *search_macro(char *name)
{
	int i;

	for (i = macroinf_idx - 1; i >= 0; i--)
		if (compstr(macroinfbuf[i].name, name) == 0)
			return(&macroinfbuf[i]);

	return(NULL);
}

int regist_typedef(typedefinf_t *tdefinf)
{
	if (typedefinf_idx >= TYPEDEFINFSIZE)
		fatal("error: typedef buffer overflow");
	typedefinfbuf[typedefinf_idx] = *tdefinf;
	return(typedefinf_idx++);
}

typedefinf_t *search_typedef(char *name)
{
	int i;

	for (i = typedefinf_idx - 1; i >= 0; i--)
		if (compstr(typedefinfbuf[i].name, name) == 0)
			return(&typedefinfbuf[i]);

	return(NULL);
}

char *regist_name(int mode, char *name)
{
	char *pos;
	int n;
	char *namebuffer;
	int *name_idx;

#ifdef DEBUG
//	print_mode(mode);
#endif
	if (mode & MODE_STATIC) {
		if (mode & MODE_LOCAL) {
			namebuffer = &lsnbuf[lsn_idx];
			name_idx = &lsn_idx;
		}
		else {
			namebuffer = &gsnbuf[gsn_idx];
			name_idx = &gsn_idx;
		}
	}
	else {
		if (mode & MODE_LOCAL) {
			namebuffer = &lnbuf[ln_idx];
			name_idx = &ln_idx;
		}
		else {
			namebuffer = &gnbuf[gn_idx];
			name_idx = &gn_idx;
		}
	}

	n = 0;
	pos = name;
	while (*pos) {
		pos++;
		n++;
	}
	n++;
	if (n + *name_idx >= NAMEBUFFSIZE) {
		fatal("identifier character buffer overflow");
	}
	*name_idx += copystr(namebuffer, name);
	*name_idx += 1; /* for null termination char */
#ifdef DEBUG
//	printf("name: %s\n", namebuffer);
#endif
	return(namebuffer);
}

int correct_var_offset(int mode, varinf_t *pvinf)
{
	int *offset;

	if (mode & MODE_STATIC) {
		offset = &goffset;
	}
	else {
		if (mode & MODE_LOCAL) {
			if (mode & MODE_PARAM)
				offset = &poffset;
			else
				offset = &loffset;
		}
		else {
			offset = &goffset;
		}
	}

	if (pvinf->type == TYPENAME) {
		pvinf->size = get_type_size(pvinf);
	}
	else if (pvinf->type == STRUCT || pvinf->type == UNION) {
		pvinf->size = get_struct_size(pvinf);
	}
	else {
		; /* do nothing */
	}
	(*offset) = pvinf->offset;
	(*offset) += pvinf->size;
	return(*offset);
}

static int chk_var(varinf_t *v1, varinf_t *v2)
{
	if ((v1->type != v2->type) || (v1->ptflag != v2->ptflag))
		return(-1);
	if ((v1->type == TYPENAME) && (v1->ttype != v2->ttype))
		return(-2);
	if ((v1->type == STRUCT) || (v1->type == UNION))
		if (compstr(v1->tname, v2->tname))
			return(-3);
	return(0);
}

static int chk_arr(varinf_t *o, varinf_t *n, int *offset)
{
	int i;

	if (o->array[0] < 0)
		return(0);

	i = 0;
	while ((o->array[i] >= 0 || n->array[i] >= 0) && o->array[i] != 0) {
		if (o->array[i] != n->array[i])
			return(-1);
		if (++i >= ARR_MAX)
			break;
	}

	if (o->array[0] == 0 && n->array[0] > 0) {
		if (o->type == TYPENAME) {
			o->size = get_type_size(n);
			o->tsize = o->tsize;
		}
		else if (o->type == STRUCT || o->type == UNION) {
			o->size = get_struct_size(n);
			o->tsize = n->tsize;
		}
		else {
			o->size = n->size;
			o->tsize = o->size;
		}
		o->offset = *offset;
		(*offset) += o->size;
	}
	return(0);
}

varinf_t *regist_var(int mode, varinf_t *pvinf)
{
	varinf_t *vinf, *topvinf;
	int *v_idx;
	int i;
	int base;
	int *offset;
	int *htbl;
	int externflg = 0;

	base = BASE_GLOBAL;
	if (mode & MODE_STATIC) {
		if (mode & MODE_LOCAL) {
//			topvinf = lsvbuf;
			topvinf = lvbuf;
//			vinf = &lsvbuf[lsv_idx];
			vinf = &lvbuf[lv_idx];
//			v_idx = &lsv_idx;
			v_idx = &lv_idx;
//			htbl = lsvhash;
			htbl = lvhash;
		}
		else {
			topvinf = gsvbuf;
			vinf = &gsvbuf[gsv_idx];
			v_idx = &gsv_idx;
			htbl = gsvhash;
		}
		offset = &goffset;
	}
	else {
		if (mode & MODE_LOCAL) {
			topvinf = lvbuf;
			vinf = &lvbuf[lv_idx];
			v_idx = &lv_idx;
			htbl = lvhash;
			if (mode & MODE_PARAM) {
				base = BASE_PARAM;
				offset = &poffset;
//				offset = &loffset;
			}
			else {
				base = BASE_LOCAL;
				offset = &loffset;
			}
		}
		else {
			topvinf = gvbuf;
			vinf = &gvbuf[gv_idx];
			v_idx = &gv_idx;
			htbl = gvhash;
			offset = &goffset;
		}
	}

	if (*v_idx >= VARBUFFSIZE) {
		fatal("variable symbol table buffer overflow");
	}

	vinf->same = NULL;
	if (struct_member == NULL) {
		varinf_t *tmpvinf;

		/* K&Rΰξϲ⤻ */
		if (mode & MODE_EXTERN && mode & MODE_PARAM)
			return(NULL);

		/* ѿ򸡺 */
		tmpvinf = search_var(topvinf, *v_idx, pvinf->name, htbl);
		if (tmpvinf) {
			/* externξ */
			if (mode & MODE_EXTERN) {
				return(NULL);
			}
			else {
				if (tmpvinf->confirmed)
					yyerror("redeclaration found");

				if ((tmpvinf->scope & ~MODE_EXTERN) ==
				    (mode & ~MODE_EXTERN)) {
					char *e1="conflicting types";
					char *e2="conflicting types (array)";
					if (chk_var(tmpvinf, pvinf))
						yyerror(e1);
					if (chk_arr(tmpvinf,pvinf,offset))
						yyerror(e2);
					tmpvinf->confirmed = 1;
					return(tmpvinf);
				}
			}
		}
		else
			externflg = (mode & MODE_EXTERN)?(1):(0);

		insert_var(topvinf, *v_idx, pvinf->name, htbl);
	}

	vinf->name = pvinf->name;
	vinf->scope = pvinf->scope;
	vinf->type = pvinf->type;
	if (vinf->type == TYPENAME) {
		vinf->size = get_type_size(pvinf);
		vinf->tsize = pvinf->tsize;
	}
	else if (vinf->type == STRUCT || vinf->type == UNION) {
		vinf->size = get_struct_size(pvinf);
		vinf->tsize = pvinf->tsize;
	}
	else {
		vinf->size = pvinf->size;
		vinf->tsize = vinf->size;
	}
	vinf->base = base;
//	vinf->offset = pvinf->offset;
	vinf->offset = *offset;
	if (pvinf->argflag == 1)
		(*offset) += sizeof(void *);
	else
		(*offset) += vinf->size;
	vinf->ptflag = pvinf->ptflag;
	vinf->argflag = pvinf->argflag;
	i = 0;
	while (i < ARR_MAX) {
		vinf->array[i] = pvinf->array[i];
		i++;
	}
	if (vinf->type == STRUCT || vinf->type == UNION) {
		vinf->tname = pvinf->tname;
	}
	else {
		vinf->ttype = pvinf->ttype;
	}
	vinf->confirmed = (externflg)?(0):(1);
	vinf->funcptr = pvinf->funcptr;
	vinf->next = pvinf->next;
	vinf->stuni = pvinf->stuni;
	(*v_idx)++;
	if (struct_member) {
		*struct_member = vinf;
		struct_member = &vinf->next;
		
	}
	if (func_args) {
		*func_args = vinf;
		func_args = &vinf->next;
	}
	return(vinf);
}

static int get_ttype_size(int type)
{
	int size;

	switch (type) {
	case TYPE_CHAR:
	case TYPE_UCHAR:
		size = sizeof(char);
		break;
	case TYPE_DOUBLE:
	case TYPE_FLOAT:
		size = sizeof(double);
		break;
	case TYPE_INT:
	case TYPE_LONG:
	case TYPE_SHORT:
	case TYPE_UINT:
	case TYPE_ULONG:
	case TYPE_USHORT:
		size = sizeof(int);
		break;
	case TYPE_VOID:
		size = sizeof(void);
		break;
	default:
		return(-1);
	}
	return(size);
}

static int get_type_size(varinf_t *pvinf)
{
	int size;
	int i;
	int array_size;

	switch (pvinf->ttype) {
	case TYPE_CHAR:
	case TYPE_UCHAR:
		if (pvinf->ptflag)
			size = sizeof(char *);
		else
			size = sizeof(char);
		break;
	case TYPE_DOUBLE:
	case TYPE_FLOAT:
		if (pvinf->ptflag)
			size = sizeof(double *);
		else
			size = sizeof(double);
		break;
//	case TYPE_FLOAT:
//		if (pvinf->ptflag)
//			size = sizeof(float *);
//		else
//			size = sizeof(float);
//		break;
	case TYPE_INT:
	case TYPE_UINT:
		if (pvinf->ptflag)
			size = sizeof(int *);
		else
			size = sizeof(int);
		break;
	case TYPE_LONG:
	case TYPE_ULONG:
		if (pvinf->ptflag)
			size = sizeof(long *);
		else
			size = sizeof(long);
		break;
	case TYPE_SHORT:
	case TYPE_USHORT:
		if (pvinf->ptflag)
			size = sizeof(short *);
		else
			size = sizeof(short);
		break;
	case TYPE_VOID:
		if (pvinf->ptflag)
			size = sizeof(void *);
		else
			size = sizeof(void);
		break;
	default:
		if (pvinf->ptflag)
			size = sizeof(int *);
		else
			size = sizeof(int);
		break;
	}
	pvinf->tsize = get_ttype_size(pvinf->ttype);
//	pvinf->tsize = size;
	array_size = 1;
	for (i = 0; i < ARR_MAX; i++) {
		if (pvinf->array[i] > 0)
			array_size *= pvinf->array[i];
	}
	size *= array_size;
	return(size);
}

static int get_struct_size(varinf_t *pvinf)
{
	int size;
	int i;
	int array_size;
	stuninf_t *stinf;

	stinf = pvinf->stuni;
	if (pvinf->ptflag)
		size = sizeof(void *);
	else
		size = stinf->size;

	pvinf->tsize = size;
	array_size = 1;
	for (i = 0; i < ARR_MAX; i++) {
		if (pvinf->array[i] > 0)
			array_size *= pvinf->array[i];
	}
	size *= array_size;
	return(size);
}

varinf_t *find_var(int mode, char *vname)
{
	varinf_t *vinf;

	if (mode & MODE_LOCAL) {
		int cur_block;
		int *tmphash;
#if 0
		/* 륹ƥå򸡺 */
//		vinf = search_var(lsvbuf, lsv_idx, vname, lsvhash);
		vinf = search_var(lvbuf, lv_idx, vname, lvhash);
		if (vinf) {
			return(vinf);
		}

		/* 򸡺 */
		vinf = search_var(lvbuf, lv_idx, vname, lvhash);
		if (vinf) {
			return(vinf);
		}
#endif
		/* 򸡺 */
		if (local_block < LOCAL_BLOCK_MAX) {
			cur_block = local_block;
		}
		else {
			cur_block = LOCAL_BLOCK_MAX - 1;
		}
		while (cur_block >= 0) {
			tmphash = lvhash_arr[cur_block];
			vinf = search_var(lvbuf, lv_idx, vname, tmphash);
			if (vinf) {
				return(vinf);
			}
			cur_block--;
		}
	}

	/* Х륹ƥå򸡺 */
	vinf = search_var(gsvbuf, gsv_idx, vname, gsvhash);
	if (vinf) {
		return(vinf);
	}

	/* Х򸡺 */
	vinf = search_var(gvbuf, gv_idx, vname, gvhash);
	if (vinf) {
		return(vinf);
	}

	return(NULL);
}

static varinf_t *search_var(varinf_t *vinf, int max, char *vname, int *htbl)
{
	int hval;
	varinf_t *p;

	hval = hash_value(vname, VARHASHSIZE);
#ifdef DEBUG
//	printf("hval       : %d\n", hval);
//	printf("htbl[hval] : %d\n", htbl[hval]);
#endif
	if (htbl[hval] < 0) {
		return(NULL);
	}
	p = &vinf[htbl[hval]];
	while (p != NULL) {
		if (compstr(p->name, vname) == 0) {
			return(p);
		}
		p = p->same;
	}

	return(NULL);
}

static int insert_var(varinf_t *vinf, int idx, char *vname, int *htbl)
{
	int hval;

	hval = hash_value(vname, VARHASHSIZE);
	if (htbl[hval] >= 0) {
		vinf[idx].same = &vinf[htbl[hval]];
	}
	htbl[hval] = idx;

	return(0);
}

stuninf_t *regist_struct(int mode, token_t *tok)
{
	stuninf_t *stinf;
	int *stuni_idx;

	if (mode & MODE_STATIC) {
		if (mode & MODE_LOCAL) {
			stinf = &lssubuf[lssu_idx];
			stuni_idx = &lssu_idx;
		}
		else {
			stinf = &gssubuf[gssu_idx];
			stuni_idx = &gssu_idx;
		}
	}
	else {
		if (mode & MODE_LOCAL) {
			stinf = &lsubuf[lsu_idx];
			stuni_idx = &lsu_idx;
		}
		else {
			stinf = &gsubuf[gsu_idx];
			stuni_idx = &gsu_idx;
		}
	}

	if (*stuni_idx >= STUNIBUFFSIZE) {
		fatal("struct symbol table buffer overflow");
	}

	stinf->name = regist_name(mode, tok->ident);
	stinf->scope = mode;
	stinf->type = tok->type;
	stinf->size = sizeof(void *);	/* Ȥꤢݥ󥿥򥻥å */
	stinf->var = NULL;
	struct_member = &stinf->var;
	(*stuni_idx)++;

	return(stinf);
}

int end_regist_struct(int mode, stuninf_t *stinf)
{
	varinf_t *vinf;
	int size;

	size = 0;
	vinf = stinf->var;
	do {
		vinf->offset = size; ////////////////////////
		size += vinf->size;
		
	} while (vinf = vinf->next, vinf);
	stinf->size = size;
	struct_member = NULL;

	return(0);
}

static stuninf_t *search_struct(stuninf_t *, int, char *);

/* Ȥꤢϵ */
stuninf_t *find_struct(int mode, char *tname)
{
	stuninf_t *stinf;

	if (mode & MODE_LOCAL) {
		/* 륹ƥå򸡺 */
		stinf = search_struct(lssubuf, lssu_idx, tname);
		if (stinf) {
			return(stinf);
		}

		/* 򸡺 */
		stinf = search_struct(lsubuf, lsu_idx, tname);
		if (stinf) {
			return(stinf);
		}
	}

	/* Х륹ƥå򸡺 */
	stinf = search_struct(gssubuf, gssu_idx, tname);
	if (stinf) {
		return(stinf);
	}

	/* Х򸡺 */
	stinf = search_struct(gsubuf, gsu_idx, tname);
	if (stinf) {
		return(stinf);
	}

	return(NULL);
}

static stuninf_t *search_struct(stuninf_t *stinf, int max, char *tname)
{
	int i;

	for (i = 0; i < max; i++) {
//		printf("stinf[%d].name = %s\n", i, stinf[i].name);
		if (compstr(stinf[i].name, tname) == 0) {
			return(&stinf[i]);
		}
	}
	return(NULL);
}

/* ¤ΤΥФꥹȤéäƸ */
varinf_t *find_struct_member(varinf_t *vinf, char *vname)
{
	varinf_t *tmpvinf;

	if (vinf->stuni == NULL)
		return(NULL);

//	printf("struct name: (%s)\n", vinf->name);
	tmpvinf = vinf->stuni->var;
	while (tmpvinf != NULL) {
//		printf("tmpvinf->name,vname: (%s,%s)\n", tmpvinf->name, vname);
		if (compstr(tmpvinf->name, vname) == 0)
			break;
		tmpvinf = tmpvinf->next;
	}
	return(tmpvinf);
}

funcinf_t *regist_function(int mode, varinf_t *pvinf, int flg)
{
	funcinf_t *funcinf, *tmpfinf;
	int *func_idx;
	int mask;

	/* ѿεɽ */
	init_local_table();

	if (mode & MODE_STATIC) {
		funcinf = &sfuncbuf[sfunc_idx];
		func_idx = &sfunc_idx;
	}
	else {
		funcinf = &gfuncbuf[gfunc_idx];
		func_idx = &gfunc_idx;
	}

	if (*func_idx >= FUNCBUFFSIZE) {
		fatal("function symbol table buffer overflow");
	}

	tmpfinf = find_function(mode, pvinf->name);
	mask = ~MODE_EXTERN;
	if (tmpfinf && !((tmpfinf->scope ^ mode) & mask)) {
		if (tmpfinf->confirmed) {
			if (!compstr(pvinf->name, "main"))
				yyerror("redefinition of \'main\'");
			return(tmpfinf);
		}
		funcinf = tmpfinf;
		func_idx = NULL;
	}

	funcinf->name = pvinf->name;
	funcinf->scope = mode;
	funcinf->argcnt = 0;
	funcinf->fsize = -1;
	funcinf->confirmed = 0;
	funcinf->ret = regist_var(mode, pvinf);
	funcinf->arg = NULL;
	func_args = &funcinf->arg;
	if (func_idx)
		(*func_idx)++;
	if (flg > 0)
		funcinf->ic = add_code_s(FENT, funcinf->name, 0, 0, 0);
	else
		funcinf->ic = flg;
	return(funcinf);
}

int end_function_param(int mode, funcinf_t *funcinf)
{
	varinf_t *vinf;
	int size;
	int argcnt;
	int argpos;
	/*int offset;*/

	size = 0;
	argcnt = 0;
	argpos = 0;
	vinf = funcinf->arg;
	while (vinf) {
		size += vinf->size;
		argpos = regist_function_param(mode, vinf);
		argcnt++;
		vinf = vinf->next;
	}
	funcinf->argsize = size;
	funcinf->argcnt = argcnt;
	funcinf->argpos = argpos - (argcnt - 1);
	/* δƤӽФǹԤ褦ѹ 20051114 */ //XXXXXXX
	/*offset = -size;
	vinf = funcinf->arg;
	while (vinf) {
		vinf->offset = offset;
		offset += vinf->size;
		vinf = vinf->next;
	}*/
//	struct_member = NULL; ///////////////////////???
//	func_args = NULL;
	return(0);
}

int end_regist_function(int mode, funcinf_t *funcinf, int flg)
{
	varinf_t *vinf;
	int size;

	size = 0;
	vinf = funcinf->arg;
	while (vinf) {
		size += vinf->size;
		vinf = vinf->next;
	}
	/* ե졼ॵ˰Υޤʤ褦ѹ 20051114 */
//	funcinf->fsize = size;
	funcinf->fsize = size - funcinf->argsize;
	func_args = NULL;
	if (flg > 0)
		add_code(FRET, size, 0, 0, 0);
#ifdef DEBUG
	printf("----- print_local_struct -----\n");
	print_local_struct();
	printf("------ print_local_variable ------\n");
	print_local_variable();
#endif
	return(0);
}

int end_prototype()
{
	func_args = NULL;
	return(0);
}

static funcinf_t *search_function(funcinf_t *, int, char *);

/* Ȥꤢϵ */
funcinf_t *find_function(int mode, char *fname)
{
	funcinf_t *finf;

	/* ƥåؿ򸡺 */
	finf = search_function(sfuncbuf, sfunc_idx, fname);
	if (finf) {
		return(finf);
	}

	/* Хؿ򸡺 */
	finf = search_function(gfuncbuf, gfunc_idx, fname);
	if (finf) {
		return(finf);
	}

	return(NULL);
}

static funcinf_t *search_function(funcinf_t *funcinf, int max, char *fname)
{
	int i;

	for (i = 0; i < max; i++) {
		if (compstr(funcinf[i].name, fname) == 0) {
			return(&funcinf[i]);
		}
	}
	return(NULL);
}

static int regist_function_param(int mode, varinf_t *vinf)
{
	type_info *func_tinfo;
	int *func_tinfo_idx;

	if (mode & MODE_STATIC) {
		func_tinfo = &sfunc_tinfo[sfunc_tinfo_idx];
		func_tinfo_idx = &sfunc_tinfo_idx;
	}
	else {
		func_tinfo = &gfunc_tinfo[gfunc_tinfo_idx];
		func_tinfo_idx = &gfunc_tinfo_idx;
	}

	if (*func_tinfo_idx >= FUNCTINFOSIZE) {
		fatal("function param info table buffer overflow");
	}

	if (vinf->type == TYPENAME)
		func_tinfo->type = vinf->ttype;
	else
		func_tinfo->type = vinf->type;
	func_tinfo->size = vinf->size;
	if (vinf->array[0] != -1)
		func_tinfo->pflg = 1;
	else
		func_tinfo->pflg = vinf->ptflag;

	/*printf("*func_tinfo_idx %d\n", *func_tinfo_idx);*/
	return((*func_tinfo_idx)++);
}

type_info *get_function_param(int mode, int idx)
{
	type_info *func_tinfo;
	int *func_tinfo_idx;

	if (mode & MODE_STATIC) {
		func_tinfo = sfunc_tinfo;
		func_tinfo_idx = &sfunc_tinfo_idx;
	}
	else {
		func_tinfo = gfunc_tinfo;
		func_tinfo_idx = &gfunc_tinfo_idx;
	}

	if (idx >= *func_tinfo_idx)
		func_tinfo = NULL;
	else
		func_tinfo = &func_tinfo[idx];

	return(func_tinfo);
}

int inc_local_block()
{
	int cur_block;
	int init_flag;

	local_block++;
	if (local_block < LOCAL_BLOCK_MAX) {
		cur_block = local_block;
		init_flag = 1;
	}
	else {
		cur_block = LOCAL_BLOCK_MAX - 1;
		init_flag = 0;
	}
	lvhash = lvhash_arr[cur_block];
	if (init_flag)
		init_hash(lvhash, VARHASHSIZE, -1);

	return(0);
}

int dec_local_block()
{
	int cur_block;

	local_block--;
	if (local_block < LOCAL_BLOCK_MAX) {
		cur_block = local_block;
	}
	else {
		cur_block = LOCAL_BLOCK_MAX - 1;
	}
	lvhash = lvhash_arr[cur_block];

	return(0);
}

static int init_local_table()
{
	ln_idx = 0;
	lsn_idx = 0;

	lv_idx = 0;
//	lsv_idx = 0;

	lsu_idx = 0;
	lssu_idx = 0;

	poffset = 0;
	loffset = 0;

	local_block = 0;
	lvhash = lvhash_arr[local_block];
	init_hash(lvhash, VARHASHSIZE, -1);
//	init_hash(lsvhash, VARHASHSIZE, -1);

	return(0);
}

int init_file_scope_table()
{
	gsn_idx = 0;

	gsv_idx = 0;

	gssu_idx = 0;

	sfunc_idx = 0;

	sfunc_tinfo_idx = 0;

	init_hash(gsvhash, VARHASHSIZE, -1);

	macroinf_idx = 0;

	typedefinf_idx = 0;

	return(0);
}

/*****************************************************************************/
/* 鲼ϥǥХåѤδؿǤ                                        */
/*****************************************************************************/
/* functions for debugging */
#ifdef DEBUG
static void print_mode(int mode)
{
	if (mode & MODE_STATIC) {
		if (mode & MODE_LOCAL) {
			printf("local static  : ");
		}
		else {
			printf("global static : ");
		}
	}
	else {
		if (mode & MODE_LOCAL) {
			if (mode & MODE_PARAM)
				printf("param         : ");
			else
				printf("local         : ");
		}
		else {
			printf("global        : ");
		}
	}
}

void print_all_function()
{
	int i;
	varinf_t *vinf;

	for (i = 0; i < gfunc_idx; i++) {
		print_mode(gfuncbuf[i].scope);
		printf("gfuncbuf[%d].name = %s\n", i, gfuncbuf[i].name);
		vinf = gfuncbuf[i].arg;
		do {
			if (vinf == NULL)
				break;
			printf("\t%s ", vinf->name);
			printf("size = %d\n", vinf->size);
		} while (vinf = vinf->next, vinf);
	}
	for (i = 0; i < sfunc_idx; i++) {
		print_mode(sfuncbuf[i].scope);
		printf("sfuncbuf[%d].name = %s\n", i, sfuncbuf[i].name);
		vinf = sfuncbuf[i].arg;
		do {
			if (vinf == NULL)
				break;
			printf("\t%s ", vinf->name);
			printf("size = %d\n", vinf->size);
		} while (vinf = vinf->next, vinf);
	}
}

void print_global_struct()
{
	int i;
	varinf_t *vinf;

	for (i = 0; i < gsu_idx; i++) {
		print_mode(gsubuf[i].scope);
		printf("gsubuf[%d].name = %s ", i, gsubuf[i].name);
		printf("size = %d\n", gsubuf[i].size);
		vinf = gsubuf[i].var;
		do {
			printf("\t%s ", vinf->name);
			printf("size = %d\n", vinf->size);
		} while (vinf = vinf->next, vinf);
	}
	for (i = 0; i < gssu_idx; i++) {
		print_mode(gssubuf[i].scope);
		printf("gssubuf[%d].name = %s ", i, gssubuf[i].name);
		printf("size = %d\n", gssubuf[i].size);
		vinf = gssubuf[i].var;
		do {
			printf("\t%s ", vinf->name);
			printf("size = %d\n", vinf->size);
		} while (vinf = vinf->next, vinf);
	}
}

void print_global_variable()
{
	int i;
	for (i = 0; i < gv_idx; i++) {
		if (gvbuf[i].type == STRUCT || gvbuf[i].type == UNION) {
			printf("struct %s\n", gvbuf[i].tname);
		}
		print_mode(gvbuf[i].scope);
		printf("gvbuf[%d].name = %s ", i, gvbuf[i].name);
		printf("size = %d\n", gvbuf[i].size);
	}
	for (i = 0; i < gsv_idx; i++) {
		if (gsvbuf[i].type == STRUCT || gsvbuf[i].type == UNION) {
			printf("struct %s\n", gsvbuf[i].tname);
		}
		print_mode(gsvbuf[i].scope);
		printf("gsvbuf[%d].name = %s ", i, gsvbuf[i].name);
		printf("size = %d\n", gsvbuf[i].size);
	}
}

static void print_local_struct()
{
	int i;
	varinf_t *vinf;

	for (i = 0; i < lsu_idx; i++) {
		print_mode(lsubuf[i].scope);
		printf("lsubuf[%d].name = %s ", i, lsubuf[i].name);
		printf("size = %d\n", lsubuf[i].size);
		vinf = lsubuf[i].var;
		do {
			printf("\t%s ", vinf->name);
			printf("size = %d\n", vinf->size);
		} while (vinf = vinf->next, vinf);
	}
	for (i = 0; i < lssu_idx; i++) {
		print_mode(lssubuf[i].scope);
		printf("lssubuf[%d].name = %s ", i, lssubuf[i].name);
		printf("size = %d\n", lssubuf[i].size);
		vinf = lssubuf[i].var;
		do {
			printf("\t%s ", vinf->name);
			printf("size = %d\n", vinf->size);
		} while (vinf = vinf->next, vinf);
	}
}

static void print_local_variable()
{
	int i;

	for (i = 0; i < lv_idx; i++) {
		if (lvbuf[i].type == STRUCT || lvbuf[i].type == UNION) {
			printf("struct %s\n", lvbuf[i].tname);
		}
		print_mode(lvbuf[i].scope);
		printf("lvbuf[%d].name = %s ", i, lvbuf[i].name);
		printf("size = %d\n", lvbuf[i].size);
	}
#if 0
	for (i = 0; i < lsv_idx; i++) {
		if (lsvbuf[i].type == STRUCT || lsvbuf[i].type == UNION) {
			printf("struct %s\n", lsvbuf[i].tname);
		}
		print_mode(lsvbuf[i].scope);
		printf("lsvbuf[%d].name = %s ", i, lsvbuf[i].name);
		printf("size = %d\n", lsvbuf[i].size);
	}
#endif
}
#endif
