#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CODE_SIZE    512
#define STACK_SIZE   512

#define TYPE_EOF     256
#define TYPE_INT     257
#define TYPE_PLUS    258
#define TYPE_MINUS   259
#define TYPE_MUL     260
#define TYPE_DIV     261
#define TYPE_REMAIN  262
#define TYPE_LPER    263
#define TYPE_RPER    264
#define TYPE_UNKNOWN 265

#define PUSH         1
#define POP          2
#define ADD          3
#define SUB          4
#define MUL          5
#define DIV          6
#define REMAIN       7
#define PRINT        8
#define INVT         9

typedef struct token {
	int type;
	int val;
} token_t;

typedef struct code {
	int op;
	int o1;
} code_t;

static FILE *ifp;
static int c = ' ';
static token_t tok;
static code_t code[CODE_SIZE];
static int ic = 0;
static int stack[STACK_SIZE];
static int sp = 0;

static int interpret();
static int add_code(int op, int o1);
static int cget();
static int get_int();
static token_t *get_token();
static void expr();
static void expr1();
static void expr2();
static void expr3();
static void expr4();
static void expr5();

static int interpret()
{
	int i;
	int val;
	for (i = 0; i < ic; i++) {
		//printf("%8d%8d\n", code[i].op, code[i].o1);
		switch (code[i].op) {
		case PUSH:
			if (sp >= STACK_SIZE) {
				printf("X^bNI[o[t[\n");
				return(-1);
			}
			stack[sp++] = code[i].o1;
			break;
		case POP:
			val = stack[--sp];
			break;
		case ADD:
			val = stack[--sp];
			val += stack[--sp];
			stack[sp++] = val;
			break;
		case SUB:
			val = stack[--sp];
			val = stack[--sp] - val;
			stack[sp++] = val;
			break;
		case MUL:
			val = stack[--sp];
			val *= stack[--sp];
			stack[sp++] = val;
			break;
		case DIV:
			val = stack[--sp];
			val = stack[--sp] / val;
			stack[sp++] = val;
			break;
		case REMAIN:
			val = stack[--sp];
			val = stack[--sp] % val;
			stack[sp++] = val;
			break;
		case PRINT:
			val = stack[--sp];
			printf("ans> %d\n", val);
			break;
		case INVT:
			val = stack[--sp] * (-1);
			stack[sp++] = val;
			break;
		default:
			break;
		}
	}
	//printf("sp = %d\n", sp);
	return(0);
}

static int add_code(int op, int o1)
{
	if (ic >= CODE_SIZE) {
		printf("R[hTCY܂\n");
	}
	code[ic].op = op;
	code[ic].o1 = o1;
	return(ic++);
}

static int cget()
{
	return(c = fgetc(ifp));
}

static int get_int()
{
	int val = c - '0';
	cget();
	while (c >= '0' && c <= '9') {
		val = val * 10 + (c - '0');
		cget();
	}
	return(val);
}

static token_t *get_token()
{
	while (c == ' ' || c == '\t') {
		cget();
	}
	if (c == '\n') {
		tok.type = TYPE_EOF;
		return(&tok);
	}
	if (c >= '0' && c <= '9') {
		tok.type = TYPE_INT;
		tok.val = get_int();
		return(&tok);
	}
	switch (c) {
	case '+':
		tok.type = TYPE_PLUS;
		break;
	case '-':
		tok.type = TYPE_MINUS;
		break;
	case '*':
		tok.type = TYPE_MUL;
		break;
	case '/':
		tok.type = TYPE_DIV;
		break;
	case '%':
		tok.type = TYPE_REMAIN;
		break;
	case '(':
		tok.type = TYPE_LPER;
		break;
	case ')':
		tok.type = TYPE_RPER;
		break;
	default:
		tok.type = TYPE_UNKNOWN;
		break;
	}
	cget();
	return(&tok);
}

static void expr()
{
	expr1();
	add_code(PRINT, 0);
}

/* + - */
static void expr1()
{
	int type;
	expr2();
	type = tok.type;
	while (type == TYPE_PLUS || type == TYPE_MINUS) {
		get_token();
		expr2();
		if (type == TYPE_PLUS) {
			//printf("+");
			add_code(ADD, 0);
		}
		else {
			//printf("-");
			add_code(SUB, 0);
		}
		type = tok.type;
	}
}

/* * / % */
static void expr2()
{
	int type;
	expr3();
	type = tok.type;
	while (type == TYPE_MUL || type == TYPE_DIV || type == TYPE_REMAIN) {
		get_token();
		expr3();
		if (type == TYPE_MUL) {
			//printf("*");
			add_code(MUL, 0);
		}
		else if (type == TYPE_DIV) {
			//printf("/");
			add_code(DIV, 0);
		}
		else {
			//printf("%");
			add_code(REMAIN, 0);
		}
		type = tok.type;
	}
}

static void expr3()
{
	int type;
	type = tok.type;
	if (type == TYPE_MINUS) {
		get_token();
		expr3();
		add_code(INVT, 0);
	}
	else {
		expr4();
	}
}

static void expr4()
{
	int type;
	type = tok.type;
	expr5();
	if (type == TYPE_INT) {
		//printf("%d", tok.val);
		add_code(PUSH, tok.val);
		get_token();
	}
}

static void expr5()
{
	int type;
	type = tok.type;
	if (type == TYPE_LPER) {
		//printf("(");
		get_token();
		expr1();
		type = tok.type;
		if (type == TYPE_RPER) {
			//printf(")");
			get_token();
		}
		else {
			//printf("[error\')\']");
		}
	}
}

int main(int argc, char **argv)
{
	printf("exp> ");
	ifp = stdin;
	get_token();
	expr();
	interpret();
	return(0);
}
